/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "config.hh"
#include <snpfile/matrix.hh>
#include <snpfile/array_matrix.hh>
using namespace BiRC::SNPFile;

#include "io.hh"
#include "tree.hh"
#include "tree_builder.hh"
#include "incompatibility.hh"
#include "score.hh"
#include "case_control_score.hh"
#include "qt_score.hh"
#include "genotype_score.hh"
#include "config.hh"

#include <gsl/gsl_errno.h>

#include <numeric>
#include <string>
#include <cmath>
#include <fstream>
#include <cstdio>
#include <sys/time.h>
#include <ext/algorithm>
#include <signal.h>

#include <boost/program_options.hpp>


namespace options {
    bool verbose = false;
    std::string positionsFileName;
    std::string haplotypesFileName;
    std::string scoresFileName("scores.txt");
    std::string pvalsFileName("p-values.txt");

    int noPerms = 0;
    int minMarkers = 10;
    double maxlen = -1.0;

    std::string scoreFunction;
    std::string interval;

    bool  quantitative = false;
    bool  unphased = true;

    int maxMutations = 1;
}

static int
getInterval(int &from, int &to, int noMarkers)
{
    if (options::interval.empty())
	{
	    from = 0;
	    to = noMarkers;
	    return 0;
	}

    std::string &s = options::interval;
    std::string::size_type pos = s.find(":",0);
    if (pos == std::string::npos)
	{
	    from = to = atol(s.c_str());
	}
    else
	{
	    from = atol(s.substr(0,pos).c_str());
	    to = atol(s.substr(pos+1,s.size()).c_str());
	}

    if (from < 0) return 1;
    if (to > noMarkers) return 1;
    return 0;
}


// signal handling to shut down properly if the user asks for termination
namespace {
    volatile bool termination_signalled = false;
}
static void 
signal_handler(int signum)
{
    termination_signalled = true;
}

static BasicScore *
selectScoringFunction(Matrix &haplotypes, std::vector<int> &positions,
		      Classification &classification)
{
    switch (options::scoreFunction[0])
	{
	case 'S': return new SMAScore(haplotypes, positions,
				      options::minMarkers, options::maxlen, 
				      classification);
	case 'B': return new BICScore(haplotypes, positions,
				      options::minMarkers, options::maxlen, 
				      classification);
	case 'N': return new newBICScore(haplotypes, positions,
					 options::minMarkers, options::maxlen, 
					 options::maxMutations,classification);
	case 'W': return new newHQScore(haplotypes, positions,
					options::minMarkers, options::maxlen, 
					options::maxMutations, classification);
	case 'P': return new ProbScore(haplotypes, positions,
				       options::minMarkers, options::maxlen,
				       classification);
	case 'H': return new HQScore(haplotypes, positions,
				     options::minMarkers, options::maxlen, 
				     classification);
	case 'A': return new AICScore(haplotypes, positions,
				      options::minMarkers, options::maxlen,
				      classification);
	case 'G': return new GiniScore(haplotypes, positions,
				       options::minMarkers, options::maxlen,
				       classification);
	case 'X': return new GenotypeScore(haplotypes, positions,
					   options::minMarkers, 
					   options::maxlen, 
					   classification,
					   options::maxMutations);
	}
    assert(!"shouldn't reach this point!");
    return 0;
}


static BasicScore *
selectScoringFunction(ImmutableMatrix &haplotypes, 
		      std::vector<int> &positions,
		      QTValue &qt_values)
{

    switch (options::scoreFunction[0])
	{
	case 'Q': return new QTQScore(haplotypes, positions,
				      options::minMarkers, options::maxlen,
				      qt_values);
	case 'B': return new QTnewBICScore(haplotypes, positions,
					   options::minMarkers, 
					   options::maxlen, 
					   options::maxMutations, qt_values);
 	case 'X': return new QTgBICScore(haplotypes, positions,
					 options::minMarkers, 
					 options::maxlen, 
					 options::maxMutations, qt_values);
	}
    assert(!"Shouldn't reach this point!");
    return 0;
}


static void
processCurrentSample(const Score *scoreFun, 
		     std::vector<long double> &scores,
		     int from, int to)
{
    const int verbose_step_size = (scores.size() < 100) ?
	1 : scores.size() / 100;

    using namespace TextManip;
    TextProgressBar progressBar;
    if (options::verbose) progressBar.start();

    for (int i = from; i < to && !termination_signalled; ++i)
	{
	    scores[i] = scoreFun->scoreLocus(i);
	    if (options::verbose and (i+1) % verbose_step_size == 0)
		progressBar.showProgress(i, scores.size());
	}

    if (options::verbose) progressBar.end();
}

#if 0
static void
permutationTest(Score *scoreFun, 
		const std::vector<long double> &scores,
		std::vector<long double> &pvals,
		int from, int to)
{
    const int small_verbose_step_size = (options::noPerms < 100) ?
	1 : options::noPerms / 100;

    using namespace TextManip;
    TextProgressBar progressBar;
    if (options::verbose) progressBar.start();

    std::fill(pvals.begin()+from, pvals.begin()+to, 0);
    for (int i = from; i < to && !termination_signalled; ++i)
	{
	    // FIXME: doesn't work for PPH where we need to sum over
	    // several trees!!!
	    std::auto_ptr<Tree> t(scoreFun->getTree(i));
	    for (int j = 0; j < options::noPerms; ++j)
		{
		    scoreFun->permute();
		    long double score = scoreFun->scoreTree(t.get());
		    if (score > scores[i]) ++pvals[i];
		    if (options::verbose and 
			(j+1) % small_verbose_step_size == 0)
			progressBar.showProgress(i, scores.size(),
						 j, options::noPerms);
		    if (termination_signalled)
			break;
		}
	}
    if (options::verbose) progressBar.end();
}
#endif


static int caseControl()
{
    std::ifstream positionsFile(options::positionsFileName.c_str());
    if (!positionsFile)
	{
	    std::cerr << "Error opening " 
		      << options::positionsFileName << std::endl;
	    return 2;
	}
    std::ifstream haplotypesFile(options::haplotypesFileName.c_str());
    if (!haplotypesFile)
	{
	    std::cerr << "Error opening "
		      << options::haplotypesFileName << std::endl;
	    return 2;
	}

    std::vector<int> positions;
    Matrix haplotypes(newArrayMatrix(0,0));
    std::vector<int> classes;

    try {
	positions = readPositions(positionsFile);
    } catch (UnsortedPositions&) {
	std::cerr << "The positions in " << options::positionsFileName 
		  << " must be sorted"
		  << std::endl;
	return 2;
    }
    positionsFile.close();

    try {
	if (options::unphased)
	    readGenotypes(haplotypesFile, haplotypes, classes);
	else
	    readHaplotypes(haplotypesFile, haplotypes, classes);
    } catch (ParseError&) {
	std::cerr << "Error reading " 
		  << options::haplotypesFileName << std::endl;
	return 2;
    }
    haplotypesFile.close();

    if (haplotypes.noCols() != (int)positions.size())
	{
	    std::cerr << "The sequence matrix has more columns than there are positions!"
		      << std::endl;
	    return 2;
	}

#if 0
    std::vector<int> affected;
    std::vector<int> unaffected;
    for (int i = 0; i < (int)classes.size(); ++i)
	{
	    if (classes[i] == 0) unaffected.push_back(i);
	    if (classes[i] == 1) affected.push_back(i);
	}
#endif

    if (options::maxlen < 0.0)
	{
	    options::maxlen = positions[positions.size()-1] - positions[0];
	}	    

    
    std::ofstream scoresFile(options::scoresFileName.c_str());
    if (!scoresFile)
	{
	    std::cerr << "Error opening file " 
		      << options::scoresFileName
		      << std::endl;
	    return 2;
	}

     if(options::scoreFunction.empty()){
	 if(classes.size()>400)
	     options::scoreFunction="H";
	 else
	     options::scoreFunction="P";
     }

    if(options::scoreFunction[0]!='B' &&
       options::scoreFunction[0]!='P' &&
       options::scoreFunction[0]!='H' &&
       options::scoreFunction[0]!='A' &&
       options::scoreFunction[0]!='S' &&
       options::scoreFunction[0]!='G' &&
       options::scoreFunction[0]!='N' &&
       options::scoreFunction[0]!='W' &&
       options::scoreFunction[0]!='X')
	{
	    std::cerr << "Not a valid name for a scoreFunction, "
		      << "use B, P, H, G, S, A, or X." << std::endl;
	    return 2;
	} 



   

    Classification classification(classes.begin(), classes.end());
    
    termination_signalled = false;
    sighandler_t prev_handler = signal(SIGINT, signal_handler);
    signal(SIGKILL, signal_handler);


    std::auto_ptr<BasicScore> scoreFun(selectScoringFunction(haplotypes,
							     positions, 
							     classification));

    std::vector<long double> scores(haplotypes.noCols());
    std::fill(scores.begin(), scores.end(), nan(""));

    int from, to;
    if (0 != getInterval(from, to, haplotypes.noCols()))
	{
	    std::cerr << "Specified interval outside the range of the markers."
		      << std::endl;
	    return 2;
	}

    processCurrentSample(scoreFun.get(), scores, from, to);

    if (options::verbose && termination_signalled)
	std::cout << std::endl
		  << "Termination signal recieved; writing results so far and shutting down." << std::endl
		  << "Warning: the results are not reliable since the process was not allowed to" << std::endl
		  << "complete, so analyse them with care!" << std::endl;
    
    printScores(scoresFile, scores);
    if (termination_signalled) return 0;

    
    std::vector<long double> pvals(haplotypes.noCols());
    std::fill(pvals.begin(), pvals.end(), nan(""));
    
    if(options::noPerms) {

	std::ofstream pvalsFile(options::pvalsFileName.c_str());
	if (!pvalsFile)
	    {
		std::cerr << "Error opening file " 
			  << options::pvalsFileName
			  << std::endl;
		return 2;
	    }

	if (options::verbose)
	    {
		std::cout << std::endl
			  << "Calculating permutation test." 
			  << std::endl;
	    }

	PermutationTestScore pScore(*scoreFun.get(), options::noPerms);
	processCurrentSample(&pScore, pvals, from, to);
	//permutationTest(scoreFun.get(), scores, pvals, from, to);

	if (options::verbose && termination_signalled)
	    std::cout << std::endl
		      << "Termination signal recieved; writing results so far and shutting down." << std::endl
		      << "Warning: the results are not reliable since the process was not allowed to" << std::endl
		      << "complete, so analyse them with care!" << std::endl;
	
	for (unsigned int i = 0; i < pvals.size(); ++i)
	    pvals[i] /= options::noPerms;
	printScores(pvalsFile, pvals);
    }

    signal(SIGINT,  prev_handler);
    signal(SIGKILL, prev_handler);

    return 0;
}


static int
quantitative()
{
      
   

    std::ifstream positionsFile(options::positionsFileName.c_str());
    if (!positionsFile)
	{
	    std::cerr << "Error opening " 
		      << options::positionsFileName << std::endl;
	    return 2;
	}
    std::ifstream haplotypesFile(options::haplotypesFileName.c_str());
    if (!haplotypesFile)
	{
	    std::cerr << "Error opening "
		      << options::haplotypesFileName << std::endl;
	    return 2;
	}
    
    std::vector<int> positions;
    Matrix haplotypes(newArrayMatrix(0,0));
    std::vector<double> values;

    try {
	positions = readPositions(positionsFile);
    } catch (UnsortedPositions&) {
	std::cerr << "The positions in " << options::positionsFileName 
		  << " must be sorted"
		  << std::endl;
	return 2;
    }
    positionsFile.close();


    try {
	if (options::unphased)
	    readGenotypesQT(haplotypesFile, haplotypes, values);
	else
	    readHaplotypesQT(haplotypesFile, haplotypes, values);
    } catch (ParseError&) {
	std::cerr << "Error reading " 
		  << options::haplotypesFileName << std::endl;
	return 2;
    }


    haplotypesFile.close();
    if (haplotypes.noCols() != (int)positions.size())
	{
	    std::cerr << "The sequence matrix has more columns than there are positions!"
		      << std::endl;
	    return 2;
	}


    std::vector<int> all_individuals;
    for (int i = 0; i < (int)values.size(); ++i)
	{
	    all_individuals.push_back(i);
	}

    if (options::maxlen < 0.0)
	{
	    options::maxlen = positions[positions.size()-1] - positions[0];
	}	        
    

    std::ofstream scoresFile(options::scoresFileName.c_str());
    if (!scoresFile)
	{
	    std::cerr << "Error opening file " 
		      << options::scoresFileName
		      << std::endl;
	    return 2;
	}
    if(options::scoreFunction.empty()) 
	{
	    options::scoreFunction = "B";
	}
    else{
	if(options::scoreFunction[0]!='B' &&
	   options::scoreFunction[0]!='X' &&
	   options::scoreFunction[0]!='Q')
	    {
		std::cerr << "Not a valid name for a scoreFunction "
			  << "on quantitative data use B, X or Q." 
			  << std::endl;
		return 2;
	    }
    }

    if((options::scoreFunction[0] == 'B' ||
	options::scoreFunction[0] == 'X') &&
       (options::maxMutations<1 || options::maxMutations>3)){
	std::cerr << "max mutations , "
		  << "should be between 1 and 3" << std::endl;
	return 2;
    }

    std::vector<long double> scores(haplotypes.noCols());
    std::fill(scores.begin(), scores.end(), nan(""));
    QTValue qt_value(values.begin(), values.end());
    

    termination_signalled = false;
    sighandler_t prev_handler = signal(SIGINT, signal_handler);
    signal(SIGKILL, signal_handler);

    std::auto_ptr<BasicScore> scoreFun(selectScoringFunction(haplotypes,
							     positions, 
							     qt_value));

    int from, to;
    if (0 != getInterval(from, to, haplotypes.noCols()))
	{
	    std::cerr << "Specified interval outside the range of the markers."
		      << std::endl;
	    return 2;
	}
    processCurrentSample(scoreFun.get(), scores, from, to);
    
    if (options::verbose && termination_signalled)
	std::cout << std::endl
		  << "Termination signal recieved; writing results so far and shutting down." << std::endl
		  << "Warning: the results are not reliable since the process was not allowed to" << std::endl
		  << "complete, so analyse them with care!" << std::endl;
    
    printScores(scoresFile, scores);
    
    // temporary hack testing #mutations:
    

//     uint max_index = 0;
//     for(uint i=1; i< scores.size(); i++){
// 	if (scores[i] > scores[max_index]){
// 	    max_index = i; 
// 	}
//     }
    
//     std::auto_ptr<BasicScore> scoreFun2(new QTnew_nmutScore(haplotypes, 
// 							    positions,
// 							    options::minMarkers, 
// 							    options::maxlen,
// 							    qt_value));
//     std::vector<long double> nmut(haplotypes.noCols());
//     std::fill(nmut.begin(), nmut.end(), nan(""));
//     processCurrentSample(scoreFun2.get(), nmut, max_index, max_index+1);

//     //std::ofstream nmutFile("nmut.txt");
//     /*
//       if (!nmutFile)
// 	{
// 	    std::cerr << "Error opening file " 
// 		      << "nmut.txt"
// 		      << std::endl;
// 	    return 2;
// 	}
//     */
//     scoresFile << nmut[max_index] << " " << scores[max_index] << std::endl;
    
    // End of temporary hack!

    if (termination_signalled) return 0;

    
    std::vector<long double> pvals(haplotypes.noCols());
    std::fill(pvals.begin(), pvals.end(), nan(""));
    
    if(options::noPerms) {

	std::ofstream pvalsFile(options::pvalsFileName.c_str());
	if (!pvalsFile)
	    {
		std::cerr << "Error opening file " 
			  << options::pvalsFileName
			  << std::endl;
		return 2;
	    }

	if (options::verbose)
	    {
		std::cout << std::endl
			  << "Calculating permutation test." 
			  << std::endl;
	    }
    
	PermutationTestScore pScore(*scoreFun.get(), options::noPerms);
	processCurrentSample(&pScore, pvals, from, to);
	//permutationTest(scoreFun.get(), scores, pvals, from, to);
	
	if (options::verbose && termination_signalled)
	    std::cout << std::endl
		      << "Termination signal recieved; writing results so far and shutting down." << std::endl
		      << "Warning: the results are not reliable since the process was not allowed to" << std::endl
		      << "complete, so analyse them with care!" << std::endl;
	
	for (unsigned int i = 0; i < pvals.size(); ++i)
	    pvals[i] /= options::noPerms;
	printScores(pvalsFile, pvals);
    }

    signal(SIGINT,  prev_handler);
    signal(SIGKILL, prev_handler);
    
    return 0;
    
}


int
main(int argc, char *argv[])
{
    using namespace TextManip;
    namespace po = boost::program_options;

    po::options_description visible("Allowed options");
    po::options_description hidden("Allowed options");
    po::options_description all("Allowed options");
    po::variables_map vm;
    visible.add_options()
	("help,h", "Produce help message.")
	("verbose,v","Turns on verbose output.")
	
	("scores,o",
	 po::value<std::string>(&options::scoresFileName)
	                              ->default_value("scores.txt"),
	 "File where scores will be written.")
	("p-values",
	 po::value<std::string>(&options::pvalsFileName)
	                              ->default_value("p-values.txt"),
	 "File where permutation test p-values will be written.")

	("scoreFunction,f",
	 po::value<std::string>(&options::scoreFunction),
	 "Score funcition (H = HQC, B = BIC, A = AIC, G = Gini, S = SMA, "
	 "P = Prob, or X = eXperimental genotype likelihood score) "
	 "(default: H if #haplotypes > 400, otherwise P). For Quantitative "
	 "data: (B = BIC independent chromosomes, X = BIC chromosomes in pairs"
	 ", Q)")
	("interval,i",
	 po::value<std::string>(&options::interval),
	 "Specify sub-interval to score (only markers in the interval will be "
	 "scored/used for permutation tests).  The interval is given in the "
	 "form <from>:<to>.")	 

	("no-permutations,n",
	 po::value<int>(&options::noPerms)->default_value(0),
	 "Number of permutations if permutation tests are to be run.")
	("minmarkers,m",
	 po::value<int>(&options::minMarkers)->default_value(10),
	 "Minimum number of markers used (default: 10).")
	("deslen,x",
	 po::value<double>(&options::maxlen)->default_value(-1),
	 "Maximum length of interval (default: whole range).")

	("quantitative,q", "QTL mapping (default false).")

	("unphased,u", 
	 po::value<bool>(&options::unphased)->default_value(true),
	 "Input is unphased (default true).")

	("max-mutations,z",
	 po::value<int>(&options::maxMutations)->default_value(1),
	 "Maximal number of mutations to consider when using the -fX score.")
	;
    hidden.add_options()
	("positions,p",
	 po::value<std::string>(&options::positionsFileName),
	 "File containing marker positions.")
	("genotypes,g",
	 po::value<std::string>(&options::haplotypesFileName),
	 "File containing marker data.")
	;
    all.add(visible).add(hidden);
    po::positional_options_description p;
    p.add("positions", 1);
    p.add("genotypes", 1);

    try {
	po::store(po::command_line_parser(argc, argv)
		  .options(all).positional(p).run(), 
		  vm, false);
	po::notify(vm);
    } catch (std::exception &e) {
	std::cerr << "Error parsing options:" << e.what()
		  << std::endl
		  << "Usage: " << argv[0] << " [options] positions genotypes"
		  << std::endl
		  << visible << std::endl;
	return 2;
    }

    if (vm.count("help")) {
	std::cout << "Usage: " << argv[0] << " [options] positions genotypes"
		  << std::endl
		  << visible << std::endl;
	return 0;
    }
    options::verbose = vm.count("verbose") > 0;
    options::quantitative = vm.count("quantitative") > 0;

    if (options::verbose)
	std::cout << bold
		  << "Running " << PACKAGE_NAME << " v" << PACKAGE_VERSION
		  << reset
		  << std::endl << std::endl;

    if (options::positionsFileName.empty())
	{
  	    std::cerr << bold << red << "Error: " << reset
		      << bold << "No position file specified." << reset
		      << std::endl << std::endl;
	    return 2;
	}
    if (options::haplotypesFileName.empty())
	{
  	    std::cerr << bold << red << "Error: " << reset
		      << bold << "No haplotype file specified." << reset
		      << std::endl << std::endl;
	    return 2;
	}

    struct timeval tv; struct timezone tz;
    gettimeofday(&tv,&tz);
    unsigned int random_seed = tv.tv_usec;
    std::srand(random_seed);

    // return NaNs rather than crash.
    gsl_set_error_handler_off();

    if(options::quantitative){
	return quantitative();
    } else {
	return caseControl();
    }
    return 0;
}
