/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "config.hh"
#include "io.hh"
#include "matrix_utils.hh"
#include "incompatibility.hh"
#include "pph.hh"
#include "tree.hh"
#include "tree_builder.hh"

#include <snpfile/array_matrix.hh>
using namespace BiRC::SNPFile;

#include <boost/program_options.hpp>
#include <fstream>
#include <string>

namespace options {
    std::string positionsFileName;
    std::string haplotypesFileName;
    int minMarkers = 10;
    double maxlen = -1.0;
    bool unphased = true;
}

int
main(int argc, char *argv[])
{
    namespace po = boost::program_options;
    po::options_description visible("Allowed options");
    po::options_description hidden("Allowed options");
    po::options_description all("Allowed options");
    po::variables_map vm;
    visible.add_options()
	("help,h", "Produce help message")
	
	("minmarkers,m",
	 po::value<int>(&options::minMarkers)->default_value(10),
	 "Minimum number of markers used (default: 10).")
	("deslen,x",
	 po::value<double>(&options::maxlen)->default_value(-1),
	 "Maximum length of interval (default: whole range).")

	("unphased,u", 
	 po::value<bool>(&options::unphased)->default_value(true),
	 "Input is unphased (default true)")
	;
    hidden.add_options()
	("positions,p",
	 po::value<std::string>(&options::positionsFileName),
	 "File containing marker positions.")
	("genotypes,g",
	 po::value<std::string>(&options::haplotypesFileName),
	 "File containing marker data.")
	;
    all.add(visible).add(hidden);
    po::positional_options_description p;
    p.add("positions", 1);
    p.add("genotypes", 1);

    try {
	po::store(po::command_line_parser(argc, argv)
		  .options(all).positional(p).run(), 
		  vm, false);
	po::notify(vm);
    } catch (std::exception &e) {
	std::cerr << "Error parsing options:" << e.what()
		  << std::endl
		  << "Usage: " << argv[0] << " [options] positions genotypes"
		  << std::endl
		  << visible << std::endl;
	return 2;
    }

    if (vm.count("help")) {
	std::cout << "Usage: " << argv[0] << " [options] positions genotypes"
		  << std::endl
		  << visible << std::endl;
	return 0;
    }

    if (options::positionsFileName == "")
	{
  	    std::cerr << "No position file specified" << std::endl;
	    return 2;
	}
    if (options::haplotypesFileName == "")
	{
  	    std::cerr << "No haplotypes file specified" << std::endl;
	    return 2;
	}


    std::ifstream positionsFile(options::positionsFileName.c_str());
    if (!positionsFile)
	{
	    std::cerr << "Error opening " 
		      << options::positionsFileName << std::endl;
	    return 2;
	}
    std::ifstream haplotypesFile(options::haplotypesFileName.c_str());
    if (!haplotypesFile)
	{
	    std::cerr << "Error opening "
		      << options::haplotypesFileName << std::endl;
	    return 2;
	}


    std::vector<int> positions;
    Matrix haplotypes(newArrayMatrix(0,0));
    std::vector<int> classes;

    try {
	positions = readPositions(positionsFile);
    } catch (UnsortedPositions&) {
	std::cerr << "The positions in " << options::positionsFileName 
		  << " must be sorted"
		  << std::endl;
	return 2;
    }
    positionsFile.close();

    // FIXME: Does this work if the data is QT?  There might be
    // problems with the first col...
    try {
	if (options::unphased)
	    readGenotypes(haplotypesFile, haplotypes, classes);
	else
	    readHaplotypes(haplotypesFile, haplotypes, classes);
    } catch (ParseError&) {
	std::cerr << "Error reading " 
		  << options::haplotypesFileName << std::endl;
	return 2;
    }
    haplotypesFile.close();
    
    std::auto_ptr<CompatibleCache> cc(options::unphased
				      ? new pph::CompatibleCache(haplotypes)
				      : new CompatibleCache(haplotypes));

    if (options::unphased)
	for (int i = 0; i < (int)positions.size(); ++i)
	    {
		int lb, rb;
		calculateIncompatibilities(haplotypes, positions, 
					   *cc, i, lb, rb,
					   options::minMarkers, 
					   options::maxlen);
		int noSTClasses = pph::getNumOfClasses(haplotypes.view(),
						       lb,rb-1);

		if (noSTClasses > 1)
		    std::cout << "Local trees for position ";
		else
		    std::cout << "local tree for position ";
		std::cout << i << " (region ["
			  << lb << ',' << (rb-1) << "]): "
			  << std::endl;


	    std::cout << std::endl << noSTClasses << std::endl << std::endl;

	    if(noSTClasses > 0)  // it is a compatible region
		{
		    if (noSTClasses > 10) noSTClasses = 10;
		    
		    int noPPHSolutions = 1 << (noSTClasses - 1);
		    Matrix pphSolution(newArrayMatrix(0,0,true));
		
		    for (int solIdx = 0; solIdx < noPPHSolutions; ++solIdx) 
			{
			    pph::getPPHSolution(solIdx, 
						noPPHSolutions, 
						pphSolution);
			    MatrixView view = pphSolution.view();
			    std::auto_ptr<Tree> t(::buildTree(view,0,
							      i,positions));
			    std::cout << *t << std::endl;
			}
		}
	    else 
		{  // incompatible region
		    Matrix ccSolution(newArrayMatrix(0,0,true));
		    // _noSTClasses is used as _errorCode here
		    pph::getCCSolution(noSTClasses, ccSolution);
		    MatrixView view = ccSolution.view();
		    std::auto_ptr<Tree> t(::buildTree(view,0,i,positions));
		    std::cout << *t << std::endl;
		}
	    std::cout << std::endl;		
	    }
    else
	for (int i = 0; i < (int)positions.size(); ++i)
	    {
		int lb, rb;
		calculateIncompatibilities(haplotypes, positions, 
					   *cc, i, lb, rb,
					   options::minMarkers, 
					   options::maxlen);
		MatrixView view = haplotypes.subMatrixColumns(lb,rb);
		sortColumns(view, positions, i);
		std::auto_ptr<Tree> t(::buildTree(view,0,i,positions));
		std::cout << "Local tree for position "
			  << i << " (region ["
			  << lb << ',' << (rb-1) << "]): "
			  << std::endl
			  << *t
			  << std::endl << std::endl;
	}

    return 0;
}
