/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "config.hh"
#include <snpfile/array_matrix.hh>
using namespace BiRC::SNPFile;

#include "case_control_score.hh"
#include "tree.hh"

#include <gsl/gsl_cdf.h>
#include <cmath>

// --- Local, generally used, functions --------------------------------
static long double Pleaf(Tree &node, const Classification &)
{
    long double res = 0.0;
    if (node.getAffected()>0) {
	res += node.getAffected() *
	    logl((long double) node.getAffected()/node.getIndividuals());
    }
    if (node.getUnaffected()>0) {
	res += node.getUnaffected() *
	    logl((long double) node.getUnaffected()/node.getIndividuals());
    }
    return res;
}


// --- Classification --------------------------------------------------
void 
Classification::permuteClasses()
{
    std::random_shuffle(i_classes.begin(), i_classes.end());

}

// --- CaseControlScore ------------------------------------------------
void
CaseControlScore::permute()
{ 
    if (!i_backup_classification.get())
	i_backup_classification.reset(new Classification(i_classification));
    i_classification.permuteClasses(); 
}

void
CaseControlScore::resetPermutation()
{ 
    if (i_backup_classification.get())
	i_classification = *i_backup_classification;
}


// --- SMAScore --------------------------------------------------------
SMAScore::SMAScore(ImmutableMatrix &haplotypes, std::vector<int> &positions,
		   int minMarkers, double maxlen,
		   Classification &classification)
    : CaseControlScore(haplotypes, positions, minMarkers,
		       maxlen, classification)
{
    i_scoresThroughTree = false;
}

long double
SMAScore::scoreLocus(int col) const
{
    int a0 = 0, a1 = 0, u0 = 0, u1 = 0;
    for (int i = 0; i < i_haplotypes.noRows(); ++i)
	if (i_classification.isAffected(i))
	    // Affected individual...
	    if (i_haplotypes(i,col) == 0)      ++a0;
	    else if (i_haplotypes(i,col) == 1) ++a1;
	    else ;//assert(false);  FIXME: deal with missing data and such
	else
	    // Unaffected individual
	    if (i_haplotypes(i,col) == 0)      ++u0;
	    else if (i_haplotypes(i,col) == 1) ++u1;
	    else ;//assert(false); FIXME: deal with missing data and such


    // In theory I know these globally, but when we handle missing
    // data it might vary across the loci.
    int a = a0 + a1;
    int u = u0 + u1;

    int n0 = a0 + u0;
    int n1 = a1 + u1;
    int n = n0 + n1;

    double exp_a0 = ((double)a*n0)/n;
    double exp_a1 = ((double)a*n1)/n;
    double exp_u0 = ((double)u*n0)/n;
    double exp_u1 = ((double)u*n1)/n;

    double xa0 = a0 - exp_a0;
    double xa1 = a1 - exp_a1;
    double xu0 = u0 - exp_u0;
    double xu1 = u1 - exp_u1;

    double x = xa0*xa0/exp_a0 + xa1*xa1/exp_a1
	     + xu0*xu0/exp_u0 + xu1*xu1/exp_u1;

    double pval = gsl_cdf_chisq_Q(x,1);

    return pval;
}

long double
SMAScore::scoreTree(Tree *) const
{
    return 0.0;			// dummy to make this class concrete...
}




// --- ProbScore -------------------------------------------------------
ProbScore::ProbScore(ImmutableMatrix &haplotypes, std::vector<int> &positions,
		     int minMarkers, double maxlen, 
		     Classification &classification)
    : CaseControlScore(haplotypes, positions, minMarkers, maxlen,
		       classification)
{
    tableInit(i_classification.size()+2);
}

void
ProbScore::tableInit(size_t size) const
{
    i_bin = new long double*[size];
    for(size_t i=0; i<size; i++) i_bin[i]= new long double[size];
    i_bin[0][0]=1;
    i_bin[1][0]=1;
    i_bin[1][1]=1;
    for(size_t i= 2; i< size; i++){
		i_bin[i][0]=1;
		i_bin[i][i]=1;
		for(size_t j= 1; j<i; j++){
			i_bin[i][j] = i_bin[i-1][j] + i_bin[i-1][j-1];
		}
    }
}

long double
ProbScore::logProbOne(size_t na, size_t nu) const
{
    size_t n = na + nu;
    long double res = 0.0;
    for(size_t i=na; i<=n; i++) {
	res += (i_bin[n][i] 
		* powl(i_classification.affectedFraction(), i) 
		* powl(i_classification.unaffectedFraction(), n-i));
    }
    return logl(res);
}

namespace {
    struct ProbabilityScoreVisitor : TreeVisitor {
	const ProbScore *probScore;
	typedef long double (ProbScore::*logProbOne_t)(size_t na, size_t nu) const;
	logProbOne_t logProbOne;

	const Classification &classification;
	const long double penalty;

	ProbabilityScoreVisitor(const ProbScore *probScore,
				logProbOne_t logProbOne,
				const Classification &classification)
	    : probScore(probScore), logProbOne(logProbOne),
	      classification(classification), penalty(0.5) 
	{
	}
	virtual void visitInnerNode(InnerNode &node);
	virtual void visitLeafNode(LeafNode &node);
    };

    void ProbabilityScoreVisitor::visitInnerNode(InnerNode &node)
    {
	node.setAffected(node.getLeftChild()->getAffected() 
			 + node.getRightChild()->getAffected());
	node.setUnaffected(node.getLeftChild()->getUnaffected()
			   + node.getRightChild()->getUnaffected());
	
	long double leaf;
	leaf = (probScore->*logProbOne)(node.getAffected(), node.getUnaffected()) 
	    + penalty * node.getLevel();

	long double left = node.getLeftChild()->getScore(); 
	long double right = node.getRightChild()->getScore(); 

	if(left < right){
	    if(leaf < left && node.getLevel()!=0) {
		node.setScore(leaf);
	    }
	    else {
		node.setScore(left);
	    }
	}
	else{
	    if(leaf < right && node.getLevel()!=0) {
		node.setScore(leaf);
	    }
	    else {
		node.setScore(right);
	    }
	}
    }
    
    void ProbabilityScoreVisitor::visitLeafNode(LeafNode &node)
    {
	int noAffected = 0;
	int noUnaffected = 0;
	LeafNode::IDsIterator i;
	for (i = node.leafIDsBegin(); i != node.leafIDsEnd(); ++i){
	    if (classification.isAffected(*i)) ++noAffected;
	    else if (classification.isUnaffected(*i)) ++noUnaffected;
	}
	node.setAffected(noAffected);
	node.setUnaffected(noUnaffected);
	node.setScore((probScore->*logProbOne)(node.getAffected(), 
					       node.getUnaffected()) 
		      + penalty * node.getLevel());
    }
}

long double
ProbScore::scoreTree(Tree *tree) const
{
    ProbabilityScoreVisitor sv(this, &ProbScore::logProbOne, i_classification);
    tree->acceptVisitor(sv);
    return -(tree->getScore());
}

// --- BIC Score -------------------------------------------------------
namespace {
    struct BICScoreVisitor : TreeVisitor {
	const Classification &classification;
	const size_t nIndividuals;
	const double m;
	BICScoreVisitor(const Classification &classification,
			const size_t nIndividuals)
	    : classification(classification), 
	      nIndividuals(nIndividuals), m(1.0) 
	{}
	virtual void visitInnerNode(InnerNode &node);
	virtual void visitLeafNode(LeafNode &node);
    };


    void BICScoreVisitor::visitInnerNode(InnerNode &node)
    {
	node.setAffected(node.getLeftChild()->getAffected() 
			 + node.getRightChild()->getAffected());
	node.setUnaffected(node.getLeftChild()->getUnaffected()
			   + node.getRightChild()->getUnaffected());
	node.setClusters(node.getLeftChild()->getClusters()
			 + node.getRightChild()->getClusters());
	
	long double leaf = Pleaf(node, classification);
	long double internal = 
	    node.getLeftChild()->getScore() +
	    node.getRightChild()->getScore();
	long double BIC_leaf = 
	    - 2.0 *leaf
	    + logl((long double)nIndividuals);
	long double BIC_internal = 
	    - 2.0 * internal
	    + node.getClusters()*logl((long double)nIndividuals);
	if(BIC_leaf<BIC_internal && node.getLevel()!=0){
	    node.setScore(leaf);
	    node.setClusters(1);
	}
	else{
	    node.setScore(internal);
	}
    }
    

    void BICScoreVisitor::visitLeafNode(LeafNode &node)
    {
	
	int noAffected = 0;
	int noUnaffected = 0;
	LeafNode::IDsIterator i;
	for (i = node.leafIDsBegin(); i != node.leafIDsEnd(); ++i)
	    if (classification.isAffected(*i)) ++noAffected;
	    else if (classification.isUnaffected(*i)) ++noUnaffected;
	node.setAffected(noAffected);
	node.setUnaffected(noUnaffected);
	node.setClusters(1);
	node.setScore(Pleaf(node, classification));
    }
}


long double
BICScore::scoreTree(Tree *t) const
{
    size_t nIndividuals = i_classification.size();
    BICScoreVisitor sv(i_classification, nIndividuals);
    t->acceptVisitor(sv);
    long double pa = ((long double)t->getAffected())/ nIndividuals;
    long double pu = ((long double)t->getUnaffected())/ nIndividuals;
    return 
	- (2.0 * (t->getAffected()*logl(pa) + t->getUnaffected()*logl(pu) )
	   - 2.0 * t->getScore() 
	   + (t->getClusters()-1) * logl((long double)nIndividuals));
}


// --- AIC Score -------------------------------------------------------
namespace {
    struct AICScoreVisitor : TreeVisitor {
	const Classification &classification;
	const size_t nIndividuals;
	const double m;
	AICScoreVisitor(const Classification &classification,
					const size_t nIndividuals)
	    : classification(classification), nIndividuals(nIndividuals), m(1.0) 
	{}
	virtual void visitInnerNode(InnerNode &node);
	virtual void visitLeafNode(LeafNode &node);
    };


    void AICScoreVisitor::visitInnerNode(InnerNode &node)
    {
	node.setAffected(node.getLeftChild()->getAffected() 
			 + node.getRightChild()->getAffected());
	node.setUnaffected(node.getLeftChild()->getUnaffected()
			   + node.getRightChild()->getUnaffected());
	node.setClusters(node.getLeftChild()->getClusters()
			 + node.getRightChild()->getClusters());
	
	long double leaf = Pleaf(node, classification);
	long double internal = 
	    node.getLeftChild()->getScore() +
	    node.getRightChild()->getScore();
	long double AIC_leaf = 
	    - 2.0*leaf 
	    + 2.0;
	long double AIC_internal = 
	    - 2.0*internal
	    + 2.0*node.getClusters();
	if(AIC_leaf<AIC_internal && node.getLevel()!=0){
	    node.setScore(leaf);
	    node.setClusters(1);
	}
	else{
	    node.setScore(internal);
	}
    }
    

    void AICScoreVisitor::visitLeafNode(LeafNode &node)
    {
	
	int noAffected = 0;
	int noUnaffected = 0;
	LeafNode::IDsIterator i;
	for (i = node.leafIDsBegin(); i != node.leafIDsEnd(); ++i)
	    if (classification.isAffected(*i)) ++noAffected;
	    else if (classification.isUnaffected(*i)) ++noUnaffected;
	node.setAffected(noAffected);
	node.setUnaffected(noUnaffected);
	node.setClusters(1);
	node.setScore(Pleaf(node, classification));
    }
}

long double
AICScore::scoreTree(Tree *t) const
{
    size_t nIndividuals = i_classification.size();
    AICScoreVisitor sv(i_classification, nIndividuals);
    t->acceptVisitor(sv);
    long double pa= ((long double)t->getAffected())/nIndividuals;
    long double pu= ((long double)t->getUnaffected())/nIndividuals;
    return - (2.0*(t->getAffected()*logl(pa) + t->getUnaffected()*logl(pu))
	      - 2.0*t->getScore() 
	      + 2.0*(t->getClusters()-1));
}

// --- HQ Score --------------------------------------------------------
namespace {
    struct HQScoreVisitor : TreeVisitor {
	const Classification &classification;
	const size_t nIndividuals;
	const double m;
	HQScoreVisitor(const Classification &classification,
				   const size_t nIndividuals)
	    : classification(classification), nIndividuals(nIndividuals), m(1.0) 
	{}
	virtual void visitInnerNode(InnerNode &node);
	virtual void visitLeafNode(LeafNode &node);
    };


    void HQScoreVisitor::visitInnerNode(InnerNode &node)
    {
	node.setAffected(node.getLeftChild()->getAffected() 
			 + node.getRightChild()->getAffected());
	node.setUnaffected(node.getLeftChild()->getUnaffected()
			   + node.getRightChild()->getUnaffected());
	node.setClusters(node.getLeftChild()->getClusters()
			 + node.getRightChild()->getClusters());
	
	long double leaf = Pleaf(node, classification);
	long double internal = 
	    node.getLeftChild()->getScore() +
	    node.getRightChild()->getScore();
	long double HQ_leaf = 
	    - 2.0 * leaf
	    + 2.0 * logl(logl((double)nIndividuals));
	long double HQ_internal = 
	    - 2.0 * internal
	    + 2.0 * node.getClusters() * logl(logl((double)nIndividuals));
	if(HQ_leaf<HQ_internal && node.getLevel()!=0){
	    node.setScore(leaf);
	    node.setClusters(1);
	}
	else{
	    node.setScore(internal);
	}
    }
    

    void HQScoreVisitor::visitLeafNode(LeafNode &node)
    {
	
	int noAffected = 0;
	int noUnaffected = 0;
	LeafNode::IDsIterator i;
	for (i = node.leafIDsBegin(); i != node.leafIDsEnd(); ++i)
	    if (classification.isAffected(*i)) ++noAffected;
	    else if (classification.isUnaffected(*i)) ++noUnaffected;
	node.setAffected(noAffected);
	node.setUnaffected(noUnaffected);
	node.setClusters(1);
	node.setScore(Pleaf(node, classification));
    }
}

long double
HQScore::scoreTree(Tree *t) const
{
    size_t nIndividuals = i_classification.size();
    HQScoreVisitor sv(i_classification, nIndividuals);
    t->acceptVisitor(sv);
    long double pa= ((long double)t->getAffected())/nIndividuals;
    long double pu= ((long double)t->getUnaffected())/nIndividuals;
    return 
	- (2.0*(t->getAffected()*logl(pa) + t->getUnaffected()*logl(pu)) 
	   - 2.0*t->getScore() 
	   + 2.0*(t->getClusters()-1) * logl(logl((long double)nIndividuals)));
}

// --- Gini Score ------------------------------------------------------
namespace {
    struct GiniScoreVisitor : TreeVisitor {
	const Classification &classification;
	GiniScoreVisitor(const Classification &classification)
	    : classification(classification)
	{}
	virtual void visitInnerNode(InnerNode &node);
	virtual void visitLeafNode(LeafNode &node);
    };

    void GiniScoreVisitor::visitInnerNode(InnerNode &node)
    {
	node.setAffected(node.getLeftChild()->getAffected()
			 + node.getRightChild()->getAffected());
	node.setUnaffected(node.getLeftChild()->getUnaffected()
			 + node.getRightChild()->getUnaffected());
	size_t n_left = node.getLeftChild()->getIndividuals();
	size_t n_right = node.getRightChild()->getIndividuals();
	size_t n = node.getIndividuals();
	node.setScore(((double)n_left)/n * node.getLeftChild()->getScore() +
		      ((double)n_right)/n * node.getRightChild()->getScore());
   }

    void GiniScoreVisitor::visitLeafNode(LeafNode &node)
    {
	int noAffected = 0;
	int noUnaffected = 0;
	LeafNode::IDsIterator i;
	for (i = node.leafIDsBegin(); i != node.leafIDsEnd(); ++i)
	    if (classification.isAffected(*i)) ++noAffected;
	    else if (classification.isUnaffected(*i)) ++noUnaffected;

	node.setAffected(noAffected);
	node.setUnaffected(noUnaffected);

	int n = noAffected + noUnaffected;
	double affectedFraction = static_cast<double>(noAffected) / n;
	node.setScore(1.0-(affectedFraction*affectedFraction+
			   (1.0 - affectedFraction)*(1.0 -affectedFraction)));
    }
}

long double
GiniScore::scoreTree(Tree *t) const
{
    GiniScoreVisitor sv(i_classification);
    t->acceptVisitor(sv);
    return - t->getScore();
}


namespace {
    struct CountVisitor : TreeVisitor {
	const Classification &classification;
	uint dfs;
	CountVisitor(const Classification &classification)
	    : classification(classification), dfs(0)
	{}
	virtual void visitInnerNode(InnerNode &node);
	virtual void visitLeafNode(LeafNode &node);
    };
    
    
    void CountVisitor::visitInnerNode(InnerNode &node)
    {
	node.setAffected(node.getLeftChild()->getAffected()
			 + node.getRightChild()->getAffected());
	node.setUnaffected(node.getLeftChild()->getUnaffected()
			   + node.getRightChild()->getUnaffected());
	node.setClusters(dfs);
	dfs++;
    }
    

    void CountVisitor::visitLeafNode(LeafNode &node)
    {
	node.setClusters(dfs);
	dfs++;
	int noAffected = 0;
	int noUnaffected = 0;
	LeafNode::IDsIterator i;
	for (i = node.leafIDsBegin(); i != node.leafIDsEnd(); ++i)
	    if (classification.isAffected(*i)) ++noAffected;
	    else if (classification.isUnaffected(*i)) ++noUnaffected;
	node.setAffected(noAffected);
	node.setUnaffected(noUnaffected);
    }
}


void get_clusters(Tree *node, std::vector<int> &clusters_aff, 
		  std::vector<int> &clusters_unaff, 
		  std::vector<bool> &is_mutation, uint cur_cluster){

    InnerNode* IN;
    LeafNode* LN;
    
    if(is_mutation[node->getClusters()]){
	clusters_aff.push_back(0);
	clusters_unaff.push_back(0);
	cur_cluster = clusters_aff.size()-1;
    }
    IN = dynamic_cast<InnerNode*>(node);
    if(IN!=0){
	get_clusters(IN->getLeftChild(), clusters_aff, clusters_unaff, is_mutation, cur_cluster);
	get_clusters(IN->getRightChild(), clusters_aff, clusters_unaff, is_mutation, cur_cluster);
    }
    else{
	LN = dynamic_cast<LeafNode*>(node);
	clusters_aff[cur_cluster] += node->getAffected();
	clusters_unaff[cur_cluster] += node->getUnaffected();
    }
}    

static long double nll(std::vector<int> nas, std::vector<int> nus)
{
    long double res = 0.0;
    for(size_t i = 0; i < nas.size(); i++){
	if (nas[i] > 0) {
	    res += nas[i] * logl((long double) nas[i] / (nas[i] + nus[i]));
	}
	if (nus[i] > 0) {
	    res += nus[i] * logl((long double) nus[i] / (nas[i] + nus[i]));
	}
    }
    return res;
}

long double
newBICScore::scoreTree(Tree *t) const
{
    size_t nIndividuals = i_classification.size();
    CountVisitor cv(i_classification);
    t->acceptVisitor(cv);
    double maxscore = - 100000.0;
    double score = 0.0;
    int tsize = t->getClusters();
    std::vector<bool> is_mutation(tsize+1, false);
    //get_clusters(t, clusters, is_mutation, 0); 
    int n_mut =0;
    //long double sigma = get_sigma(clusters);
    long double pa= ((long double)t->getAffected())/nIndividuals;
    long double pu= ((long double)t->getUnaffected())/nIndividuals;
    long double leaf = (t->getAffected()*logl(pa) + t->getUnaffected()*logl(pu));
    for(int i = 1; i<tsize; i++){ /// Start == 0 ???
	is_mutation[i]=true;
	std::vector<int> clusters_aff(1, 0);
	std::vector<int> clusters_unaff(1, 0);
	get_clusters(t, clusters_aff, clusters_unaff, is_mutation, 0); 

	score = - 2.0 *leaf + 2.0 * nll(clusters_aff, clusters_unaff) -
	    (clusters_aff.size()-1) * logl((long double)nIndividuals);
	
	if(score>maxscore){
	    maxscore = score;
	    n_mut = 1;
	}
	if(muts_to_check>1){
	    for(int j = i+1; j<tsize; j++){
		is_mutation[j]=true;
		std::vector<int> clusters_aff(1, 0);
		std::vector<int> clusters_unaff(1, 0);
		get_clusters(t, clusters_aff, clusters_unaff, is_mutation, 0); 
		score = - 2.0 *leaf + 2.0 * nll(clusters_aff, clusters_unaff) -
		    (clusters_aff.size()-1) * logl((long double)nIndividuals);
		if(score>maxscore){
		    maxscore = score;
		    n_mut = 2;
		}
		
		if(muts_to_check>2){
		    for(int k = j+1; k<tsize; k++){
			is_mutation[k]=true;
			std::vector<int> clusters_aff(1, 0);
			std::vector<int> clusters_unaff(1, 0);
			get_clusters(t, clusters_aff, clusters_unaff, 
				     is_mutation, 0); 
			score = - 2.0 *leaf + 2.0 * 
			    nll(clusters_aff, clusters_unaff) -
			    (clusters_aff.size()-1) * 
			    logl((long double)nIndividuals);
			if(score>maxscore){
			    maxscore = score;
			    n_mut = 3;
		}
			is_mutation[k]=false;
		    }
		}
		is_mutation[j]=false;  
	    }
	}	
	is_mutation[i]=false;
    }
    return maxscore;
}


long double
newHQScore::scoreTree(Tree *t) const
{
    size_t nIndividuals = i_classification.size();
    CountVisitor cv(i_classification);
    t->acceptVisitor(cv);
    double maxscore = - 100000.0;
    double score = 0.0;
    int tsize = t->getClusters();
    std::vector<bool> is_mutation(tsize+1, false);
    long double pa= ((long double)t->getAffected())/nIndividuals;
    long double pu= ((long double)t->getUnaffected())/nIndividuals;
    long double leaf = (t->getAffected()*logl(pa) + t->getUnaffected()*logl(pu));
    for(int i = 1; i<tsize; i++){ /// Start == 0 ???
	is_mutation[i]=true;
	std::vector<int> clusters_aff(1, 0);
	std::vector<int> clusters_unaff(1, 0);
	get_clusters(t, clusters_aff, clusters_unaff, is_mutation, 0); 

	score = - 2.0 *leaf + 2.0 * nll(clusters_aff, clusters_unaff) -
	    2.0*(clusters_aff.size()-1) *logl(logl((long double)nIndividuals));
	
	if(score>maxscore){
	    maxscore = score;
	}
	if(muts_to_check>1){
	    for(int j = i+1; j<tsize; j++){
		is_mutation[j]=true;
		std::vector<int> clusters_aff(1, 0);
		std::vector<int> clusters_unaff(1, 0);
		get_clusters(t, clusters_aff, clusters_unaff, is_mutation, 0); 
		score = - 2.0 *leaf + 2.0 * nll(clusters_aff, clusters_unaff) -
		    2.0*(clusters_aff.size()-1) * 
		    logl(logl((long double)nIndividuals));
		if(score>maxscore){
		    maxscore = score;
		}
		
		if(muts_to_check>2){
		    for(int k = j+1; k<tsize; k++){
			is_mutation[k]=true;
			std::vector<int> clusters_aff(1, 0);
			std::vector<int> clusters_unaff(1, 0);
			get_clusters(t, clusters_aff, clusters_unaff, is_mutation, 0); 
			score = - 2.0 *leaf + 2.0 * 
			    nll(clusters_aff, clusters_unaff) -
			    2.0*(clusters_aff.size()-1) * 
			    logl(logl((long double)nIndividuals));
			if(score>maxscore){
			    maxscore = score;
			}
			is_mutation[k]=false;
		    }
		}
		is_mutation[j]=false;  
	    }
	}	
	is_mutation[i]=false;
    }
    return maxscore;
}



// --- Depricated functions... -----------------------------------------
long double
sma(const MatrixView &m, int col, const Classification &cls)
{
    int a0 = 0, a1 = 0, u0 = 0, u1 = 0;
    for (int i = 0; i < m.noRows(); ++i)
		if (cls.isAffected(i))
			// Affected individual...
			if (m(i,col) == 0)      ++a0;
			else if (m(i,col) == 1) ++a1;
			else ;//assert(false); FIXME: deal with missing data and such
		else
			// Unaffected individual
			if (m(i,col) == 0)      ++u0;
			else if (m(i,col) == 1) ++u1;
			else ;//assert(false); FIXME: deal with missing data and such

    // In theory I know these globally, but when we handle missing
    // data it might vary across the loci.
    int a = a0 + a1;
    int u = u0 + u1;

    int n0 = a0 + u0;
    int n1 = a1 + u1;
    int n = n0 + n1;

    double exp_a0 = ((double)a*n0)/n;
    double exp_a1 = ((double)a*n1)/n;
    double exp_u0 = ((double)u*n0)/n;
    double exp_u1 = ((double)u*n1)/n;

    double xa0 = a0 - exp_a0;
    double xa1 = a1 - exp_a1;
    double xu0 = u0 - exp_u0;
    double xu1 = u1 - exp_u1;

    double x = xa0*xa0/exp_a0 + xa1*xa1/exp_a1
	     + xu0*xu0/exp_u0 + xu1*xu1/exp_u1;

    double pval = gsl_cdf_chisq_Q(x,1);

    return pval;
}

long double
sma(const Matrix &m, int col, const Classification &cls)
{
    return sma(m.view(), col, cls);
}



static size_t lastBinSize = 0;
static long double **bin = 0;

static void 
table_init(const size_t size){
    if (size == lastBinSize) return;
    if (bin) {
	for (size_t i = 0; i < lastBinSize; ++i)
	    delete[] bin[i];
	delete[] bin;
    }
    bin = new long double*[size];
    for(size_t i=0; i<size; i++)
	bin[i]= new long double[size];
    bin[0][0]=1;
    bin[1][0]=1;
    bin[1][1]=1;
    for(size_t i= 2; i< size; i++){
	bin[i][0]=1;
	bin[i][i]=1;
	for(size_t j= 1; j<i; j++){
	    bin[i][j] = bin[i-1][j] + bin[i-1][j-1];
	}
    }
}


static long double logProbOne(size_t na, size_t nu, 
			      const Classification &classification){ 
    size_t n = na + nu;
    long double res = 0.0;
    for(size_t i=na; i<=n; i++){
		res += (bin[n][i] 
			* powl(classification.affectedFraction(), i) 
			* powl(classification.unaffectedFraction(), n-i));
    }
    return logl(res);
}

static long double logProbOneHG(size_t na, size_t nu, size_t fa, size_t fu){
    //if(na==0)
    //return 0.0;
    long double res =0.0;
    na = na+1;
    nu = nu+1;
    fa = fa+1;
    fu = fu+1;
    size_t f = fa + fu;
    size_t n = na + nu;
    //    for(int i = na; i <=  std::min(n, fa); i++){
    //    	res += (Bin[fa][i] * Bin[fu][n-i]);
    //    }
    //std::cout << fa << " " << fu << std::endl ;std::max(na, nu)
    for(size_t i=na; i<=n; i++){
		res += (bin[fa][i] * bin[fu][n-i]);
    }
    //    std::cout << res << std::endl;
    return logl(res) - logl(bin[f][n]);
}


namespace {
    struct OldProbabilityScoreVisitor : TreeVisitor {
	const Classification &classification;
	const int nSamples;
	const bool bin;
	const long double penalty;
	OldProbabilityScoreVisitor(const Classification &classification, 
				   const int nSamples, const bool bin)
	    : classification(classification), nSamples(nSamples), bin(bin), 
	penalty(0.5) {}
	virtual void visitInnerNode(InnerNode &node);
	virtual void visitLeafNode(LeafNode &node);
    };

    void OldProbabilityScoreVisitor::visitInnerNode(InnerNode &node)
    {
	node.setAffected(node.getLeftChild()->getAffected() 
			 + node.getRightChild()->getAffected());
	node.setUnaffected(node.getLeftChild()->getUnaffected()
			   + node.getRightChild()->getUnaffected());
	
	long double leaf;
	if(bin){
	    leaf = logProbOne(node.getAffected(), node.getUnaffected(), 
			      classification) 
		+ penalty * node.getLevel();
	}
	else
	    leaf = logProbOneHG(node.getAffected(), node.getUnaffected(),
				nSamples, nSamples)
		+ penalty * node.getLevel(); 
	
	long double left = node.getLeftChild()->getScore(); 
	long double right = node.getRightChild()->getScore(); 

	if(left < right){
	    if(leaf < left && node.getLevel()!=0) {
		node.setScore(leaf);
	    }
	    else {
		node.setScore(left);
	    }
	}
	else{
	    if(leaf < right && node.getLevel()!=0) {
		node.setScore(leaf);
	    }
	    else {
		node.setScore(right);
	    }
	}
    }
    
    void OldProbabilityScoreVisitor::visitLeafNode(LeafNode &node)
    {
	int noAffected = 0;
	int noUnaffected = 0;
	LeafNode::IDsIterator i;
	for (i = node.leafIDsBegin(); i != node.leafIDsEnd(); ++i){
	    if (classification.isAffected(*i)) ++noAffected;
	    else if (classification.isUnaffected(*i)) ++noUnaffected;
	}
	node.setAffected(noAffected);
	node.setUnaffected(noUnaffected);
	if(bin){
	    node.setScore(logProbOne(node.getAffected(), node.getUnaffected(),
				     classification) 
			  + penalty * node.getLevel());
	}
	else
	    node.setScore(logProbOneHG(noAffected, noUnaffected,
				       nSamples, nSamples)
			  + penalty * node.getLevel()); 
    }
}

long double
probabilityScore(Tree *t, const Classification &classification, 
		 const int nIndividuals, const bool bin)
{
    table_init(classification.size()+2);
    OldProbabilityScoreVisitor sv(classification, nIndividuals/2, bin);
    t->acceptVisitor(sv);
    return -(t->getScore());
}







long double
BICscore(Tree *t, const Classification &classification, const int nIndividuals)
{
    BICScoreVisitor sv(classification, nIndividuals);
    t->acceptVisitor(sv);
    long double pa= ((long double)t->getAffected())/ nIndividuals;
    long double pu= ((long double)t->getUnaffected())/ nIndividuals;
    return 
	- (2.0 * (t->getAffected()*logl(pa) + t->getUnaffected()*logl(pu) )
	   - 2.0 * t->getScore() 
	   + (t->getClusters()-1) * logl((long double)nIndividuals));
}





long double
HQscore(Tree *t, const Classification &classification, const int nIndividuals)
{
    HQScoreVisitor sv(classification, nIndividuals);
    t->acceptVisitor(sv);
    long double pa= ((long double)t->getAffected())/nIndividuals;
    long double pu= ((long double)t->getUnaffected())/nIndividuals;
    return 
	- (2.0*(t->getAffected()*logl(pa) + t->getUnaffected()*logl(pu)) 
	   - 2.0*t->getScore() 
	   + 2.0*(t->getClusters()-1) * logl(logl((long double)nIndividuals)));
}






long double
AICscore(Tree *t, const Classification &classification, 
		  const int nIndividuals)
{
    AICScoreVisitor sv(classification, nIndividuals);
    t->acceptVisitor(sv);
    long double pa= ((long double)t->getAffected())/nIndividuals;
    long double pu= ((long double)t->getUnaffected())/nIndividuals;
    return - (2.0*(t->getAffected()*logl(pa) + t->getUnaffected()*logl(pu))
	      - 2.0*t->getScore() 
	      + 2.0*(t->getClusters()-1));
}



long double
Giniscore(Tree *t, const Classification &classification)
{
    GiniScoreVisitor sv(classification);
    t->acceptVisitor(sv);
    return - t->getScore();
}

