/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef CASE_CONTROL_SCORE_HH_INCLUDED
#define CASE_CONTROL_SCORE_HH_INCLUDED

/** \file
 * \brief Functions for scoring case/control datasets.
 */

#include "score.hh"
#include <snpfile/matrix.hh>

/// Representation of a affected/unaffected classification of haplotypes.
class Classification {
    /// Vector of affected/unaffected status -- 1 for cases, 0 for controls.
    std::vector<int> i_classes;
    /// Fraction of affected in the classification.
    long double i_affectedFraction;
    /// Fraction of unaffected in the classification.
    long double i_unaffectedFraction;
    /// Total number of affected.
    size_t i_noAffected;
    /// Total number of unaffected.
    size_t i_noUnaffected;

public:
    /**
     * \brief Construct new classification based on a sequence of
     * classes.
     *
     * The sequence of classes is given through a pair of iterators.
     * Each iterator should give the status of an individual, that is
     * a pair of chromosones. Thus the size of the classification is
     * twice the size of the distance between given iterators.
     * The classes in the sequence are copied into the classification
     * object, so changes to the sequence later on will not affect the
     * classification object.
     *
     * \param classesBegin      Iterator to the first class.
     * \param classesEnd        Iterator one past the last class.
     */
    template <typename Itr>
    Classification(Itr classesBegin, Itr classesEnd)
        : i_classes(classesBegin, classesEnd)
    {
        i_noAffected   = std::count(i_classes.begin(), i_classes.end(), 1);
        i_noUnaffected = std::count(i_classes.begin(), i_classes.end(), 0);
        size_t n = i_noAffected + i_noUnaffected;
        i_affectedFraction   = static_cast<long double>(i_noAffected) / n;
        i_unaffectedFraction = static_cast<long double>(i_noUnaffected) / n;
    }

    /// Get total number of chromosones in the classification.
    size_t size() const { return 2*i_classes.size(); }

    /// Get number of affected chromosones in the classification.
    size_t noAffected()   const { return 2*i_noAffected; }
    /// Get number of unaffected chromosones in the classification.
    size_t noUnaffected() const { return 2*i_noUnaffected; }

    /// Get fraction of affected in the classification.
    long double affectedFraction()   const { return i_affectedFraction; }
    /// Get fraction of unaffected in the classification.
    long double unaffectedFraction() const { return i_unaffectedFraction; }

    /**
     * \brief           Test if a given chromosone is affected.
     * \param i         Chromosone to check.
     * \returns         True if \a i is affected, false otherwise.
     */
    bool isAffected(size_t i)   const { return i_classes.at(i / 2) ==  1; }
    /**
     * \brief           Test if a given chromosone is unaffected.
     * \param i         Chromosone to check.
     * \returns         True if \a i is unaffected, false otherwise.
     */
    bool isUnaffected(size_t i) const { return i_classes.at(i / 2) ==  0; }
    /**
     * \brief           Test if a given chromosone is missing a classification.
     * \param i         Chromosone to check.
     * \returns         True if \a i is is not classified, false otherwise.
     */
    bool isMissing(size_t i)    const { return i_classes.at(i / 2) < 0; }

    /**
     * \brief Permute classification for a permutation test.
     *
     * The permutation destroys the original classification.
     */
    void permuteClasses();
};



/// Abstract class for scoring case/control datasets
class CaseControlScore : public BasicScore {
protected:
    /// Classification of dataset.
    Classification &i_classification;
    std::auto_ptr<Classification> i_backup_classification;

    CaseControlScore(BiRC::SNPFile::ImmutableMatrix &haplotypes, 
		     std::vector<int> &positions,
                     int minMarkers, double maxlen,
                     Classification &classification)
        : BasicScore(haplotypes, positions, minMarkers, maxlen),
          i_classification(classification),
          i_backup_classification(0)
    {
    }

    virtual void permute();
    virtual void resetPermutation();
};


/**
 * \brief Single marker case/control association score.
 *
 * This is simply a 2x2 contingency table, chi-square association test.
 */
class SMAScore : public CaseControlScore {
public:
    SMAScore(BiRC::SNPFile::ImmutableMatrix &haplotypes, 
	     std::vector<int> &positions,
             int minMarkers, double maxlen, Classification &classification);

    /// Calculate the single marker score for \a locus.
    virtual long double scoreLocus(int locus) const;

protected:
    /// Dummy method to make this class concrete.
    virtual long double scoreTree(Tree *tree) const;
};


/// Score with the probability score
class ProbScore : public CaseControlScore {
    /// Table used in calculating the score...
    mutable long double **i_bin;
    /// Initialize the table for the current data size.
    void tableInit(size_t size) const;

    long double logProbOne(size_t na, size_t nu) const;

public:
    ProbScore(BiRC::SNPFile::ImmutableMatrix &haplotypes, 
	      std::vector<int> &positions,
              int minMarkers, double maxlen, 
              Classification &classification);

protected:
    /// Calculate the probability score for \a tree.
    virtual long double scoreTree(Tree *tree) const;
};


/// BIC score
class BICScore : public CaseControlScore {
public:
    BICScore(BiRC::SNPFile::ImmutableMatrix &haplotypes, 
	     std::vector<int> &positions,
             int minMarkers, double maxlen, 
             Classification &classification)
        : CaseControlScore(haplotypes, positions, minMarkers,
                           maxlen, classification)
    {
    }

protected:
    /// Calculate the probability score for \a tree.
    virtual long double scoreTree(Tree *tree) const;
};

/// AIC score
class AICScore : public CaseControlScore {
public:
    AICScore(BiRC::SNPFile::ImmutableMatrix &haplotypes, 
	     std::vector<int> &positions,
             int minMarkers, double maxlen, 
             Classification &classification)
        : CaseControlScore(haplotypes, positions, minMarkers,
                           maxlen, classification)
    {
    }

protected:
    /// Calculate the probability score for \a tree.
    virtual long double scoreTree(Tree *tree) const;
};

/// HQ score
class HQScore : public CaseControlScore {
public:
    HQScore(BiRC::SNPFile::ImmutableMatrix &haplotypes, 
	    std::vector<int> &positions,
            int minMarkers, double maxlen,
            Classification &classification)
        : CaseControlScore(haplotypes, positions, minMarkers,
                           maxlen, classification)
    {
    }

protected:
    /// Calculate the probability score for \a tree.
    virtual long double scoreTree(Tree *tree) const;
};

/// Gini score
class GiniScore : public CaseControlScore {
public:
    GiniScore(BiRC::SNPFile::ImmutableMatrix haplotypes, 
	      std::vector<int> &positions,
              int minMarkers, double maxlen,
              Classification &classification)
        : CaseControlScore(haplotypes, positions, minMarkers,
                           maxlen, classification)
    {
    }

protected:
    /// Calculate the probability score for \a tree.
    virtual long double scoreTree(Tree *tree) const;
};

///new  BIC score
class newBICScore : public CaseControlScore {
    int muts_to_check;
public:
    newBICScore(BiRC::SNPFile::ImmutableMatrix &haplotypes,
		std::vector<int> &positions,
		int minMarkers, double maxlen, int nmut, 
		Classification &classification)
        : CaseControlScore(haplotypes, positions, minMarkers,
                           maxlen, classification), muts_to_check(nmut)
    {
    }
protected:
    /// Calculate the probability score for \a tree.
    virtual long double scoreTree(Tree *tree) const;
};



///new  HQ score
class newHQScore : public CaseControlScore {
    int muts_to_check;
public:
    newHQScore(BiRC::SNPFile::ImmutableMatrix &haplotypes, 
	       std::vector<int> &positions,
	       int minMarkers, double maxlen, int nmut, 
	       Classification &classification)
        : CaseControlScore(haplotypes, positions, minMarkers,
                           maxlen, classification), muts_to_check(nmut)
    {
    }
protected:
    /// Calculate the probability score for \a tree.
    virtual long double scoreTree(Tree *tree) const;
};





#endif
