/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "config.hh"
#include <snpfile/matrix.hh>
using namespace BiRC::SNPFile;
#include "compatible_cache.hh"
#include <algorithm>

namespace {
    inline bool compatibleColumns(const ImmutableMatrix &haplotypes,
				  size_t i, size_t j)
    {
		bool seen00, seen01, seen10, seen11;
		seen00 = seen01 = seen10 = seen11 = false;
		const size_t nr = haplotypes.noRows();

		for (size_t k = 0; k < nr; ++k)
			{
				int valI = haplotypes(k,i);
				int valJ = haplotypes(k,j);
				if      (valI == 0 && valJ == 0) seen00 = true;
				else if (valI == 0 && valJ == 1) seen01 = true;
				else if (valI == 1 && valJ == 0) seen10 = true;
				else if (valI == 1 && valJ == 1) seen11 = true;

				if (seen00 && seen01 && seen10 && seen11) return false;
			}

		return true;
    }
}

CompatibleCache::CompatibleCache(const ImmutableMatrix &haplotypes)
    : i_haplotypes(haplotypes)
{
    i_right.resize(haplotypes.noCols());
    i_left.resize(haplotypes.noCols());
    std::fill(i_right.begin(), i_right.end(), -1);
    std::fill(i_left.begin(), i_left.end(), -1);
}

CompatibleCache::~CompatibleCache()
{
}

void
CompatibleCache::calc_right(int i) const
{
    const size_t nc = i_haplotypes.noCols();
    size_t j = i+1;
    while (j < nc && compatibleColumns(i_haplotypes, i, j))
	++j;
    i_right[i] = j-1;	// j is now the first column _not_
                        // compatible with i.
}

void
CompatibleCache::calc_left(int i) const
{
    int j = i-1;
    while (j >= 0 && right(j) >= i)
	--j;
    i_left[i] = j+1;	// j is now the first column _not_
			// compatible with i.
}

