/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef COMPATIBLE_CACHE_HH_INCLUDED
#define COMPATIBLE_CACHE_HH_INCLUDED

/** \file
 * \brief Cache for knowing the extend a region at best can be
 * extended to obtain a compatible list of loci.
 */

#include <vector>

#include "config.hh"
namespace BiRC { namespace SNPFile { 
	class Matrix; 
	class ImmutableMatrix;
	class MatrixView;
    } 
}

/**
 * \brief Cache for knowing the extend a region at best can be
 * extended to obtain a compatible list of loci.
 */
class CompatibleCache {
protected:
    const BiRC::SNPFile::ImmutableMatrix &i_haplotypes;

    /**
     * \brief Right-vector.
     *
     * i_right[i] =
     * 	\f$\max\{j \,|\, \forall k, i \leq k \leq j: compatible(i,k) \}\f$
     */
    mutable std::vector<int> i_right;
    /**
     * \brief left-vector.
     *
     * i_left[i] = 
     * 	\f$\min\{j \,|\, \forall k, j \leq k \leq i: i_right[k] \geq i \}\f$
     */
    mutable std::vector<int> i_left;

    /// calculate the right-vector if it is not in the cache...
    virtual void calc_right(int i) const;
    /// calculate the left-vector if it is not in the cache...
    virtual void calc_left(int i)  const;

public:
    CompatibleCache(const BiRC::SNPFile::ImmutableMatrix &haplotypes);
    virtual ~CompatibleCache();

    /**
     * \brief Right-vector.
     *
     * \param i		Locus index.
     * \returns 
     *       \f$\max\{j \,|\, \forall k, i < k \leq j: compatible(i,k) \}\f$
     */
    int right(int i) const
    {
	if (i_right[i] < 0) calc_right(i);
	return i_right[i];
    }

    /**
     * \brief Left-vector.
     *
     * \param i		Locus index.
     * \returns
     *       \f$\min\{j \,|\, \forall k, j \leq k < i: right(k) \geq i \}\f$
     */
    int left(int i) const
    {
	if (i_left[i] < 0) calc_left(i);
	return i_left[i];
    }
};

#endif
