/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "config.hh"
#include "io.hh"
#include "tree.hh"
#include "tree_builder.hh"
#include "incompatibility.hh"
#include "case_control_score.hh"
#include "pph.hh"

#include <snpfile/snpfile.hh>
#include <snpfile/file_matrix.hh>
#include <snpfile/metadata_access.hh>
using namespace BiRC::SNPFile;

#include "matrix_utils.hh"

#include <boost/program_options.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/shared_ptr.hpp>

#include <iostream>
#include <fstream>
#include <stack>
#include <vector>

#include <cmath>
#include <cstdio>
#include <string>
#include <cassert>
#include <cstdlib>

#include <gsl/gsl_sf_gamma.h>



namespace options
{
    bool verbose = false;

    std::string region1FileName;
    std::string region2FileName;

    std::string scoresFileName("interaction-scores.txt");

    int minMarkers = 10;
    double maxlen = -1.0;

    std::string scoreFunction;
}



namespace
{
    typedef std::vector< boost::shared_ptr< Tree > > treesAtRegion_t;
}



treesAtRegion_t
buildTrees(CompatibleCache &cc, ImmutableMatrix &data,
           const std::vector<int> &positions, int i)
{
    treesAtRegion_t trees;

    int leftBreak, rightBreak;
    calculateIncompatibilities(data, positions, cc,
                               i, leftBreak, rightBreak,
                               options::minMarkers,
                               (options::maxlen > 0) ?
                               options::maxlen :
                               positions[positions.size()-1]-positions[0]);

    if(data.isPhased())
    {
        MatrixView w = data.subMatrix(0, data.noRows(), 
                                      leftBreak, rightBreak);
        sortColumns(w, positions, i);
        trees.push_back(boost::shared_ptr< Tree >(buildTree(w, 0, i, positions)));
    }
    else //Unphased data
    {
        /// Get the number of classes in the shadow tree for the
        /// given region
        int noSTClasses = pph::getNumOfClasses(data.view(), 
                                               leftBreak, rightBreak-1);
	    
        if(noSTClasses > 0)  // it is a compatible region
        {
            if (noSTClasses > 10) noSTClasses = 10;
		    
            int noPPHSolutions = 1 << (noSTClasses - 1); // 2**(noSTClasses-1)
	    Matrix pphSolution(newArrayMatrix(0,0,true));
		
            for(int solIdx = 1; solIdx <= noPPHSolutions; ++solIdx) 
            {
                pph::getPPHSolution(solIdx, 
                                    noPPHSolutions, 
                                    pphSolution);

                MatrixView view = pphSolution.view();

                sortColumns(view, positions, i);
                trees.push_back(boost::shared_ptr< Tree >(buildTree(view, 0, i, positions)));
            }
        }
        else 
        {  // incompatible region
	    Matrix ccSolution(newArrayMatrix(0,0,true));
            // _noSTClasses is used as _errorCode here
            pph::getCCSolution(noSTClasses, ccSolution);
            MatrixView view = ccSolution.view();

            sortColumns(view, positions, i);
            trees.push_back(boost::shared_ptr< Tree >(buildTree(view, 0, i, positions)));
        }
    }
    
    return trees;
}



namespace
{
    typedef long double score_t;

    struct Penalty
    {
        virtual ~Penalty() {}
        virtual score_t penalty(int samples) const = 0;
    };

    struct AIC : Penalty
    {
        virtual ~AIC() {}
        virtual score_t penalty(int samples) const
        {
            return 2;
        }
    };

    struct BIC : Penalty
    {
        virtual ~BIC() {}
        virtual score_t penalty(int samples) const
        {
            return logl(samples);
        }
    };

    struct HQC : Penalty
    {
        virtual ~HQC() {}
        virtual score_t penalty(int samples) const
        {
            return 2 * logl(logl(samples));
        }
    };

    struct SimpleTree;

    typedef boost::shared_ptr<SimpleTree> treePtr_t;

    struct SimpleTree
    {
        SimpleTree *parent;
        treePtr_t leftChild;
        treePtr_t rightChild;
        int matrixIdx;
        bool leaf;
        int markerPosition;



        SimpleTree(int matrixIdx)
            : parent(0),
              leftChild(),
              rightChild(),
              matrixIdx(matrixIdx),
              leaf(true)
        {
            //std::cerr << this << " ** Child construction" << std::endl;
        }

        SimpleTree(treePtr_t &left, treePtr_t &right, int markerPosition)
            : parent(0),
              leftChild(left),
              rightChild(right),
              matrixIdx(left->matrixIdx),
              leaf(false),
              markerPosition(markerPosition)
        {
            //std::cerr << this << " ** Parent construction" << std::endl;
        }

        ~SimpleTree()
        {
            //std::cerr << this << " ** Destruction" << std::endl;
        }

        bool isLeaf() const
        {
            return leaf;
        }

        bool isInner() const
        {
            return !leaf;
        }

        bool canMerge() const
        {
            return isInner() && leftChild->isLeaf() && rightChild->isLeaf();
        }

        void setLeaf()
        {
            leaf = true;
        }

        void setInner()
        {
            leaf = false;
        }
    };



    std::ostream &
    operator<<(std::ostream &os, const std::vector<int> &v)
    {
        std::vector<int>::const_iterator i;
        const char *sep = "";
        for(i = v.begin(); i != v.end(); ++i)
	    { os << sep << *i; sep = " "; }
        return os;
    }



    std::ostream &
    operator<<(std::ostream &os, const std::set<int> &v)
    {
        std::set<int>::const_iterator i;
        const char *sep = "";
        for(i = v.begin(); i != v.end(); ++i)
	    { os << sep << *i; sep = " "; }
        return os;
    }



    treePtr_t translateTree(const Tree *t, std::set<SimpleTree *> &leaves)
    {
        const InnerNode *inner = dynamic_cast<const InnerNode *>(t);
        if(inner)
        {
            treePtr_t left = translateTree(inner->getLeftChild(), leaves);
            treePtr_t right = translateTree(inner->getRightChild(), leaves);
            treePtr_t retVal(new SimpleTree(left, right, int(inner->getPos())));
            left->parent = right->parent = retVal.get();
            return retVal;
        }
        else
        {
            const LeafNode *leaf = dynamic_cast<const LeafNode *>(t);
            if(leaf)
            {
                treePtr_t retVal(new SimpleTree(leaves.size()));
                leaves.insert(retVal.get());
                return retVal;
            }
            else
            {
                assert(!"Unknown type of tree node");
                return treePtr_t();
            }
        }
    }



    struct NodeRep
    {
        SimpleTree *tree;
        int treeNu;
        int dist;
        bool sign;

        NodeRep(SimpleTree *tree, int treeNu, int centerIdx)
            : tree(tree),
              treeNu(treeNu),
              dist(abs(centerIdx - tree->markerPosition)),
              sign(centerIdx < tree->markerPosition)
        {}

        bool operator< (const NodeRep &lhs) const
        {
            if(dist < lhs.dist)
                return true;
            else if(dist == lhs.dist)
            {
                if(treeNu < lhs.treeNu)
                    return true;
                else if(treeNu == lhs.treeNu)
                {
                    if(sign < lhs.sign)
                        return true;
                }
            }

            return false;
        }
    };



    int classifyIndividuals(const Tree *tree, std::vector<int> &places, int multiplier, int idx = 0)
    {
        const InnerNode *inner = dynamic_cast<const InnerNode *>(tree);
        if(inner)
        {
            int leftSize  = classifyIndividuals(inner->getLeftChild(), 
                                                places, multiplier, idx);
            int rightSize = classifyIndividuals(inner->getRightChild(),
                                                places, multiplier, idx + leftSize);
            return leftSize + rightSize;
        }
        else
        {
            const LeafNode *leaf = dynamic_cast<const LeafNode *>(tree);
            if(leaf)
            {
                LeafNode::IDsIterator it = leaf->leafIDsBegin();
                for(; it != leaf->leafIDsEnd(); ++it)
                    places[*it] += multiplier * idx;

                return 1;
            }
            else
            {
                assert(!"Unknown type of tree node");
                return -1;
            }
        }
    }



    void distributeIndividuals(const Tree *tree1,
                               const Tree *tree2,
                               const Classification &classification,
                               std::vector<int> &A,
                               std::vector<int> &U)
    {
        std::vector<int> individualPlaces;
        individualPlaces.resize(classification.size(), 0);
        int tree1Leaves = classifyIndividuals(tree1, individualPlaces, 1);
        classifyIndividuals(tree2, individualPlaces, tree1Leaves);

        for(size_t i = 0; i < classification.size(); i++)
        {
            if(classification.isAffected(i))
                A[individualPlaces[i]]++;
            else if(classification.isUnaffected(i))
                U[individualPlaces[i]]++;
            else
                ;//Unknown type: Ignore.
        }
    }



    void findInitialCandidates(const treePtr_t &tree, std::set<NodeRep> &candidates, int treeNu, int center)
    {
        if(tree->canMerge())
            candidates.insert(NodeRep(tree.get(), treeNu, center));

        if(tree->leftChild.get() != 0)
        {
            assert(tree->rightChild.get() != 0);
            findInitialCandidates(tree->leftChild,  candidates, treeNu, center);
            findInitialCandidates(tree->rightChild, candidates, treeNu, center);
        }
    }



    score_t scoreCluster(int A, int U)
    {
        score_t Ca = A;
        score_t Cu = U;
        score_t C  = Ca + Cu;
        score_t retVal = 0;
        if(A > 0)
            retVal += Ca * logl(Ca / C);
        if(U > 0)
            retVal += Cu * logl(Cu / C);
        //std::cerr << "A " << Ca << ", U = " << Cu << ", C = " << C << std::endl;
        return -2 * retVal;
    }
}


static score_t
scoreTreeInteraction(const Tree *tree1,
                     int idx1,
                     const Tree *tree2,
                     int idx2,
                     const Penalty &penaltyFunction,
                     const Classification &classification)
{
    //Set up tree datastructures
    treePtr_t t1, t2;
    std::set<SimpleTree *> leaves1, leaves2;

    t1 = translateTree(tree1, leaves1);
    t2 = translateTree(tree2, leaves2);

    int matrixSize = leaves1.size() * leaves2.size();
    int effectiveMatrixWidth = leaves1.size();
    int effectiveMatrixHeight = leaves2.size();
    int matrixOffsetFactor = leaves1.size();

    std::vector<int> A, U;
    A.resize(matrixSize, 0);
    U.resize(matrixSize, 0);
    //std::cerr << "Matrix-size = " << leaves1.size() << "x" << leaves2.size() << std::endl;

    distributeIndividuals(tree1, tree2, classification, A, U);

    //Calculate initial score
    score_t baseScore = 0;
    for(size_t i = 0; i < A.size(); i++)
        baseScore += scoreCluster(A[i], U[i]);

    score_t penaltyFactor = penaltyFunction.penalty(classification.size());
    score_t penalty = effectiveMatrixHeight * effectiveMatrixWidth * penaltyFactor;
    score_t score = baseScore + penalty;

    //std::cerr << "Original score " << baseScore << " + " << penalty << " = " << score << std::endl;

    //Set up for iteration over candidates for merging.
    std::set<NodeRep> candidates, badCandidates;
    findInitialCandidates(t1, candidates, 1, idx1);
    findInitialCandidates(t2, candidates, 2, idx2);

    //Merge greedily
    bool improved = false;
    while(!candidates.empty() || (improved && !badCandidates.empty()))
    {
        if(candidates.empty())
        {
            candidates = badCandidates;
            badCandidates.clear();
            improved = false;
        }

        //std::cerr << "Iterate. Candidates = " << candidates.size() << ", size = " << effectiveMatrixWidth << "x" << effectiveMatrixHeight << std::endl;

        //First calculate how merge would change score...
        const NodeRep c = *candidates.rbegin();
        candidates.erase(c);
        score_t newBaseScore = baseScore;
        score_t newPenalty = 0;
        int parentIdx = c.tree->matrixIdx;
        int childIdx1 = c.tree->leftChild->matrixIdx;
        int childIdx2 = c.tree->rightChild->matrixIdx;
        if(c.treeNu == 1) // Merge column scores
        {
            std::set<SimpleTree *>::const_iterator it = leaves2.begin();
            for(; it != leaves2.end(); ++it)
            {
                int j = (*it)->matrixIdx;
                int offset1 = j * matrixOffsetFactor + childIdx1;
                assert(offset1 < matrixSize);
                int offset2 = j * matrixOffsetFactor + childIdx2;
                assert(offset2 < matrixSize);
                int A1 = A[offset1];
                int A2 = A[offset2];
                int U1 = U[offset1];
                int U2 = U[offset2];
                newBaseScore += scoreCluster(A1 + A2, U1 + U2)
                    - scoreCluster(A1, U1)
                    - scoreCluster(A2, U2);
            }
            newPenalty = effectiveMatrixHeight * (effectiveMatrixWidth - 1) * penaltyFactor;
        }
        else // Merge row scores
        {
            std::set<SimpleTree *>::const_iterator it = leaves1.begin();
            for(; it != leaves1.end(); ++it)
            {
                int j = (*it)->matrixIdx;
                int offset1 = childIdx1 * matrixOffsetFactor + j;
                assert(offset1 < matrixSize);
                int offset2 = childIdx2 * matrixOffsetFactor + j;
                assert(offset2 < matrixSize);
                int A1 = A[offset1];
                int A2 = A[offset2];
                int U1 = U[offset1];
                int U2 = U[offset2];
                newBaseScore += scoreCluster(A1 + A2, U1 + U2)
                    - scoreCluster(A1, U1)
                    - scoreCluster(A2, U2);
            }
            newPenalty = (effectiveMatrixHeight - 1) * effectiveMatrixWidth * penaltyFactor;
        }
        score_t newScore = newBaseScore + newPenalty;
        //std::cerr << "New score " << newBaseScore << " + " << newPenalty << " = " << newScore << std::endl;


        if(newScore < score) //Merge accepted. Commit/update datastructures.
        {
            //Update matrix and tree.
            if(c.treeNu == 1) // Merge columns
            {
                //std::cerr << "Accepted column merge" << std::endl;
                std::set<SimpleTree *>::const_iterator it = leaves2.begin();
                for(; it != leaves2.end(); ++it)
                {
                    int j = (*it)->matrixIdx;
                    int offset = j * matrixOffsetFactor + parentIdx;
                    assert(offset < matrixSize);
                    int offset1 = j * matrixOffsetFactor + childIdx1;
                    assert(offset1 < matrixSize);
                    int offset2 = j * matrixOffsetFactor + childIdx2;
                    assert(offset2 < matrixSize);
                    A[offset]
                        = A[offset1]
                        + A[offset2];
                    U[offset]
                        = U[offset1]
                        + U[offset2];
                }
                leaves1.erase(c.tree->leftChild.get());
                leaves1.erase(c.tree->rightChild.get());
                leaves1.insert(c.tree);
                effectiveMatrixWidth--;
            }
            else // Merge rows
            {
                //std::cerr << "Accepted row merge" << std::endl;
                std::set<SimpleTree *>::const_iterator it = leaves1.begin();
                for(; it != leaves1.end(); ++it)
                {
                    int j = (*it)->matrixIdx;
                    int offset = parentIdx * matrixOffsetFactor + j;
                    assert(offset < matrixSize);
                    int offset1 = childIdx1 * matrixOffsetFactor + j;
                    assert(offset1 < matrixSize);
                    int offset2 = childIdx2 * matrixOffsetFactor + j;
                    assert(offset2 < matrixSize);
                    A[offset]
                        = A[offset1]
                        + A[offset2];
                    U[offset]
                        = U[offset1]
                        + U[offset2];
                }
                leaves2.erase(c.tree->leftChild.get());
                leaves2.erase(c.tree->rightChild.get());
                leaves2.insert(c.tree);
                effectiveMatrixHeight--;
            }
            c.tree->setLeaf();

            //Update candidates
            if(c.tree->parent != 0 && c.tree->parent->canMerge())
                candidates.insert(NodeRep(c.tree->parent, c.treeNu, c.treeNu == 1? idx1 : idx2));

            improved = true;
            baseScore = newBaseScore;
            penalty = newPenalty;
            score = newScore;
        }
        else //Merge not accepted.
        {
            //std::cerr << "Not accepted!" << std::endl;
            badCandidates.insert(c);
        }
    }

    return score;
}



static void
interactionScores(ImmutableMatrix &data1, const std::vector<int> &positions1,
                  ImmutableMatrix &data2, const std::vector<int> &positions2,
                  const Penalty &penalty,
                  const Classification &classification,
                  std::ostream &os)
{
    //std::cerr << "In interactionScores" << std::endl;
    treesAtRegion_t::const_iterator r_itr1, r_itr2;
    TextProgressBar progress;
    CompatibleCache cc1(data1);
    CompatibleCache cc2(data2);

    if(options::verbose) progress.start();

    for(int i = 0; i < data1.noCols(); ++i)
	{
        const treesAtRegion_t r1Trees = buildTrees(cc1, data1, positions1, i);

        for(int j = 0; j < data2.noCols(); ++j)
		{
            const treesAtRegion_t r2Trees = buildTrees(cc2, data2, positions2, j);

		    if(options::verbose)
                progress.showProgress(i,
                                      data1.noCols(),
                                      j,
                                      data2.noCols());

            score_t scoreSum = 0.0;
            int scoreCount = 0;

            treesAtRegion_t::const_iterator t_itr1, t_itr2;
            for(t_itr1 = r1Trees.begin(); t_itr1 != r1Trees.end(); ++t_itr1)
            {
                for(t_itr2 = r2Trees.begin(); t_itr2 != r2Trees.end(); ++t_itr2)
                {
                    //std::cerr << "Score" << std::endl;
                    scoreSum += scoreTreeInteraction(t_itr1->get(), i, t_itr2->get(), j,
                                                     penalty,
                                                     classification);
                    scoreCount++;
                }
            }

            os << (scoreSum / scoreCount) << ' ';
		}

	    os << std::endl;
	}

    if(options::verbose) progress.end();
}



namespace po = boost::program_options;
void
printUsage(const char *progname, 
           po::options_description &desc, 
           std::ostream &os = std::cerr)
{
    os << "Usage: " << progname << " [options] region1 region2"
       << std::endl
       << desc << std::endl;
    
}

void printArgError(const char *progname, 
                   po::options_description &desc, 
                   const std::string &error,
                   std::ostream &os = std::cerr)
{
    using namespace TextManip;
    os << bold << red
       << "Error: " << reset
       << bold
       << error << reset
       << std::endl << std::endl;
    printUsage(progname, desc, os);
}




int
main(int argc, char *argv[])
{
    using namespace TextManip;

    po::options_description desc("Allowed options");
    po::variables_map vm;
    desc.add_options()
        ("help,h", "Produce help message.")
        ("verbose,v", "Turns on verbose output.")

        ("region1", po::value<std::string>(&options::region1FileName),
         "First region.")
        ("region2", po::value<std::string>(&options::region2FileName),
         "Second region.")

        ("scores,o",
         po::value<std::string>(&options::scoresFileName)
         ->default_value("scores.txt"),
         "File where scores will be written.")
        ("minmarkers,m",
         po::value<int>(&options::minMarkers)->default_value(10),
         "Minimum number of markers used (default: 10).")
        ("deslen,x",
         po::value<double>(&options::maxlen)->default_value(-1),
         "Maximum length of interval (default: whole range).")
        ("scoreFunction,f",
         po::value<std::string>(&options::scoreFunction),
         "Score funcition (H = HQC, B = BIC, A = AIC")
        ;
    po::positional_options_description p;
    p.add("region1", 1);
    p.add("region2", 1);

    try
    {
        po::store(po::command_line_parser(argc, argv)
                  .options(desc).positional(p).run(), 
                  vm, false);
        po::notify(vm);
    }
    catch (std::exception &e)
    {
        printArgError(argv[0], desc, std::string("While parsing: ") + e.what());
        return EXIT_FAILURE;
    }

    if(vm.count("help"))
    {
        printUsage(argv[0], desc, std::cout);
        return EXIT_SUCCESS;
    }

    options::verbose = vm.count("verbose") > 0;


    if(options::scoreFunction.size() != 1)
    {
        printArgError(argv[0], desc, "No score function given.");
        return EXIT_FAILURE;
    }

    std::auto_ptr<Penalty> penalty;

    if(options::scoreFunction[0]=='A')
	{
        penalty = std::auto_ptr<Penalty>(new AIC());
    }
    else if(options::scoreFunction[0]=='B')
    {
        penalty = std::auto_ptr<Penalty>(new BIC());
    }
    else if(options::scoreFunction[0]=='H')
    {
        penalty = std::auto_ptr<Penalty>(new HQC());
    }
    else
    {
        printArgError(argv[0], desc,
                      "Not a valid name for a scoreFunction, use A, B or H.");
	    return EXIT_FAILURE;
    }

    if(options::region1FileName == "")
	{
        printArgError(argv[0], desc, "No file name for the first region specified.");
	    return EXIT_FAILURE;
	}
    if(options::region2FileName == "")
	{
        printArgError(argv[0], desc, "No file name for the second region specified.");
	    return EXIT_FAILURE;
	}

    try
    {
        using namespace BiRC::SNPFile;

        SNPFile region1(options::region1FileName.c_str());
        SNPFile region2(options::region2FileName.c_str());

        // FIXME: Validate input files (same number of individuals and
        // such).

	ImmutableMatrix region1genotypes(newReadOnlyFileMatrix(region1));
	ImmutableMatrix region2genotypes(newReadOnlyFileMatrix(region2));

        std::vector<bool> diseaseStatus;
        std::vector<int> region1positions, region2positions;

        getMetadata(region1, "phenotype", diseaseStatus);
        getMetadata(region1, "positions", region1positions);
        getMetadata(region2, "positions", region2positions);

        Classification classification(diseaseStatus.begin(), 
                                      diseaseStatus.end());


        std::ofstream scoresFile(options::scoresFileName.c_str());
        if (!scoresFile)
	    {
            std::cerr << "Error opening file " 
                      << options::scoresFileName
                      << std::endl;
            return EXIT_FAILURE;
	    }
	
        // main computation part...
        if (options::verbose)
            std::cout << bold
                      << "Running " << PACKAGE_NAME << " v" << PACKAGE_VERSION
                      << std::endl
                      << "Mapping interactions."
                      << reset
                      << std::endl << std::endl;


        if (options::verbose)
            std::cout << "Scoring interactions." << std::endl;
	
        interactionScores(region1genotypes, region1positions,
                          region2genotypes, region2positions,
                          *penalty,
                          classification, scoresFile);


    }
    catch(BiRC::SNPFile::SNPFileError &e)
    { 
        std::cerr << "Error accessing SNP file: "
                  << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch(std::exception &e)
    { 
        std::cerr << "Error: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }

    return EXIT_SUCCESS;
}
