/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "genotype_score.hh"
#include "tree.hh"

#include <snpfile/matrix.hh>
using namespace BiRC::SNPFile;

#include <limits>
#include <functional>
#include <cassert>
#include <cmath>

#include <gsl/gsl_sf_gamma.h>



GenotypeScore::GenotypeScore(ImmutableMatrix &haplotypes, 
                             std::vector<int> &positions,
                             int minMarkers, double maxlen,
                             Classification &classification,
                             int noMutations)
    : CaseControlScore(haplotypes, positions, 
                       minMarkers, maxlen, classification),
      // we divide by two here, to move the classification from
      // haplotypes to genotypes.
      nullModel(gsl_sf_lnbeta(classification.noAffected()/2+1,
                              classification.noUnaffected()/2+1)),
      i_noMutations(noMutations)
{
    //Set sizes of scores tables
    scoreVsMutationCount.resize(noMutations);
    noModelsVsMutationCount.resize(noMutations);

    noTypes = (i_noMutations + 1)*(i_noMutations + 2) / 2;
    nA.resize(noTypes);
    nU.resize(noTypes);

    //Calculate typeId lookup table
    typeIdTable.reset(new boost::scoped_array<int>[noMutations + 1]);
    for(int i = 0; i <= noMutations; i++)
	{
	    typeIdTable[i].reset(new int[noMutations + 1]);
	    for(int j = 0; j <= noMutations; j++)
		{
		    if(i < j)
                typeIdTable[i][j] = j * (j + 1) / 2 + i;
		    else
                typeIdTable[i][j] = i * (i + 1) / 2 + j;
		}
	}
}



namespace {
    inline long double
    logspace_add (long double logx, long double logy)
    {
        return std::max(logx, logy) + log1p(exp(-fabs(logx - logy)));
    }
}



void
GenotypeScore::print(std::ostream &os) const
{
    //Since mutant array don't exist anymore, following is irrelevant
    //but should possibly be replaced with something equivalent.
    //os << "i\th\tm\ts" << std::endl;
    //for (size_t j = 0; j < i_classification.size(); ++j)
    //    os << (j/2) << '\t' << j << '\t' << i_mutant[j]
    //       << '\t' << i_classification.isAffected(j) << std::endl;
    //os << std::endl;

    long double totalScore = -std::numeric_limits<double>::infinity();
    unsigned int noModels = 0;

    for(int i = 0; i < i_noMutations; i++)
	{
	    if(totalScore == -std::numeric_limits<double>::infinity())
            totalScore = scoreVsMutationCount[i];
	    else
            totalScore = logspace_add(totalScore, scoreVsMutationCount[i]);
	    noModels += noModelsVsMutationCount[i];

	    os << "With only " << (i + 1) << " mutation(s): "
	       << "noModels = " << noModelsVsMutationCount[i]
	       << " score = " << (scoreVsMutationCount[i] - std::log(noModelsVsMutationCount[i]) - nullModel)
	       << std::endl;

	    if(i > 0)
		{
		    os << "With 1 .. " << (i + 1) << " mutation(s): "
		       << "noModels = " << noModels
		       << " score = " << (totalScore - std::log(noModels) - nullModel)
		       << std::endl;
		}

	}

    for(int i = 0; i < noTypes; i++)
        os << "i = " << i << ", A[i] = " << nA[i] << ", U[i] = " << nU[i] << std::endl;
}



void
GenotypeScore::updateSuperIndividual(SuperIndividual *si, int m1, int m2) const
{
    const int oldId = typeIdTable[si->leftMutation][si->rightMutation];
    const int newId = typeIdTable[m1][m2];
    si->leftMutation = m1;
    si->rightMutation = m2;

    nA[oldId] -= si->nA;
    nA[newId] += si->nA;

    nU[oldId] -= si->nU;
    nU[newId] += si->nU;
}



void
GenotypeScore::colourSubtree(int idx, int mutant) const
{
    Tree *tree = nodeList[idx];

    nodeColour[idx] = mutant;

    if (dynamic_cast<LeafNode*>(tree))
	{
	    LeafInfo &li = leafInfo[idx];
	    int i = 0;
	    for(; li.ind[i] != NULL && li.ind[i]->owner != li.id && li.ind[i]->owner != -1; i++)
            updateSuperIndividual(li.ind[i].get(), mutant, li.ind[i]->rightMutation);

	    if(li.ind[i] != NULL && li.ind[i]->owner == -1)
            updateSuperIndividual(li.ind[i++].get(), mutant, mutant);

	    for(; i < noLeaves && li.ind[i] != NULL; i++) 
            updateSuperIndividual(li.ind[i].get(), li.ind[i]->leftMutation, mutant);
	}
    else if (dynamic_cast<InnerNode*>(tree))
	{
	    colourSubtree(leftChild[idx],  mutant);
	    colourSubtree(rightChild[idx], mutant);
	}
    else
        assert(!"Shouldn't reach this point");
}



void
GenotypeScore::scoreRecursively(int mutation, int minNode, int maxNode) const
{
    int colour = mutation + 1;
 
    for(int i = minNode; i < maxNode; i++)
	{
	    int prevColour = nodeColour[i];

	    colourSubtree(i, colour);

	    scoreCurrent(mutation);

	    if(mutation + 1 < i_noMutations)
            scoreRecursively(mutation + 1, i + 1, maxNode);

	    colourSubtree(i, prevColour);
	}
}



long double
GenotypeScore::getCurrentScore() const
{
    long double retVal = 0;
    for(int i = 0; i < noTypes; i++)
        retVal += gsl_sf_lnbeta(nA[i] + 1, nU[i] + 1);

    return retVal;
}



void
GenotypeScore::scoreCurrent(int noMutations) const
{
    noModelsVsMutationCount[noMutations]++;

    if (scoreVsMutationCount[noMutations] == -std::numeric_limits<double>::infinity())
        scoreVsMutationCount[noMutations] = getCurrentScore();
    else
        scoreVsMutationCount[noMutations] = logspace_add(scoreVsMutationCount[noMutations],getCurrentScore());
}



int 
GenotypeScore::flattenTree(Tree *tree, int next) const
{
    InnerNode* node = dynamic_cast<InnerNode *>(tree);

    nodeList[next] = tree;

    if(!node)
	{
	    leftChild[next] = -1;
	    rightChild[next] = -1;

	    return next + 1;
	}

    int leftNext = flattenTree(node->getLeftChild(), next + 1);
    leftChild[next] = next + 1;

    int rightNext = flattenTree(node->getRightChild(), leftNext);
    rightChild[next] = leftNext;

    return rightNext;
}



namespace {
    class UpdateLeafMap
    {
        int *map;
        unsigned value;

    public:

        UpdateLeafMap(int *map, unsigned value)
            : map(map), value(value)
        {}

        void operator() (int i)
        {
            map[i] = value;
        }
    };
}



void
GenotypeScore::buildLeafInfo(const Tree *tree) const
{
    int leafMap[i_classification.size()];

    //Create SuperIndividuals.
    int leafNu1 = 0;
    for(size_t i = 0; i < nodeList.size(); i++)
	{
	    LeafNode *leaf1 = dynamic_cast<LeafNode *>(nodeList[i]);
	    if(!leaf1)
		{
		    leafInfo[i].id = -1;
		    continue;
		}

	    std::for_each(leaf1->leafIDsBegin(), leaf1->leafIDsEnd(),
                      UpdateLeafMap(leafMap, i));

	    leafInfo[i].id = leafNu1;
	    leafInfo[i].ind.reset(new boost::shared_ptr<SuperIndividual>[noLeaves]);

	    int leafNu2 = 0;
	    for(size_t j = 0; j < nodeList.size(); j++)
		{
		    LeafNode *leaf2 = dynamic_cast<LeafNode *>(nodeList[j]);
		    if(!leaf2)
                continue;

		    if(i <= j)
			{
			    SuperIndividual *si = new SuperIndividual;
			    si->owner = i == j ? -1 : leafNu1;
			    si->nA = 0;
			    si->nU = 0;
			    si->leftMutation = 0;
			    si->rightMutation = 0;
			    leafInfo[i].ind[leafNu2].reset(si);
			}
		    else
			{
			    leafInfo[i].ind[leafNu2] = leafInfo[j].ind[leafNu1];
			}

		    leafNu2++;
		}

	    leafNu1++;
	}

    //Fill out nA and nU info in SuperIndividuals.
    for(size_t i = 0; i < i_classification.size(); i+=2)
	{
	    boost::shared_ptr<SuperIndividual> &si = leafInfo[leafMap[i]].ind[leafInfo[leafMap[i+1]].id];
	    if(i_classification.isAffected(i))
            si->nA++;
	    else if(i_classification.isUnaffected(i))
            si->nU++;
	    else
            ;//Unknown. Ignore...
	}

    //Remove empty SuperIndividuals
    leafNu1 = 0;
    for(size_t i = 0; i < nodeList.size(); i++)
	{
	    if(leafInfo[i].id == -1)
            continue;

	    boost::scoped_array<boost::shared_ptr<SuperIndividual> > &ind = leafInfo[i].ind;

	    int target = 0;
	    for(int j = 0; j < noLeaves; j++)
		{
		    if(ind[j]->nA != 0 || ind[j]->nU != 0)
			    ind[target++] = ind[j];
		}
	    for(int j = target; j < noLeaves; j++)
            ind[j].reset();

	    leafNu1++;
	}
}



long double
GenotypeScore::scoreTree(Tree *tree) const
{
    if(!dynamic_cast<InnerNode *>(tree) || i_noMutations == 0)
        return 0; // this model equals the null model

    int noNodes = tree->getNoNodes();
    noLeaves = tree->getNoNodes() - tree->getNoInnerNodes();
    nodeList.resize(noNodes);
    leftChild.resize(noNodes);
    rightChild.resize(noNodes);
    nodeColour.resize(noNodes);

    leafInfo.reset(new LeafInfo[noNodes]);

    std::fill(nodeColour.begin(), nodeColour.end(), 0);

    std::fill(scoreVsMutationCount.begin(), scoreVsMutationCount.end(), -std::numeric_limits<double>::infinity());
    std::fill(noModelsVsMutationCount.begin(), noModelsVsMutationCount.end(), 0);

#if 0
    std::cout << "No affected " << tree->getAffected() << std::endl
              << "No unaffected " << tree->getUnaffected() << std::endl
              << "No Clusters " << tree->getClusters() << std::endl
              << "No nodes " << tree->getNoNodes() << std::endl
              << "No inner nodes " << tree->getNoInnerNodes() << std::endl
              << "No individuals, pr classification " << i_mutant.size() << std::endl;
#endif

    //Build tree and leaf datastructures.
    flattenTree(tree, 0);

    buildLeafInfo(tree);

    //Reset counts.
    std::fill(nA.begin(), nA.end(), 0);
    std::fill(nU.begin(), nU.end(), 0);
    nA[0] = i_classification.noAffected() / 2;
    nU[0] = i_classification.noUnaffected() / 2;

    //Calculate scores.
    //Note: minIndex of 1 => Don't mutate on root
    //      minIndex of 2 => Also skip left child of root
    scoreRecursively(0, 2, noNodes);

    //Garbage collect leaf datastructures...
    leafInfo.reset();

    //Calculate total score.
    long double totalScore = -std::numeric_limits<double>::infinity();
    unsigned int noModels = 0;

    for(int i = 0; i < i_noMutations; i++)
	{
	    if(totalScore == -std::numeric_limits<double>::infinity())
            totalScore = scoreVsMutationCount[i];
	    else
            totalScore = logspace_add(totalScore, scoreVsMutationCount[i]);
	    noModels += noModelsVsMutationCount[i];
	}

    totalScore -= std::log(noModels);
    
    //std::cout << (*tree) << ' ' << totalScore << std::endl;
    return totalScore - nullModel;
}
