/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef GENOTYPE_SCORE_INCLUDED
#define GENOTYPE_SCORE_INCLUDED

#include "case_control_score.hh"
#include "tree.hh"

#include <boost/scoped_array.hpp>
#include <boost/shared_ptr.hpp>

#include <iostream>


class GenotypeScore : public CaseControlScore {
public:
    GenotypeScore(BiRC::SNPFile::ImmutableMatrix &haplotypes,
		  std::vector<int> &positions,
                  int minMarkers, double maxlen, 
                  Classification &classification,
                  int noMutations);

    /// Print the state of this to \a os for debugging purposes.
    void print(std::ostream &os) const;

    /// Calculate the probability score for \a tree.
    virtual long double scoreTree(Tree *tree) const;

private:
    /// the score of the null-model, for comparison when scoring
    const long double nullModel;
    /// List of all internal nodes.
    mutable std::vector<Tree *> nodeList;
    /// List of left children.
    mutable std::vector<int> leftChild;
    /// List of right children.
    mutable std::vector<int> rightChild;
    /// Colour of inner node.
    mutable std::vector<int> nodeColour;
    /// Number of mutations to consider simultaneously
    const int i_noMutations;


    /// Number of types
    mutable int noTypes;
    /// Id of mutations types lookup
    mutable boost::scoped_array< boost::scoped_array< int > > typeIdTable;
    /// Number of affected individuals with a given type
    mutable std::vector<int> nA;
    /// Number of unaffected individuals with a given type
    mutable std::vector<int> nU;

    /// Scores for a given number of mutations
    mutable std::vector<long double> scoreVsMutationCount;
    /// No models for a given number of mutations
    mutable std::vector<int> noModelsVsMutationCount;


    /// Merges several similar individuals into one super-individual
    struct SuperIndividual
    {
        int owner; //Id of owning leaf, -1 means only one leaf has access and is owner
        int nA;
        int nU;
        int leftMutation;
        int rightMutation;
    };

    /// Pointers form leaves, to superIndividuals
    struct LeafInfo
    {
        boost::scoped_array<boost::shared_ptr<SuperIndividual> > ind;
        int id;
    };

    /// Functor to help construct the super individual structure.
    class MakeSuperIndividual;
    /// Associates leaves, with their info.
    mutable boost::scoped_array<LeafInfo> leafInfo;
    /// Number of leaves in tree.
    mutable int noLeaves;


    /// Traverse tree and compute the score for each edge, summing the
    /// scores in \a totalScore.
    void scoreRecursively(int mutations, int minNode, int maxNode) const;
    /// Changes colour of a superIndividual
    void updateSuperIndividual(SuperIndividual *si, int m1, int m2) const;
    /// Colour all leaves in \a tree as \a mutant
    void colourSubtree   (int idx, int mutant) const;
    /// Score according to the current colouring.
    void scoreCurrent(int noMutations) const;
    /// Create internal flat structure of tree
    int flattenTree(Tree *tree, int next) const;
    /// Prepares the leafInfo array
    void buildLeafInfo(const Tree *tree) const;
    /// Get the current score value
    long double getCurrentScore() const;

};

inline std::ostream &
operator<<(std::ostream &os, const GenotypeScore &gs)
{
    gs.print(os);
    return os;
}

#endif
