/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "config.hh"
#include "incompatibility.hh"
#include <snpfile/matrix.hh>
#include <snpfile/array_matrix.hh>
using namespace BiRC::SNPFile;
#include "compatible_cache.hh"

namespace {

    struct ClosestTo {
        const std::vector<int> &loci;
        int locus;
        int leftIndex, rightIndex;
        ClosestTo(const std::vector<int> &loci, int locusIndex)
            : loci(loci), locus(loci[locusIndex]),
              leftIndex(locusIndex-1), rightIndex(locusIndex+1)
        {}
        int next()
        {
            if (leftIndex < 0 && rightIndex >= (int)loci.size())
                        return -1;

            if (leftIndex < 0)                  return rightIndex++;
            if (rightIndex >= (int)loci.size()) return leftIndex--;

            double deltaLeft  = locus - loci[leftIndex];
            double deltaRight = loci[rightIndex] - locus;
            if (deltaLeft < deltaRight)         return leftIndex--;
            else                                return rightIndex++;
        }
    };


    inline bool compatibleColumns(const ImmutableMatrix &haplotypes,
				  int i, int j)
    {
                bool seen00, seen01, seen10, seen11;
                seen00 = seen01 = seen10 = seen11 = false;

                for (int k = 0; k < haplotypes.noRows(); ++k)
                        {
                        int valI = haplotypes(k,i);
                        int valJ = haplotypes(k,j);
                        if      (valI == 0 && valJ == 0) seen00 = true;
                        else if (valI == 0 && valJ == 1) seen01 = true;
                        else if (valI == 1 && valJ == 0) seen10 = true;
                        else if (valI == 1 && valJ == 1) seen11 = true;

                        if (seen00 && seen01 && seen10 && seen11) return false;
            }

                return true;
    }

    inline bool compatible(const ImmutableMatrix &haplotypes, 
			   int next, int left, int right)
    {
                for (int i = left; i < right; ++i)
                        if (!compatibleColumns(haplotypes, next, i))
                                return false;
                return true;
    }
}

void
calculateIncompatibilities(const ImmutableMatrix &haplotypes,
                           const std::vector<int> &positions,
                           int i, int &leftBreak, int &rightBreak,
                           int minMarkers, double maxlen)
{
    int leftIdx, rightIdx, nextIdx;
    leftIdx = rightIdx = i;
  
    if (maxlen < 0) maxlen = positions[positions.size()-1] - positions[0];
    
    ClosestTo ct(positions,i);
    while ((nextIdx = ct.next()) >= 0 && 
           compatible(haplotypes, nextIdx, leftIdx, rightIdx+1))
      {
          leftIdx  = std::min(leftIdx, nextIdx);
          rightIdx = std::max(rightIdx, nextIdx);
      }
    
    if (nextIdx > rightIdx)
        // incomp to the right, try to extend to the left
        for (int j = leftIdx; j >= 0; --j)
            if (compatible(haplotypes, j, leftIdx, rightIdx+1))
                leftIdx = j;
            else
                break;
    else
        // incomp to the left, try to extend to the right
        for (int j = rightIdx+1; j < static_cast<int>(haplotypes.noCols()); ++j)
            if (compatible(haplotypes, j, leftIdx, rightIdx+1))
                rightIdx = j;
            else
                break;
    
    int usedMarkers = rightIdx + 1 - leftIdx;
    while(usedMarkers < minMarkers && 
          nextIdx >= 0 &&
          positions[std::max(rightIdx, nextIdx)]-positions[i]<=maxlen &&
          positions[i]-positions[std::min(leftIdx, nextIdx)]<=maxlen)
        {
            if( nextIdx < leftIdx || nextIdx > rightIdx){
                leftIdx  = std::min(leftIdx, nextIdx);
                rightIdx = std::max(rightIdx, nextIdx);
                usedMarkers++;
            }
            nextIdx= ct.next();
        }
    
    leftBreak  = leftIdx;
    rightBreak = rightIdx + 1;
}




void
calculateIncompatibilities(const ImmutableMatrix &haplotypes,
                           const std::vector<int> &positions,
                           const CompatibleCache &cc,
                           int i, int &leftBreak, int &rightBreak,
                           int minMarkers, double maxlen)
{
    int leftIdx, rightIdx, nextIdx;
    leftIdx = rightIdx = i;

    if (maxlen < 0) maxlen = positions[positions.size()-1] - positions[0];

    ClosestTo ct(positions,i);
    while ((nextIdx = ct.next()) >= 0)
        {
            if (nextIdx < i)
                {
                    if (cc.right(nextIdx) < rightIdx) break;
                    leftIdx  = nextIdx;
                }
            else
                {
                    if (cc.left(nextIdx) > leftIdx) break;
                    rightIdx = nextIdx;
                }
        }
    
    if (nextIdx > rightIdx)
        // incomp to the right, try to extend to the left
        for (int j = leftIdx; j >= 0; --j)
            if (cc.right(j) >= rightIdx)
                leftIdx = j;
            else
                break;
    else
        // incomp to the left, try to extend to the right
        for (int j = rightIdx+1; j < static_cast<int>(haplotypes.noCols()); ++j)
            if (cc.left(j) <= leftIdx)
		rightIdx = j;
            else
		break;
    
    int usedMarkers = rightIdx + 1 - leftIdx;
    while(usedMarkers < minMarkers && 
          nextIdx >= 0 &&
          positions[std::max(rightIdx, nextIdx)]-positions[i] <= maxlen &&
          positions[i]-positions[std::min(leftIdx, nextIdx)] <= maxlen)
        {
            if (nextIdx < leftIdx || nextIdx > rightIdx) {
                leftIdx  = std::min(leftIdx, nextIdx);
                rightIdx = std::max(rightIdx, nextIdx);
                usedMarkers++;
            }
            nextIdx = ct.next();
        }
    
    leftBreak  = leftIdx;
    rightBreak = rightIdx + 1;
}

