/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef INCOMPATIBILITY_HH_INCLUDED
#define INCOMPATIBILITY_HH_INCLUDED

/** \file
 *  \brief Methods for calculating regions to build trees for.
 *
 * Conceptually, the regions are maximal compatible regions, but it is
 * possible to specify a minimal number of markers to include (even if
 * they are incompatible) or a maximal region length (even if the
 * region can be expanded to a wide compatible region).
 */

#include "config.hh"
namespace BiRC { namespace SNPFile { 
	class ImmutableMatrix;
	class Matrix;
	class MatrixView;
    }
}
class CompatibleCache;

#include <vector>

/**
 * \brief Calculate the region, around a point, over which to build a tree.
 *
 * Find the index of the leftmost and the rightmost marker to include
 * in the region for a tree at a given marker index.
 *
 * Conceptually, the regions are maximal compatible regions, but it is
 * possible to specify a minimal number of markers to include (even if
 * they are incompatible) or a maximal region length (even if the
 * region can be expanded to a wide compatible region).
 *
 * \param haplotypes		A matrix view representing the haplotypes.
 * \param positions		Positions for all markers.
 * \param i			Index of the marker this region is for.
 * \param leftBreak		Reference that will be assigned the index of
 * 				the leftmost marker in the region.
 * \param rightBreak		Reference that will be assigned the index
 * 				one past the rightmost marker in the region.
 * \param minMarkers		The minimal markers to include in the region,
 * 				even if this means the region contains
 *				incompatible markers.
 * \param maxlen		The maximal length the region is allowed to be
 *				in the units used by \a positions.
 */
void
calculateIncompatibilities(const BiRC::SNPFile::ImmutableMatrix &haplotypes,
			   const std::vector<int> &positions,
			   int i, int &leftBreak, int &rightBreak,
			   int minMarkers, double maxlen);


/**
 * \brief Calculate the region, around a point, over which to build a tree.
 *
 * Find the index of the leftmost and the rightmost marker to include
 * in the region for a tree at a given marker index.
 *
 * Conceptually, the regions are maximal compatible regions, but it is
 * possible to specify a minimal number of markers to include (even if
 * they are incompatible) or a maximal region length (even if the
 * region can be expanded to a wide compatible region).
 *
 * \param haplotypes		A matrix view representing the haplotypes.
 * \param positions		Positions for all markers.
 * \param cc			Cache for speeding up the region computation.
 * \param i			Index of the marker this region is for.
 * \param leftBreak		Reference that will be assigned the index of
 * 				the leftmost marker in the region.
 * \param rightBreak		Reference that will be assigned the index
 * 				one past the rightmost marker in the region.
 * \param minMarkers		The minimal markers to include in the region,
 * 				even if this means the region contains
 *				incompatible markers.
 * \param maxlen		The maximal length the region is allowed to be
 *				in the units used by \a positions.
 */
void
calculateIncompatibilities(const BiRC::SNPFile::ImmutableMatrix &haplotypes,
			   const std::vector<int> &positions,
			   const CompatibleCache &cc,
			   int i, int &leftBreak, int &rightBreak,
			   int minMarkers, double maxlen);


#endif
