/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

// test to use while refactoring the score framework -- tests that the
// new score classes compute the same scores as the old score
// functions.


#include "config.hh"
#include "io.hh"
#include "incompatibility.hh"
#include "compatible_cache.hh"

#include <snpfile/matrix.hh>
#include <snpfile/array_matrix.hh>
using namespace BiRC::SNPFile;

#include <iostream>

#undef NDEBUG
#include <cassert>

namespace {
    inline bool compatibleColumns(const ImmutableMatrix &haplotypes, 
				  int i, int j)
    {
	bool seen00, seen01, seen10, seen11;
	seen00 = seen01 = seen10 = seen11 = false;

	for (int k = 0; k < haplotypes.noRows(); ++k)
	    {
		int valI = haplotypes(k,i);
		int valJ = haplotypes(k,j);
		if (valI == 0 && valJ == 0) seen00 = true;
		if (valI == 0 && valJ == 1) seen01 = true;
		if (valI == 1 && valJ == 0) seen10 = true;
		if (valI == 1 && valJ == 1) seen11 = true;

		if (seen00 && seen01 && seen10 && seen11) return false;
	    }

	return true;
    }

    inline bool compatible(const Matrix &haplotypes, 
			   int i, int left, int right)
    {
	for (int j = left; j <= right; ++j)
	    if (!compatibleColumns(haplotypes, i, j))
                return false;
	return true;
    }
}


static void
testCF()
{
    std::vector<int> positions;
    Matrix haplotypes(newArrayMatrix(0,0));
    std::vector<int> classes;

    positions = readPositions("testdata/CF/cf-positions.txt");
    readHaplotypes("testdata/CF/cf-haplotypes.txt", haplotypes, classes);
    assert((int)positions.size() == haplotypes.noCols());

    CompatibleCache cc(haplotypes);

    for (int i = 0; i < static_cast<int>(haplotypes.noCols()); ++i)
	assert(compatible(haplotypes, i, cc.left(i), cc.right(i)));

    for (int i = 0; i < static_cast<int>(haplotypes.noCols()); ++i)
	{
	    int oldLB, oldRB;
	    int newLB, newRB;
	    calculateIncompatibilities(haplotypes, positions, i,
				       oldLB, oldRB, 0, -1);
	    calculateIncompatibilities(haplotypes, positions, cc, i,
				       newLB, newRB, 0, -1);

	    assert(oldLB == newLB);
	    assert(oldRB == newRB);
	}
}

static void
testInteractions()
{
    std::vector<int> positions;
    Matrix haplotypes(newArrayMatrix(0,0));
    std::vector<int> classes;

    positions = readPositions("testdata/interaction/positions.txt");
    readHaplotypes("testdata/interaction/haplotypes.txt", 
		   haplotypes, classes);
    assert((int)positions.size() == haplotypes.noCols());

    CompatibleCache cc(haplotypes);

    for (int i = 0; i < static_cast<int>(haplotypes.noCols()); ++i)
	assert(compatible(haplotypes, i, cc.left(i), cc.right(i)));

    for (int i = 0; i < static_cast<int>(haplotypes.noCols()); ++i)
	{
	    int oldLB, oldRB;
	    int newLB, newRB;
	    calculateIncompatibilities(haplotypes, positions, i,
				       oldLB, oldRB, 0, -1);
	    calculateIncompatibilities(haplotypes, positions, cc, i,
				       newLB, newRB, 0, -1);

	    assert(oldLB == newLB);
	    assert(oldRB == newRB);
	}
}

int
main()
{
    testCF();
    testInteractions();

    return 0;
}

