/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "config.hh"
#include "io.hh"
#include <snpfile/array_matrix.hh>
using namespace BiRC::SNPFile;
#include "matrix_utils.hh"
#include "tree.hh"
#include "tree_builder.hh"
#include "incompatibility.hh"
#include "case_control_score.hh"
#include "pph.hh"

#include <iostream>
#include <fstream>
#include <set>
#include <cmath>
#include <cstdio>
#include <string>
#include <cassert>

#include <gsl/gsl_errno.h>

#include <gsl/gsl_sf_gamma.h>
#include <boost/program_options.hpp>
#include <boost/shared_ptr.hpp>



namespace options {
    bool verbose = false;

    std::string region1positionsFileName;
    std::string region1haplotypesFileName;
    std::string region2positionsFileName;
    std::string region2haplotypesFileName;

    std::string scoresFileName("interaction-scores.txt");

    int minMarkers = 10;
    double maxlen = -1.0;

    bool unphased = true;
}



namespace {
    typedef std::vector< std::vector< boost::shared_ptr< Tree > > > treesAtRegion_t;
}



treesAtRegion_t
buildTrees(ImmutableMatrix &data, 
	   const std::vector<int> &positions)
{
    treesAtRegion_t trees;
    CompatibleCache cc(data);

    TextProgressBar progress;

    if (options::verbose) progress.start();
    for (int i = 0; i < data.noCols(); ++i)
	{
        std::vector< boost::shared_ptr< Tree > > treesAtPosition;

	    if (options::verbose) progress.showProgress(i, data.noCols());

	    int leftBreak, rightBreak;
        calculateIncompatibilities(data, positions, cc,
				       i, leftBreak, rightBreak,
				       options::minMarkers,
				       (options::maxlen > 0) ?
				       options::maxlen :
				       positions[positions.size()-1]-positions[0]);
        if(!options::unphased)
        {
            MatrixView w = data.subMatrix(0, data.noRows(), 
                                          leftBreak, rightBreak);
            sortColumns(w, positions, i);
            treesAtPosition.push_back(boost::shared_ptr< Tree >(buildTree(w, 0, i, positions)));
        }
        else //Unphased data
        {
            /// Get the number of classes in the shadow tree for the
            /// given region
            int noSTClasses = pph::getNumOfClasses(data.view(), 
                                                   leftBreak, rightBreak-1);
	    
            if(noSTClasses > 0)  // it is a compatible region
            {
                if (noSTClasses > 10) noSTClasses = 10;
		    
                int noPPHSolutions = 1 << (noSTClasses - 1); // 2**(noSTClasses-1)
		
		Matrix pphSolution(newArrayMatrix(0,0,true));
		
                for (int solIdx = 0; solIdx < noPPHSolutions; ++solIdx) 
                {
                    pph::getPPHSolution(solIdx, 
                                        noPPHSolutions, 
                                        pphSolution);

                    MatrixView view = pphSolution.view();

                    sortColumns(view, positions, i);
                    treesAtPosition.push_back(boost::shared_ptr< Tree >(buildTree(view, 0, i, positions)));
                }
            }
            else 
            {  // incompatible region
		Matrix ccSolution(newArrayMatrix(0,0,true));
                // _noSTClasses is used as _errorCode here
                pph::getCCSolution(noSTClasses, ccSolution);
                MatrixView view = ccSolution.view();

                sortColumns(view, positions, i);
                treesAtPosition.push_back(boost::shared_ptr< Tree >(buildTree(view, 0, i, positions)));
            }
        }

        trees.push_back(treesAtPosition);

	}
    if (options::verbose) progress.end();
    
    return trees;
}



namespace {


    
    typedef std::vector<Tree *> flatTree_t;



    std::ostream &
    operator<<(std::ostream &os, const std::vector<int> &v)
    {
        std::vector<int>::const_iterator i;
        const char *sep = "";
        for (i = v.begin(); i != v.end(); ++i)
	    { os << sep << *i; sep = " "; }
        return os;
    }



    std::ostream &
    operator<<(std::ostream &os, const std::set<int> &v)
    {
        std::set<int>::const_iterator i;
        const char *sep = "";
        for (i = v.begin(); i != v.end(); ++i)
	    { os << sep << *i; sep = " "; }
        return os;
    }



    std::ostream &
    operator<<(std::ostream &os, const flatTree_t &t)
    {
        flatTree_t::const_iterator i;
        for (i = t.begin(); i != t.end(); ++i)
            os << '[' << **i << "] ";
        return os;
    }



    template<typename V>
    void visitDepthFirst(Tree *t, V &v)
    {
        InnerNode *inner = dynamic_cast<InnerNode *>(t);
        if(inner)
        {
            v.visitInnerNode(*inner);
            visitDepthFirst(inner->getLeftChild(), v);
            visitDepthFirst(inner->getRightChild(), v);
        }
        else
        {
            LeafNode *leaf = dynamic_cast<LeafNode *>(t);
            if(leaf)
            {
                v.visitLeafNode(*leaf);
            }
            else
            {
                assert(!"Unknown type of tree node");
            }
        }
    }



    class FlattenTree : public TreeVisitor
    {
        flatTree_t &result;
        
    public:
        FlattenTree(flatTree_t &result, Tree *root)
            : result(result)
        {
            visitDepthFirst(root, *this);
        }

        virtual ~FlattenTree() {}

        virtual void visitInnerNode(InnerNode &n)
        {
            result.push_back(&n);
        }

        virtual void visitLeafNode(LeafNode &n) {}
    };



    const int WILD_TYPE = 0;
    const int MUTATION_TYPE = 1;

    class ColourTree : public TreeVisitor
    {
        std::vector<int> &result;
        int colour;

    public:
        ColourTree(std::vector<int> &result, int colour, Tree *root)
            : result(result), colour(colour)
        {
            visitDepthFirst(root, *this);
        }

        virtual ~ColourTree() {}

        virtual void visitInnerNode(InnerNode &n) {}

        virtual void visitLeafNode(LeafNode &n)
        {
            for(LeafNode::IDsIterator it = n.leafIDsBegin(); it != n.leafIDsEnd(); ++it)
                result[*it] = colour;
        }
    };



    void colourTreeBasedOnPrevious(std::vector<int> &result, Tree* currentNode, Tree* previousNode)
    {
        InnerNode *previousInner = dynamic_cast<InnerNode *>(previousNode);

        if(previousNode == NULL)
        {
            ColourTree(result, MUTATION_TYPE, currentNode);
        }
        else if(currentNode == previousInner->getLeftChild())
        {
            ColourTree(result, WILD_TYPE, previousInner->getRightChild());
        }
        else
        {
            ColourTree(result, WILD_TYPE, previousNode);
            ColourTree(result, MUTATION_TYPE, currentNode);
        }
    }



    long double scoreColouring(const Classification &classification, const std::vector<int> &col1, const std::vector<int> &col2)
    {
        const size_t TYPE_COUNT = 9;
        int A[TYPE_COUNT] = {0, 0, 0, 0, 0, 0, 0, 0, 0};
        int U[TYPE_COUNT] = {0, 0, 0, 0, 0, 0, 0, 0, 0};
        for(size_t i = 0; i < classification.size(); i += 2)
        {
            int type = 3 * (col1[i] + col1[i + 1]) + (col2[i] + col2[i + 1]);
            if(classification.isAffected(i))
                A[type]++;
            else if(classification.isUnaffected(i))
                U[type]++;
            else
                assert(0); //Unknown. Ignore.
        }

        long double retVal = 0.0;

        for(size_t i = 0; i < TYPE_COUNT; i++)
        {
            //std::cout << "A[" << i << "] = " << A[i] << ", U[" << i << "] = " << U[i] << std::endl;
            retVal += gsl_sf_lnbeta(A[i] + 1, U[i] + 1);
        }
        
        return retVal;
    }



    inline long double
    logspaceAdd (long double logx, long double logy)
    {
        return std::max(logx, logy) + log1p(exp(-fabs(logx - logy)));
    }
}



static long double
scoreTreeInteraction(Tree *tree1, Tree *tree2, const Classification &classification)
{
    flatTree_t t1, t2;
    int noModels = 0;

    FlattenTree(t1, tree1);
    FlattenTree(t2, tree2);

    std::vector<int> mutation1, mutation2;
    mutation1.resize(classification.size(), WILD_TYPE);
    mutation2.resize(classification.size(), WILD_TYPE);

    long double retVal = -std::numeric_limits<double>::infinity();

    Tree *prevIt1 = NULL, *prevIt2 = NULL;

    //Note: begin() + 2 ==> Don't mutate on root or left child of root
    for(flatTree_t::iterator it1 = t1.begin() + 2; it1 != t1.end(); ++it1)
    {
        //std::cerr << "XXX Setting mutation 1 to " << **it1 << std::endl;
        colourTreeBasedOnPrevious(mutation1, *it1, prevIt1);

        //Note: begin() + 2 ==> Don't mutate on root or left child of root
        for(flatTree_t::iterator it2 = t2.begin() + 2; it2 != t2.end(); ++it2)
        {
            //std::cerr << "ZZZ Setting mutation 2 to " << **it2 << std::endl;
            colourTreeBasedOnPrevious(mutation2, *it2, prevIt2);
            noModels++;
            long double localScore = scoreColouring(classification, mutation1, mutation2);
            if(retVal == -std::numeric_limits<double>::infinity())
                retVal = localScore;
            else
                retVal = logspaceAdd(retVal, localScore);

            prevIt2 = *it2;
        }

        prevIt1 = *it1;
    }

    long double nullModel = gsl_sf_lnbeta(classification.noAffected() / 2 + 1, classification.noUnaffected() / 2 + 1);

    return retVal - nullModel - std::log(noModels);
}



static void
interactionScores(treesAtRegion_t &region1,
                  treesAtRegion_t &region2,
                  const Classification &classification,
                  std::ostream &os)
{
    treesAtRegion_t::const_iterator r_itr1, r_itr2;
    TextProgressBar progress;

    if (options::verbose) progress.start();

    for (r_itr1 = region1.begin(); r_itr1 != region1.end(); ++r_itr1)
	{
        const std::vector< boost::shared_ptr< Tree > > &r1Trees = *r_itr1;

	    for (r_itr2 = region2.begin(); r_itr2 != region2.end(); ++r_itr2)
		{
            const std::vector< boost::shared_ptr< Tree > > &r2Trees = *r_itr2;

		    if (options::verbose)
                progress.showProgress(r_itr1 - region1.begin(),
                                      region1.size(),
                                      r_itr2 - region2.begin(),
                                      region2.size());

            long double scoreSum = 0.0;
            int scoreCount = 0;

            std::vector< boost::shared_ptr< Tree > >::const_iterator t_itr1, t_itr2;
            for(t_itr1 = r1Trees.begin(); t_itr1 != r1Trees.end(); ++t_itr1)
            {
                for(t_itr2 = r2Trees.begin(); t_itr2 != r2Trees.end(); ++t_itr2)
                {
                    scoreSum = scoreTreeInteraction(t_itr1->get(), t_itr2->get(), classification);
                    scoreCount++;
                }
            }

            os << (scoreSum / scoreCount) << ' ';
		}

	    os << std::endl;
	}

    if (options::verbose) progress.end();
}



namespace po = boost::program_options;
void
printUsage(const char *progname, 
           po::options_description &desc, 
           std::ostream &os = std::cerr)
{
    os << "Usage: " << progname << " [options] positions genotypes"
       << std::endl
       << desc << std::endl;
    
}



int
main(int argc, char *argv[])
{
    using namespace TextManip;

    po::options_description desc("Allowed options");
    po::variables_map vm;
    desc.add_options()
        ("help,h", "Produce help message.")
        ("verbose,v", "Turns on verbose output.")

        ("reg1pos", po::value<std::string>(&options::region1positionsFileName),
         "Positions' file for first region.")
        ("reg1gen", po::value<std::string>(&options::region1haplotypesFileName),
         "Genotypes' file for first region.")
        ("reg2pos", po::value<std::string>(&options::region2positionsFileName),
         "Positions' file for second region.")
        ("reg2gen", po::value<std::string>(&options::region2haplotypesFileName),
         "Genotypes' file for second region.")

        ("scores,o",
         po::value<std::string>(&options::scoresFileName)
         ->default_value("scores.txt"),
         "File where scores will be written.")
        ("minmarkers,m",
         po::value<int>(&options::minMarkers)->default_value(10),
         "Minimum number of markers used (default: 10).")
        ("deslen,x",
         po::value<double>(&options::maxlen)->default_value(-1),
         "Maximum length of interval (default: whole range).")

        ("unphased,u",
         po::value<bool>(&options::unphased)->default_value(true),
         "Input is unphased (default true).")
        ;
    po::positional_options_description p;
    p.add("reg1pos", 1);
    p.add("reg1gen", 1);
    p.add("reg2pos", 1);
    p.add("reg2gen", 1);

    try {
        po::store(po::command_line_parser(argc, argv)
                  .options(desc).positional(p).run(), 
                  vm, false);
        po::notify(vm);
    } catch (std::exception &e) {
        std::cerr << "Error parsing options:" << e.what() << std::endl;
        printUsage(argv[0], desc);
        return 2;
    }

    if (vm.count("help")) {
        printUsage(argv[0], desc, std::cout);
        return 0;
    }
    options::verbose = vm.count("verbose") > 0;

    if (options::region1positionsFileName == "")
	{
  	    std::cerr << bold << red << "Error: " << reset
                  << bold
                  << "No position file for first region specified."
                  << reset
                  << std::endl << std::endl;
	    printUsage(argv[0], desc);
	    return 2;
	}
    if (options::region1haplotypesFileName == "")
	{
  	    std::cerr << bold << red << "Error: " << reset
                  << bold
                  << "No haplotype file for first region specified."
                  << reset
                  << std::endl << std::endl;
	    printUsage(argv[0], desc);
	    return 2;
	}

    if (options::region2positionsFileName == "")
	{
  	    std::cerr << bold << red << "Error: " << reset
                  << bold
                  << "No position file for second region specified."
                  << reset
                  << std::endl << std::endl;
	    printUsage(argv[0], desc);
	    return 2;
	}
    if (options::region2haplotypesFileName == "")
	{
  	    std::cerr << bold << red << "Error: " << reset
                  << bold
                  << "No haplotype file for second region specified."
                  << reset
                  << std::endl << std::endl;
	    printUsage(argv[0], desc);
	    return 2;
	}

    std::ifstream region1positionsFile(options::region1positionsFileName.c_str());
    if (!region1positionsFile)
	{
	    std::cerr << "Error opening " 
                  << options::region1positionsFileName
                  << std::endl;
	    return 2;
	}
    std::ifstream region1haplotypesFile(options::region1haplotypesFileName.c_str());
    if (!region1haplotypesFile)
	{
	    std::cerr << "Error opening " 
                  << options::region1haplotypesFileName
                  << std::endl;
	    return 2;
	}

    std::ifstream region2positionsFile(options::region2positionsFileName.c_str());
    if (!region2positionsFile)
	{
	    std::cerr << "Error opening " 
                  << options::region2positionsFileName
                  << std::endl;
	    return 2;
	}
    std::ifstream region2haplotypesFile(options::region2haplotypesFileName.c_str());
    if (!region2haplotypesFile)
	{
	    std::cerr << "Error opening "
                  << options::region2haplotypesFileName
                  << std::endl;
	    return 2;
	}


    // reading and validating input
    std::vector<int> region1positions, region2positions;
    Matrix region1haplotypes(newArrayMatrix(0,0));
    Matrix region2haplotypes(newArrayMatrix(0,0));
    std::vector<int> classes;

    try {
        region1positions = readPositions(region1positionsFile);
    } catch (UnsortedPositions&) {
        std::cerr << "The positions in " 
                  <<  options::region1positionsFileName
                  << " must be sorted"
                  << std::endl;
        return 2;
    }
    region1positionsFile.close();

    try {
        region2positions = readPositions(region2positionsFile);
    } catch (UnsortedPositions&) {
        std::cerr << "The positions in " 
                  << options::region2positionsFileName
                  << " must be sorted"
                  << std::endl;
        return 2;
    }
    region2positionsFile.close();

    try {
        if (options::unphased)
            readGenotypes(region1haplotypesFile, region1haplotypes, classes);
        else
            readHaplotypes(region1haplotypesFile, region1haplotypes, classes);
    } catch (ParseError&) {
        std::cerr << "Error reading " 
                  << options::region1haplotypesFileName
                  << std::endl;
        return 2;
    }
    region1haplotypesFile.close();

    try {
        if (options::unphased)
            readGenotypes(region2haplotypesFile, region2haplotypes, classes);
        else
            readHaplotypes(region2haplotypesFile, region2haplotypes, classes);
    } catch (ParseError&) {
        std::cerr << "Error reading " 
                  << options::region2haplotypesFileName
                  << std::endl;
        return 2;
    }
    region2haplotypesFile.close();

    if (region1haplotypes.noRows() != region2haplotypes.noRows())
	{
	    std::cerr << "Both regions must have the same number of haplotypes"
                  << std::endl;
	    return 2;
	}


    // main computation part...
    std::ofstream scoresFile(options::scoresFileName.c_str());
    if (!scoresFile)
	{
	    std::cerr << "Error opening file " 
                  << options::scoresFileName
                  << std::endl;
	    return 2;
	}
    
    if (options::verbose)
        std::cout << bold
                  << "Running " << PACKAGE_NAME << " v" << PACKAGE_VERSION
                  << std::endl
                  << "Mapping interactions."
                  << reset
                  << std::endl << std::endl;


    if (options::verbose) std::cout << "Building marginal trees." << std::endl;
    Classification classification(classes.begin(), classes.end());
    treesAtRegion_t region1 = buildTrees(region1haplotypes, region1positions);
    treesAtRegion_t region2 = buildTrees(region2haplotypes, region2positions);

    // return NaNs rather than crash.
    gsl_set_error_handler_off();

    if (options::verbose)
        std::cout << "Scoring interactions." << std::endl;
    interactionScores(region1, region2, classification, scoresFile);

    region1.clear();
    region2.clear();
}
