/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "config.hh"
#include "io.hh"
#include <snpfile/matrix.hh>
using namespace BiRC::SNPFile;

#include <algorithm>
#include <cmath>
#include <iomanip>
#include <fstream>
#include <string>

namespace TextManip {
  const char *clearLine  = "[100D[K";
  const char *bold       = "[1m";
  const char *underscore = "[4m";
  const char *reset      = "[0m";
  const char *green      = "[32m";
  const char *red        = "[31m";
  const char *blue       = "[34m";
}


namespace {
    template <typename Itr>
    bool isSorted(Itr begin, Itr end)
    {
	if (begin+1 >= end) return true;
	for (Itr i = begin; i < end-1; ++i)
	    if (*i > *(i+1)) return false;
	return true;
    }
}

std::vector<int> 
readPositions(std::istream &is)
{
    std::vector<int> result;
    double floatPos;
    int    intPos;

    while (is >> floatPos) {
	intPos = static_cast<int>(floatPos);
	if (intPos != floatPos)
	    std::cerr << "Warning: support for floating point positions is "
		      << "deprecated, rounding " << floatPos << " to "
		      << intPos << '.' << std::endl;
	result.push_back(intPos);
    }
    if (!isSorted(result.begin(), result.end())) throw UnsortedPositions();

    return result;
}

std::vector<int> 
readPositions(const char *fname)
{
    std::ifstream file(fname);
    if (!file) throw UnableToOpenFile();
    return readPositions(file);
}


namespace {
    void split(const std::string& str, std::vector<std::string>& tokens)
    {
	const std::string& delimiters = " \t";
	std::string::size_type lastPos = str.find_first_not_of(delimiters, 0);
	std::string::size_type pos = str.find_first_of(delimiters, lastPos);

	while (std::string::npos != pos || std::string::npos != lastPos)
	    {
		tokens.push_back(str.substr(lastPos, pos - lastPos));
		lastPos = str.find_first_not_of(delimiters, pos);
		pos = str.find_first_of(delimiters, lastPos);
	    }
    }

    // FIXME: should check that the tokens are actually integers
    inline int string2int(std::string &s) 
    { 
		if (s == "?")  return -9;
		else           return atoi(s.c_str()); 
    }
    inline double string2double(std::string &s) { return atof(s.c_str()); }
}

void readHaplotypes(std::istream &is, 
		    Matrix &matrix, 
		    std::vector<int> &classes)
{
    std::vector< std::vector<int> > haplotypes;
    while (is)
	{
	    std::string line; 
	    std::vector<std::string> tokens;

	    getline(is, line);
	    split(line, tokens);

	    std::vector<int> haplotype(tokens.size());
	    std::transform(tokens.begin(), tokens.end(), 
			   haplotype.begin(), string2int);

	    if (haplotype.size() == 0) continue; // empty line

	    if (haplotypes.size() > 0)
		if (haplotypes[0].size() != haplotype.size())
		    throw ParseError();
	    haplotypes.push_back(haplotype);
	}

    if (haplotypes.empty()) throw ParseError();

    int noHaplotypes = (int)haplotypes.size();
    int noMarkers    = (int)haplotypes[0].size() - 1; // leftmost
						      // column is the
						      // classification

    matrix.resize(noHaplotypes, noMarkers);
    classes.resize(noHaplotypes / 2);

    for (int i = 0; i < noHaplotypes; ++i)
	{
	    classes[i / 2] = haplotypes[i][0]; //Yes, each position
                                           //will be update twice.
                                           //That's ok, it should
                                           //be the same value...
	    for (int j = 0; j < noMarkers; ++j)
		matrix(i,j) = haplotypes[i][j+1];
	}
    matrix.setPhased(true);
}

void readHaplotypes(const char *fname,
		    Matrix &matrix, 
		    std::vector<int> &classes)
{
    std::ifstream file(fname);
    if (!file) throw UnableToOpenFile();
    readHaplotypes(file, matrix, classes);
}



void readHaplotypesQT(std::istream &is, 
		      Matrix &matrix, 
		      std::vector<double> &values)
{
    std::vector< std::vector<int> > haplotypes;
    bool add_now = true;
    while (is)
	{
	    std::string line; 
	    std::vector<std::string> tokens;

	    getline(is, line);
	    split(line, tokens);

	    std::vector<int> haplotype(tokens.size());

	    if (haplotype.size() == 0) continue; // empty line

	    std::transform(tokens.begin()+1, tokens.end(), 
			   haplotype.begin(), string2int);
	    if(add_now){
		values.push_back(string2double(tokens[0]));
	    }
	    add_now = not add_now;
	    if (haplotypes.size() > 0)
		if (haplotypes[0].size() != haplotype.size())
		    throw ParseError();
	    haplotypes.push_back(haplotype);
	}

    if (haplotypes.empty()) throw ParseError();

    int noHaplotypes = (int)haplotypes.size();
    int noMarkers    = (int)haplotypes[0].size() - 1; // leftmost
						      // column is the
						      // classification

    assert((int)values.size() == noHaplotypes/2);

    matrix.resize(noHaplotypes, noMarkers);

    for (int i = 0; i < noHaplotypes; ++i)
	{
	    for (int j = 0; j < noMarkers; ++j)
		matrix(i,j) = haplotypes[i][j];
	}

    matrix.setPhased(true);
}

void readHaplotypesQT(const char *fname,
		      Matrix &matrix,
		      std::vector<double> &values)
{
    std::ifstream file(fname);
    if (!file) throw UnableToOpenFile();
    readHaplotypesQT(file, matrix, values);
}



// FIXME: get rid of all this code redundancy caused by case/control |
// qt * haplotypes | genotypes

void readGenotypes(std::istream &is, 
		   Matrix &matrix, 
		   std::vector<int> &classes)
{
    std::vector< std::vector<int> > genotypes;
    while (is)
	{
	    std::string line; 
	    std::vector<std::string> tokens;

	    getline(is, line);
	    split(line, tokens);

	    std::vector<int> genotype(tokens.size());
	    std::transform(tokens.begin(), tokens.end(), 
			   genotype.begin(), string2int);

	    if (genotype.size() == 0) continue; // empty line

	    if (genotypes.size() > 0)
		if (genotypes[0].size() != genotype.size())
		    throw ParseError();
	    genotypes.push_back(genotype);
	}

    if (genotypes.empty()) throw ParseError();

    int noGenotypes = (int)genotypes.size();
    int noMarkers   = (int)genotypes[0].size() - 1; // leftmost
                            // column is the
						    // classification

    matrix.resize(noGenotypes, noMarkers);
    classes.resize(noGenotypes);

    for (int i = 0; i < noGenotypes; ++i)
	{
	    classes[i] = genotypes[i][0];
	    for (int j = 0; j < noMarkers; ++j)
		matrix(i,j) = genotypes[i][j+1];
	}
    matrix.setPhased(false);
}

void readGenotypes(const char *fname,
		   Matrix &matrix, 
		   std::vector<int> &classes)
{
    std::ifstream file(fname);
    if (!file) throw UnableToOpenFile();
    readGenotypes(file, matrix, classes);
}


void readGenotypesQT(std::istream &is, 
		     Matrix &matrix,
		     std::vector<double> &values)
{
    std::vector< std::vector<int> > genotypes;
    std::vector< double > tmp_values;
    while (is)
	{
	    std::string line; 
	    std::vector<std::string> tokens;

	    getline(is, line);
	    split(line, tokens);
	    
	    std::vector<int> genotype(tokens.size());
	    
	    if (genotype.size() == 0) continue; // empty line
	    
	    std::transform(tokens.begin()+1, tokens.end(), 
			   genotype.begin(), string2int);
	    tmp_values.push_back(string2double(tokens[0]));
	    if (genotypes.size() > 0)
		if (genotypes[0].size() != genotype.size())
		    throw ParseError();
	    genotypes.push_back(genotype);
	}

    if (genotypes.empty()) throw ParseError();

    int noGenotypes = (int)genotypes.size();
    int noMarkers   = (int)genotypes[0].size() - 1; // leftmost
						    // column is the
						    // classification

    matrix.resize(noGenotypes, noMarkers);

    for (int i = 0; i < noGenotypes; ++i)
	{
	    values[2*i] = values[2*i+1] = tmp_values[i];
	    for (int j = 0; j < noMarkers; ++j)
		{
		    matrix(i,j) = genotypes[i][j];
		}
	}
    matrix.setPhased(false);
}

void readGenotypesQT(const char *fname,
		      Matrix &matrix, 
		     std::vector<double> &values)
{
    std::ifstream file(fname);
    if (!file) throw UnableToOpenFile();
    readGenotypesQT(file, matrix, values);
}


void
printScores(std::ostream &os, const std::vector<long double> &scores)
{
    std::vector<long double>::const_iterator i;
    for (i = scores.begin(); i != scores.end(); ++i)
	os << *i  << ' ';
    os << std::endl;
}


void
TextProgressBar::start() const
{
    i_os << "Progress: [";
    for (int i = 0; i < i_progressWidth; ++i) i_os << ' ';
    i_os << "] (  0%)" << std::flush;
}

void
TextProgressBar::showProgress(int progress, int maxProgress) const
{
    using namespace TextManip;
    int noBlocks = i_progressWidth*(progress+1) / maxProgress;
    int noEmptyBlocks = i_progressWidth - noBlocks;

    i_os << clearLine << "Progress: [";
    for (int j = 0; j < noBlocks; ++j)      i_os << '=';
    for (int j = 0; j < noEmptyBlocks; ++j) i_os << ' ';
    i_os << "] ("
	 << std::right
	 << std::setw(3)
	 << std::setprecision(2)
	 << (100*(progress+1)/maxProgress) << "%)"
	 << std::flush;
}

void 
TextProgressBar::showProgress(int progress, int maxProgress,
			      int subProgress, int maxSubProgress) const
{
    using namespace TextManip;
    int subProgressWidth = static_cast<int>(strlen("|     ]")) - 2;
    int progressWidth = i_progressWidth - (subProgressWidth + 2);
    
    int noBlocks = progressWidth*(progress+1) / maxProgress;
    int noEmptyBlocks = progressWidth - noBlocks;
    int noSubBlocks = subProgressWidth*(subProgress+1) / maxSubProgress;
    int noEmptySubBlocks = subProgressWidth - noSubBlocks;

    i_os << clearLine << "Progress: [";
    for (int j = 0; j < noBlocks; ++j)         i_os << '=';
    for (int j = 0; j < noEmptyBlocks; ++j)    i_os << ' ';
    i_os << '|';
    for (int j = 0; j < noSubBlocks; ++j)      i_os << '.';
    for (int j = 0; j < noEmptySubBlocks; ++j) i_os << ' ';
    i_os << "] ("
	 << std::right
	 << std::setw(3)
	 << std::setprecision(2)
	 << (100*(progress+1)/maxProgress) << "%)"
	 << std::flush;
}


void
TextProgressBar::end() const
{
    using namespace TextManip;
    i_os << clearLine << "Progress: [";
    for (int i = 0; i < i_progressWidth; ++i) std::cout << '=';
    i_os << "] (100%)" << std::endl << std::flush;
}
