/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef IO_HH_INCLUDED
#define IO_HH_INCLUDED

/** \file
 * \brief Functions for reading in data and printing results
 * and user feedback.
 */

#include <iostream>
#include <vector>
#include <stdexcept>

namespace BiRC { namespace SNPFile { 
	class ImmutableMatrix;
	class Matrix;
	class MatrixView;
    }
}

/// Text codes for manipulating text output.
namespace TextManip {
    /// Clear the current line.
    extern const char *clearLine;
    /// Write the following text in boldface.
    extern const char *bold;
    /// Underscore the following text.
    extern const char *underscore;
    /// Reset all colouring and font-changes.
    extern const char *reset;
    /// Write the following text in green.
    extern const char *green;
    /// Write the following text in red.
    extern const char *red;
    /// Write the following text in blue.
    extern const char *blue;
}

/// Exception thrown if positions read in are not in increasing order.
struct UnsortedPositions : std::logic_error {
    UnsortedPositions() : std::logic_error("Positions must be sorted.") {}
};

/// Exception thrown if loading haplotypes failed.
struct ParseError : std::logic_error {
    ParseError() : std::logic_error("Error parsing haplotype data.") {}
};

/// Exception thrown if a requested file could not be opened.
struct UnableToOpenFile : std::logic_error {
    UnableToOpenFile() : std::logic_error("Unable to open file.") {}
};

/**
 * \brief Load positions from a stream.
 *
 * Read in a sequence of positions from a stream.  The positions
 * should be in increasing order, otherwise the UnsortedPositions
 * exception will be thrown.
 *
 * \param is		Input stream to read the positions from.
 * \returns		Vector of positions.  Guaranteed sorted.
 * \throws		UnsortedPositions
 */
std::vector<int> readPositions(std::istream &is);

/**
 * \brief Load positions from a named file.
 *
 * Read in a sequence of positions from a stream.  The positions
 * should be in increasing order, otherwise the UnsortedPositions
 * exception will be thrown.
 *
 * \param fname		File name for the file to read from.
 * \returns		Vector of positions.  Guaranteed sorted.
 * \throws		UnsortedPositions
 * \throws		UnableToOpenFile
 */
std::vector<int> readPositions(const char *fname);


/**
 * \brief Load a sequence of haplotypes, with case/control status,
 * from a stream.
 *
 * Read in a sequence of haplotypes from a stream.  The first column
 * is considered a case/control status indicator, where 1 is
 * interpreted as case and 0 as control.
 *
 * \param is		Input stream to read the haplotypes from.
 * \param haplotypes	Matrix that will be filled in with the read haplotypes.
 * \param classes	Vector that will be filled in with case/control data.
 *                      The vector will be ordered to match the rows in 
 *                      \a haplotypes.
 *
 * \throws		ParseError
 */
void readHaplotypes(std::istream &is, 
		    BiRC::SNPFile::Matrix &haplotypes, 
		    std::vector<int> &classes);

/**
 * \brief Load a sequence of haplotypes, with case/control status,
 * from a file.
 *
 * Read in a sequence of haplotypes from a file.  The first column
 * is considered a case/control status indicator, where 1 is
 * interpreted as case and 0 as control.
 *
 * \param fname		File name for the input file.
 * \param haplotypes	Matrix that will be filled in with the read haplotypes.
 * \param classes	Vector that will be filled in with case/control data.
 *                      The vector will be ordered to match the rows in 
 *                      \a haplotypes.
 *
 * \throws		ParseError
 */
void readHaplotypes(const char *fname,
		    BiRC::SNPFile::Matrix &haplotypes, 
		    std::vector<int> &classes);


/**
 * \brief Load a sequence of haplotypes, with a quantitative trait,
 * from a stream.
 *
 * Read in a sequence of haplotypes from a stream.  The first column
 * is considered a quantitative trait value.
 *
 * \param is		Input stream to read the haplotypes from.
 * \param haplotypes	Matrix that will be filled in with the read haplotypes.
 * \param values	Vector that will be filled in with the trait.
 *                      The vector will be ordered to match the rows in 
 *                      \a haplotypes.
 *
 * \throws		ParseError
 */
void readHaplotypesQT(std::istream &is, 
		      BiRC::SNPFile::Matrix &haplotypes, 
		      std::vector<double> &values);

/**
 * \brief Load a sequence of haplotypes, with a quantitative trait,
 * from a file.
 *
 * Read in a sequence of haplotypes from a file.  The first column
 * is considered a quantitative trait value.
 *
 * \param fname		File name of input file.
 * \param haplotypes	Matrix that will be filled in with the read haplotypes.
 * \param values	Vector that will be filled in with the trait.
 *                      The vector will be ordered to match the rows in 
 *                      \a haplotypes.
 *
 * \throws		ParseError
 */
void readHaplotypesQT(const char *fname,
		      BiRC::SNPFile::Matrix &haplotypes, 
		      std::vector<double> &values);

/**
 * \brief Load a sequence of unphased genotypes, with case/control
 * status, from a stream.
 *
 * Read in a sequence of genotypes from a stream.  The first column
 * is considered a case/control status indicator, where 1 is
 * interpreted as case and 0 as control.
 *
 * \param is		Input stream to read the haplotypes from.
 * \param genotypes	Matrix that will be filled in with the read genotypes.
 * \param classes	Vector that will be filled in with case/control data.
 *                      The vector will be ordered to match the rows in 
 *                      \a genotypes.
 *
 * \throws		ParseError
 */
void readGenotypes(std::istream &is, 
		   BiRC::SNPFile::Matrix &genotypes, 
		   std::vector<int> &classes);

/**
 * \brief Load a sequence of genotypes, with case/control status,
 * from a file.
 *
 * Read in a sequence of genotypes from a file.  The first column
 * is considered a case/control status indicator, where 1 is
 * interpreted as case and 0 as control.
 *
 * \param fname		File name for the input file.
 * \param genotypes	Matrix that will be filled in with the read genotypes.
 * \param classes	Vector that will be filled in with case/control data.
 *                      The vector will be ordered to match the rows in 
 *                      \a genotypes.
 *
 * \throws		ParseError
 */
void readGenotypes(const char *fname,
		   BiRC::SNPFile::Matrix &genotypes, 
		   std::vector<int> &classes);


/**
 * \brief Load a sequence of genotypes, with a quantitative trait,
 * from a stream.
 *
 * Read in a sequence of genotypes from a stream.  The first column
 * is considered a quantitative trait value.
 *
 * \param is		Input stream to read the genotypes from.
 * \param genotypes	Matrix that will be filled in with the read genotypes.
 * \param values	Vector that will be filled in with the trait.
 *                      The vector will be ordered to match the rows in 
 *                      \a genotypes.
 *
 * \throws		ParseError
 */
void readGenotypesQT(std::istream &is, 
		     BiRC::SNPFile::Matrix &genotypes, 
		     std::vector<double> &values);

/**
 * \brief Load a sequence of genotypes, with a quantitative trait,
 * from a file.
 *
 * Read in a sequence of genotypes from a file.  The first column
 * is considered a quantitative trait value.
 *
 * \param fname		File name of input file.
 * \param genotypes	Matrix that will be filled in with the read genotypes.
 * \param values	Vector that will be filled in with the trait.
 *                      The vector will be ordered to match the rows in 
 *                      \a genotypes.
 *
 * \throws		ParseError
 */
void readGenotypesQT(const char *fname,
		     BiRC::SNPFile::Matrix &genotypes, 
		     std::vector<double> &values);


/**
 * \brief Print scores to a stream.
 *
 * \param os		Stream to print to.
 * \param scores	Vector of scores to print.
 */
void printScores(std::ostream &os, 
		 const std::vector<long double> &scores);

/// Textual progress bar for verbose output.
class TextProgressBar {
    /// Width in characters for the progress bar.
    int i_progressWidth;
    /// Stream the progress bar is shown on.
    std::ostream &i_os;
    
public:
    /**
     * \brief Create progress bar with a given width and on a given stream.
     *
     * \param progressWidth	Character width for the bar.  Default 60.
     * \param os		Output stream. Default std::cout.
     */
    TextProgressBar(int progressWidth = 60, std::ostream &os = std::cout) 
	: i_progressWidth(progressWidth), i_os(os)
    {}
    
    /// Start the process this bar should show progress for.
    void start() const;
    /// Complete the process this bar shows progress for.
    void end() const;

    /**
     * \brief Display the current progress.
     *
     * \param progress		The current progress.
     * \param maxProgress	The progress corresponding to 100%.
     */
    void showProgress(int progress, int maxProgress) const;
 
    /**
     * \brief Display the current progress.
     *
     * The progress includes the progress of a sub-process.
     *
     * \param progress		The current (full) progress.
     * \param maxProgress	The progress corresponding to (the full) 100%.
     * \param subProgress	The current progress of the sub-task.
     * \param maxSubProgress	The progress corresponding to 100% of the
     *                          sub-task.
     */
    void showProgress(int progress, int maxProgress,
		      int subProgress, int maxSubProgress) const;
};

#endif
