/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */



#include "matrix_utils.hh"
using namespace BiRC::SNPFile;


bool rawRowLess(const MatrixView &m, int i, int j)
{
    const int n = m.noCols();
    for (int k = 0; k < n; ++k)
    {
        int x = m(i, k);
        int y = m(j, k);
        if (x > y) return true;
        if (x < y) return false;
    }
    return false;
}



namespace {
    struct row_cmp : std::binary_function<bool, int, int>
    {
        MatrixView &view;
        row_cmp(MatrixView &view) : view(view) {}
        bool operator()(int i, int j) { return rawRowLess(view, i,j); }
    };

    struct col_cmp : std::binary_function<bool,int,int> {
        MatrixView &view;
        const std::vector<int> &pos;
        int marker;
        col_cmp(MatrixView &view, const std::vector<int> &pos, int marker) 
            : view(view), pos(pos), marker(marker) {}
        bool operator()(int i, int j) 
        { 
            return fabs(pos[i]-pos[marker]) < fabs(pos[j]-pos[marker]); 
        }
    };
}



void sortRows(MatrixView &m)
{
    std::vector<int> permutation;
    permutation.resize(m.noRows());
    for(size_t i = 0; i < permutation.size(); i++)
        permutation[i] = i;

    std::sort(permutation.begin(), permutation.end(), row_cmp(m));

    std::vector<int> &prevRowIndices = m.getRowMappings();
    std::vector<int> newRowIndices;

    newRowIndices.resize(m.noRows());

    for(size_t i = 0; i < permutation.size(); i++)
        newRowIndices[i] = prevRowIndices[permutation[i]];

    prevRowIndices = newRowIndices;
}



// FIXME: this stuff doesn't work if there is missing data...

int splitOnColumn(MatrixView &m, int j)
{
    std::vector<int> &rowIndices = m.getRowMappings();

    int firstOne = m.noRows();
    while (firstOne > 1 && m(firstOne - 1, j) == 1) 
        --firstOne;

    for (int i = 0; i < firstOne; ++i)
        if (m(i, j) == 1)
        {
            std::swap(rowIndices[i], rowIndices[--firstOne]);
            while (firstOne > 1 && m(firstOne-1, j) == 1) 
                --firstOne;
            assert(firstOne == m.noCols() || m(firstOne, j) == 1);
            assert(firstOne == 0 || m(firstOne-1, j) == 0);
        }

#if 0
    for (int i = 0; i < firstOne; ++i)
        assert(m(i, j) == 0);
    for (int i = firstOne; i < noRows(); ++i)
        assert(m(i, j) == 1);
#endif

    return firstOne;
}
