/*
 * SNPFile -- SNP data management for association mapping.
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus, Denmark
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */



#ifndef MATRIX_UTILS_HH_INCLUDED
#define MATRIX_UTILS_HH_INCLUDED



#include "config.hh"
#include <snpfile/matrix.hh>
#include <snpfile/array_matrix.hh>
#include <vector>
#include <cmath>




/**
 * \brief Predicate to test if one row is
 * lexicographically less than another.
 *
 * \param	i		The first row index.
 * \param	j		The second row index.
 *
 * \returns			True if row i is less than j;
 *                          false otherwise.
 */
bool rawRowLess(const BiRC::SNPFile::MatrixView &m, int i, int j);



/// Sort the rows in this view according to the ordering
/// given by rawRowLess().
void sortRows(BiRC::SNPFile::MatrixView &m);



/**
 * \brief Split the rows in two, conditioning on a column,
 * such that the first part of the matrix all have 0 in
 * that column, and the remaining all have 1.
 *
 * \param j			Index of the reference column.
 * \returns			The index of the row containing
 *                          the first 1 in \a j.
 */
int splitOnColumn(BiRC::SNPFile::MatrixView &m, int j);



/// Class used in sortCols()
template <typename PosType>
struct col_cmp : std::binary_function<bool,int,int>
{
    BiRC::SNPFile::MatrixView &view;
    const std::vector<PosType> &pos;
    int marker;

    col_cmp(BiRC::SNPFile::MatrixView &view,
            const std::vector<PosType> &pos, 
            int marker) 
        : view(view), pos(pos), marker(marker) 
    {}

    bool operator()(int i, int j) 
    { 
        return std::fabs(pos[i]-pos[marker])
            < std::fabs(pos[j]-pos[marker]); 
    }
};



/**
 * Sort the columns of this view according to nearness to
 * \a marker.
 *
 * The columns are sorted such that markers closer to \a
 * marker will be in the first part of the matrix, and
 * markers further away later in the matrix.
 *
 * \param pos		Vector of marker positions.
 * \param marker		The reference marker positions
 *                          are compared with.
 */
template <typename PosType>
inline void
sortColumns(BiRC::SNPFile::MatrixView &m,
	    const std::vector<PosType> &pos, int marker)
{
    std::vector<int> &colIndices = m.getColumnMappings();
    std::sort(colIndices.begin(),
              colIndices.end(),
              col_cmp<PosType>(m, pos, marker));
}



#endif
