/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2007 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * PPH: Copyright (C) 2007 by Zhihong Ding <zhding@ucdavis.edu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "pph.hh"
#include <cmath>
#include <snpfile/matrix.hh>
using namespace BiRC::SNPFile;

#include "pph_sort_util.hh"



#define FREE 0		// free link or fix link
#define FIX 1
#define TREE 0		// tree edge or shadow edge
#define SHADOW 1
#define TCONN 0		// T connector or H connector
#define HCONN 1
#define TREEROOT 0	// the column number of the root of
					// the shadow tree is 0
#define OEL 0		// OldEntryList
#define COOEL 1		// CopyofOldEntryList
#define NEL 2		// NewEntryList
#define CL 3		// CheckList

/// For ENT phasing method
#define F_MARKED 1
#define F_REMOVED 2
#define RAND_SOL 1
#define MIN_SOL 2
#define SAVE_NEW 1
#define SAVE_OVER 2
#define VERBOSE 0  // control the amount of output
#define EXP 0


namespace {

    class CConnector
    {
    public:
        int col;			// column number
        int edgetype;		// TREE or SHADOW
        int conntype;		// TCONN or HCONN
        int linktype;		// FREE or FIX
        CConnector *linkto;	// the connector that this connector links to

        /// initialize an object of class CConnector
        void init(int thiscol, int thisedge, int thisconn, 
                  int thislinktype, CConnector *thislinkto);
    };

    /// class CMatrix is a data structure used to represent a matrix
    class CMatrix
    {
    public:
        int *col; // all entries in one row of a matrix
    };
    
    /// for ENT phasing method
    struct Gen;
    struct CComp{  // connected component
        CComp(int i) {id = i; recomb = false;}
        int id, num_sol;
        int size;
        bool recomb;
        double cur_cov;
	
        std::vector<Gen*> member;

        Gen ** node;
        char ** min_cf1;
        char ** min_cf2;
    };

    struct Gen{   // genotype
        Gen();

        ~Gen() {}
	
        CComp * cc;
        int id;
        int h_count[2];
	
        int inherit[2];
        int sav_inherit[2];
        bool inherit_set[2];
    
        Gen * pp[2];  // point to parents
        std::vector<Gen*> child;
        char * seq; // genotype sequence
	
        char * t[2], * h[2], * p[2], * h_init[2];
    
        int flag; // flag for ordering connected components (F_MARKED/F_REMOVED)
        bool old_saved, orphan;
    };
    struct Hap{ int id, coverage; };


    /// Test if the region from column _colStart to column _colEnd of the
    /// genotype matrix _haplotypes_ has PPH solution
    bool hasPPHSolution(const Matrix &haplotypes, int colStart, int colEnd);
    
    void getOnePPHSln (int solIdx);  // get PPH solution _solIdx_
                                     // for the set of full genotypes
	void imputeMissingValue();  // phase the set of genotypes with missing value

	/// main function to solve PPH problem
    int pph_main(const MatrixView &haplotypes, int colStart, int colEnd);

    /// Initialization of connectors, lists, arrays used in the main function
    void pph_init (void);
	/// build initial tree from matrix _mymat_
    int initialtree(void);
	/// create OldEntryList, NewEntryList, etc. for one row
    void createentrylist(int currentrow); 

    int FirstPath (int currentrow); // an implementation of Procedure FirstPath
    int SecondPath (int currentrow); // an implementation of Procedure SecondPath
    int DirectSecondPath (int colTEp,int Ci);	// an implementation
    // of Procedure DirectSecondPath
    void FixTree (int currentrow);		// an implementation
    // of Procedure FixTree
    int NewEntries (int currentrow);		// an implementation
    // of Procedure NewEntries

    /// find next 2 entry left to startcol in the specified entry list
    int getnextentry (int whichlist, int startcol);	
	/// find the class root of the specified edge
    CConnector *getclassroot (int thiscol, int thisedge);
	/// find the parent connector of this connector
	CConnector *getParentConn (CConnector *thisConn);	

    /// retrieve the region from column _colStart to column _colEnd of matrix
    /// _haplotypes and store the region in matrix _inputmat
    int retrievematrix(const MatrixView &haplotypes, int colStart, int colEnd);	

    /*
     * Description: Re-arrange the input matrix S to get an ordered
     * matrix - mymat Columns of S are sorted by their leaf-counts
     * decreasingly. A leaf-count for each column C is defined as
     * follows: each 2 (or 1) in column C contributes a count of 1 (or
     * 2) to the leaf-count for column C. Rows of S are sorted by the
     * position of rightmost 1 entry in each row decreasingly, with
     * the first row containing the rightmost 1 entry in S.
     */
    int rearrangematrix(void);

    /// Count the number of classes in the shadow tree
    int countClasses(void);	
	// release memory according to different errorcode
    void releasemem (int errorcode);	
    
    /// For ENT phasing method
    void main_ENT(void);  // main function for ENT
    void read_genotypes(void);
    void init_ENT(void);
    bool fill_template(Gen * g);

    void drop_inheritence(CComp * c);
    void outputENTsol(void);  //_ output haplotypes (ENT solution) 
    double mylog(int cov);
    int hap_string2int(char * hap , int start);
    int hap_string2int(char * p , char * c , int start);
    int compute_free(int pos);
    int compute_locked(int pos);

    int add_hap(int id);
    int del_hap(int id);
    void remove_hap(int i);
    double get_coverage();
    char compliment(char c);

    int traverse(CComp * c, int i, int seq_start, double cur_cov, double * min_cov, int mode);
    CComp* get_cc(Gen * g, CComp * c);
    void merge_cc(CComp * c1, CComp * c2);
    void cc_swap(int cc, int a, int b);
    void cc_save(CComp * c, int start, int mode);
    bool cc_save_min(CComp * c, int i, int start, double new_cov, double * cur_cov);
    void insert_sort_cc(CComp * c, CComp ** arr, int size);
    void specialcase(void);  // phase every genotype and missing value randomly when
    // every genotype in the input file contains missing data

    
    CConnector *TTConn;			// T connectors of tree edges in the shadow tree
    CConnector *THConn;			// H connectors of tree edges
    CConnector *STConn;			// T connectors of shadow edges
    CConnector *SHConn;			// H connectors of shadow edges
    CConnector ConnTreeRoot;	// root of the whole shadow tree
    CConnector ConnNoWhere;		// a connector for program purpose

	int noGenotype, noGenotypeMV; // number of all genotypes and
    // number of genotypes with missing values, noGenotype = rownumber + noGenotypeMV
	int rownumber, columnnumber;  // number of full genotypes and their lengths
	CMatrix *inputmat;  // matrix containing full genotypes, index starts from 1
	CMatrix *genMatMV;  // matrix containing genotypes with missing values
    // index of rows and columns in genMatMV starts from 0
	std::vector<int> fullGenMap;  // genotype(i) in the set of full genotypes(_inputmat_)
    // is genotype(fullGenMap[i])in all input genotypes
	std::vector<int> mvGenMap;    
	CMatrix *orghapmat;  // PPH sln to the set of full genotypes(_inputmat_), index starts from 1
	CMatrix *completeSln;  // complete solution of the input genotype matrix
    // including phasing for both full genotypes and genotypes with missing values

	CMatrix *mymat;	// matrix used in PPH algorithm, obtained in func rearrangematrix by 
    // rearranging rows and columns of _inputmat_ according to leaf-counts, and 
    // rightmost 1 entry position (same size as _inputmat_)

    int *leafcount;		// leaf-count of each column in the input matrix (inputmat)
    int *rightmost1;	// position of rightmost 1 entry of each row in input matrix (inputmat)
    int *colindex;		// column i in mymat corresponds to column colindex[i] in inputmat
    int *rowindex;		// row i in mymat corresponds to row rowindex[i] in inputmat
    bool *boolnewentry;			// whether column i is a new entry in the shadow tree
    bool *boolInInitialTree;	// whether column i is in initial shadow tree
    bool *boolNewEntryList;		// whether column i is in NewEntryList of this row
    bool *boolOldEntryList;
    bool *boolCopyofOldEntryList;
    bool *boolCheckList;
    int *childof;		// childof[Ci] is the same as Ci.child in the algorithm

    int rowroot;	// column number of the root of each row in the shadow tree
    int colTE1;		// TE1, TEt, and E2 are used in Procedure FixTree and NewEntries
    int colTEt;		// colTE1 = col(TE1), colTEt = col(TEt)
    CConnector *pE2;	// a pointer pointing to the T connector of E2

    /// Global variables for ENT phasing method
    int noRuns = 0;  // number of times the main_ENT function is executed
    int nr_cc = 0;
    int nr_snp = 0; // total number of SNPs in sequence
    int nr_gen = 0; // total number of genotypes
    int nr_hap = 0;
    int nr_free = 0;  // free window size      // for each group of f unphased SNPs,
    int nr_locked = 0; // locked window size   // add the preceding l phased SNPs to create a window
    // (leftmost window has no locked SNPs)
    double total_cov;
    bool MODE_drop = false;
    int  MODE_count = 1;  // 0 - count all, 1 - count founders only, 2 - count children only
    int num_dropped = 0;
    int num_merrors = 0;
    std::vector<CComp*> cc_graph; // connected components table containing all genotypes
    Gen * gen_array;
    Hap * hap_table;


    bool hasPPHSolution(const ImmutableMatrix &haplotypes, 
			int colStart, int colEnd)
    {
        int errorCode = pph_main(haplotypes.view(), colStart, colEnd);
        releasemem(errorCode);
        if (errorCode == 0)  return true;
        else  return false;
    }


    int pph_main(const MatrixView &haplotypes, int colStart, int colEnd)
    {
        int errorcode;
        int rowcounter, colcounter;
	
        errorcode = retrievematrix(haplotypes, colStart, colEnd);
        if (errorcode != 0) 
            return errorcode;
	
        // rearrage rows and columns in the input matrix (inputmat) according to
        // leaf-counts, and positions of rightmost 1 entries
        errorcode = rearrangematrix();
        if (errorcode != 0)
            return errorcode;
	
        /// initialize connectors, lists, arrays in the shadow tree 
        pph_init();
	
        /// build initial tree from matrix _mymat_
        errorcode = initialtree();
        if (errorcode != 0) 
            return errorcode;
	
        /// process matrix mymat one row at a time
        for (rowcounter = 1; rowcounter <= rownumber; rowcounter ++)
		{
            rowroot = rightmost1[rowindex[rowcounter]];  // column number of the root of this row
		
            /// create OldEntryList, CopyofOldEntryList, NewEntryList, and initialize CheckList
            createentrylist(rowcounter);
		
            /// initialize the linked list of tree edges on the first path
            for (colcounter = 1; colcounter <= columnnumber; colcounter ++)
                childof[colcounter] = -1;
		
            errorcode = FirstPath (rowcounter);
            if (errorcode != 0)
                return errorcode;
		
            errorcode = SecondPath (rowcounter);
            if (errorcode != 0) 
                return errorcode;
		
            FixTree (rowcounter);
		
            errorcode = NewEntries (rowcounter);
            if (errorcode != 0) 
                return errorcode;
		}	
	
        return 0;
    }


    int retrievematrix(const MatrixView &haplotypes, int colStart, int colEnd) 
    {
        noGenotype = haplotypes.noRows();
        columnnumber = colEnd - colStart + 1;  // length of each genotype

        rownumber = 0;  // number of full genotypes
        noGenotypeMV = 0;  // number of genotypes with missing values
        bool hasMissingValue;
        for (int rowIdx = 0; rowIdx < noGenotype; ++rowIdx) {
            hasMissingValue = false;
            for (int colIdx = 0; colIdx < columnnumber; ++colIdx) {
                if (haplotypes(rowIdx, colStart+colIdx) < 0) {
                    hasMissingValue = true;
                    break;
                }
            }
            if (hasMissingValue)  noGenotypeMV++;
            else  rownumber ++;
        }

        if (rownumber > 0)
            inputmat = new CMatrix[rownumber + 1];  // index starts from 1
        genMatMV = new CMatrix[noGenotypeMV];  // index starts from 0
        fullGenMap.resize(rownumber);  // gen(i) in _inputmat is gen(fullGenMap[i]) in all gens
        mvGenMap.resize(noGenotypeMV);
        int mvGenIdx = 0, fullGenIdx = 0;
        std::vector<int> genotype(columnnumber);
        for (int rowIdx = 0; rowIdx < noGenotype; ++rowIdx) {
            hasMissingValue = false;
            for (int colIdx = 0; colIdx < columnnumber; ++colIdx) {
                genotype[colIdx] = haplotypes(rowIdx, colStart+colIdx);
                if (genotype[colIdx] < 0)  hasMissingValue = true;
            }
            if (hasMissingValue) {
                genMatMV[mvGenIdx].col = new int[columnnumber];
                mvGenMap[mvGenIdx] = rowIdx;
                for (int colIdx = 0; colIdx < columnnumber; ++colIdx)
                    genMatMV[mvGenIdx].col[colIdx] = genotype[colIdx];
                mvGenIdx++ ;
            }
            else {
                inputmat[fullGenIdx+1].col = new int[columnnumber+1];
                inputmat[fullGenIdx+1].col[0] = -1;
                fullGenMap[fullGenIdx] = rowIdx;
                for (int colIdx = 0; colIdx < columnnumber; ++colIdx)
                    inputmat[fullGenIdx+1].col[colIdx+1] = genotype[colIdx];
                fullGenIdx++ ;
            }
        }
        if (rownumber == 0)  // every genotype in _haplotypes has missing data
            return -1;  // ERR-1
        return 0;
    }	
    
    
    void getOnePPHSln(int solIdx)
    {
        int rowcounter, colcounter, i, j;	// index i for row, index j for column

        /// Flip classes in the shadow tree to obtain solution _solIdx_
        int number = solIdx;
        int flipFlag[11];
        for (i = 0; i < 11; i ++) {
            if (number >= 1) {
                flipFlag[i] = number%2;
                number = number/2;
            }
            else
                flipFlag[i] = 0;
        }
        int classIdx = 1;
        colcounter = 1;
        CConnector *crTE;
        CConnector *crSE;
        CConnector *tmpConn;
        for (colcounter = 1; colcounter <= columnnumber; colcounter ++) {
            if ((THConn[colcounter].linktype == FREE) && (!boolnewentry[colcounter])) {	
                /// the H connector of te(colcounter) is a class root
                crTE = getclassroot (colcounter, TREE);
                crSE = getclassroot (colcounter, SHADOW);

                if (flipFlag[classIdx] == 1) {
                    /// flip the class of colcounter
                    tmpConn = crTE->linkto, crTE->linkto = crSE->linkto, crSE->linkto = tmpConn;
                }
                classIdx ++;
            }
        }

        /// Generate the haplotype matrix of solution solIdx to the transformed genotype matrix
        CMatrix *hapmat;  // index starts from 1
        hapmat = new CMatrix[2*rownumber + 1];
        CConnector *p1;
        bool boolSthFound;
        for (rowcounter = 1; rowcounter <= rownumber; rowcounter ++) 
        {
            hapmat[2*rowcounter - 1].col = new int[columnnumber + 1];
            hapmat[2*rowcounter].col = new int[columnnumber + 1];

            /* find the rightmost 2 entry in this row */
            boolSthFound = false;
            colcounter = columnnumber;
            while ((!boolSthFound) && (colcounter > 0)) {
                if (mymat[rowcounter].col[colcounter] == 2)
                    boolSthFound = true;
                else
                    colcounter --;
            }	// colcounter is the rightmost 2 entry in this row
            if (colcounter == 0)
                p1 = &ConnTreeRoot;
            else
                p1 = &THConn[colcounter];
            /* p1 points to the H connector of the tree edge of the rightmost 2 entry */

            for (colcounter = columnnumber; colcounter >= 1; colcounter --) {
                if (mymat[rowcounter].col[colcounter] == 0) {
                    hapmat[2*rowcounter - 1].col[colcounter] = 0;
                    hapmat[2*rowcounter].col[colcounter] = 0;
                }
                else if (mymat[rowcounter].col[colcounter] == 1) {
                    hapmat[2*rowcounter - 1].col[colcounter] = 1;
                    hapmat[2*rowcounter].col[colcounter] = 1;
                }
                else {		// mymat[rowcounter].col[colcounter] == 2
                    if (colcounter == p1->col) {
                        hapmat[2*rowcounter - 1].col[colcounter] = 1;
                        hapmat[2*rowcounter].col[colcounter] = 0;

                        /* find next tree edge on the path from p1 to TREEROOT */
                        boolSthFound = false;
                        while ((!boolSthFound) && (p1->col > TREEROOT)) {
                            if ((p1->linkto->conntype == TCONN) && (p1->linkto->edgetype == TREE))
                                boolSthFound = true;
                            p1 = p1->linkto;
                        }
                    }
                    else {
                        hapmat[2*rowcounter - 1].col[colcounter] = 0;
                        hapmat[2*rowcounter].col[colcounter] = 1;
                    }
                }
            }	// end for each column of this row
        }	// end for each row of mymat

        orghapmat = new CMatrix[2*rownumber + 1];  // PPH sln to original set of full genotypes(_inputmat_)
        /// column i in mymat corresponds to column colindex[i] in inputmat
        /// row i in mymat corresponds to row rowindex[i] in inputmat
        for (rowcounter = 1; rowcounter <= rownumber; rowcounter ++) {
            orghapmat[2*rowcounter - 1].col = new int[columnnumber + 1];
            orghapmat[2*rowcounter].col = new int[columnnumber + 1];
        }
        for (i = 1; i <= rownumber; i++) {
            for (j = 1; j <= columnnumber; j++) {
                orghapmat[2*rowindex[i]-1].col[colindex[j]] = hapmat[2*i-1].col[j];
                orghapmat[2*rowindex[i]].col[colindex[j]] = hapmat[2*i].col[j];
            }
        }

        for (i = 1; i <= 2*rownumber; i++)
            delete [] hapmat[i].col;
        delete [] hapmat;
    }


    void imputeMissingValue()
    {
        int i, j;	// index i for row, index j for column
        /// Create a list of distinctive haplotypes from _orghapmat_
        CMatrix *distinctHaps;
        distinctHaps = new CMatrix[2*rownumber];  // start from index 0
        std::vector<int> freqDistinctHaps(2*rownumber);
        int noDistinctHaps = 0;
        int foundHap = -1;
        bool sameHap, newHap;
        for (i = 0; i < 2*rownumber; i++) {
            newHap = true;  // orghapmat[i+1] is not in distinctHaps
            for (int k = 0; k < noDistinctHaps; k++) {
                sameHap = true;
                for (j = 0; j < columnnumber; j++) {
                    if (orghapmat[i+1].col[j+1] != distinctHaps[k].col[j]) {
                        sameHap = false;
                        break;
                    }
                }
                if (sameHap) {
                    newHap = false;  // orghapmat[i+1] is the same as distinctHaps[k]
                    foundHap = k;
                    break;
                }
            }
            if (newHap) {
                distinctHaps[noDistinctHaps].col = new int[columnnumber];
                freqDistinctHaps[noDistinctHaps] = 1;
                for (j = 0; j < columnnumber; j++)
                    distinctHaps[noDistinctHaps].col[j] = orghapmat[i+1].col[j+1];
                noDistinctHaps ++ ;
            }
            else
                freqDistinctHaps[foundHap] ++ ;
        }

        /// Bubble-sort the list by each haplotype's frequency, highest freq first
        /// distinctHaps[_sortedFreqMap[0]_] has the highest freq
        std::vector<int> sortedFreqMap(noDistinctHaps);
        for (i = 0; i < noDistinctHaps; i++)  sortedFreqMap[i] = i;
        bool boolSwap, boolSortDone = false;
        int tmp;
        while (!boolSortDone) {	
            boolSwap = false;
            for (i = 0; i < noDistinctHaps-1; i++) {
                if (freqDistinctHaps[sortedFreqMap[i]] < freqDistinctHaps[sortedFreqMap[i+1]]) {
                    tmp = sortedFreqMap[i];
                    sortedFreqMap[i] = sortedFreqMap[i+1];
                    sortedFreqMap[i+1] = tmp;
                    boolSwap = true;
                }
            }
            if (!boolSwap)
                boolSortDone = true;
        }

        /// impute missing values and phase genotypes with missing value
        CMatrix *phasedGenMV;  // phased genotypes with missing value
        phasedGenMV = new CMatrix[2*noGenotypeMV];  // start from index 0
        bool compatible, identical;
        int hapIdx, hap2Idx;
        std::vector<int> secondHap(columnnumber);
        srand(time(NULL));
        for (i = 0; i < noGenotypeMV; i++) {
            hapIdx = -1, hap2Idx = -1;
            for (int k = 0; k < noDistinctHaps; k++) {
                compatible = true;  // distinctHaps[sortedFreqMap[k]] can be used to phase genMatMV[i]
                for (j = 0; j < columnnumber; j++) {
                    if ((genMatMV[i].col[j] == 0) && (distinctHaps[sortedFreqMap[k]].col[j] != 0)) {
                        compatible = false;
                        break;
                    }
                    else if ((genMatMV[i].col[j] == 1) && (distinctHaps[sortedFreqMap[k]].col[j] != 1)) {
                        compatible = false;
                        break;
                    }
                }
                if (compatible) {
                    hapIdx = sortedFreqMap[k];
                    break;
                }
            }
            if (hapIdx >= 0) {  // use distinctHaps[hapIdx] to phase genMatMV[i]
                for (j = 0; j < columnnumber; j++) {
                    if (genMatMV[i].col[j] == 2)
                        secondHap[j] = 1 - distinctHaps[hapIdx].col[j];
                    else
                        secondHap[j] = genMatMV[i].col[j];
                }
                for (int k = 0; k < noDistinctHaps; k++) {
                    identical = true;  // distinctHaps[sortedFreqMap[k2]] == secondHap
                    for (j = 0; j < columnnumber; j++) {
                        if ( (secondHap[j] >= 0) &&
                             (distinctHaps[sortedFreqMap[k]].col[j] != secondHap[j]) )
                        {
                            identical = false;
                            break;
                        }
                    }
                    if (identical) {
                        hap2Idx = sortedFreqMap[k];  // use distinctHaps[hap2Idx] to phase genMatMV[i]
                        break;
                    }
                }
                if (hap2Idx < 0) {  // no haplotype in _distinctHaps_ can be used
                    // with distinctHaps[hapIdx] to phase genMatMV[i]
                    for (j = 0; j < columnnumber; j++)
                        if (secondHap[j] < 0)
                            secondHap[j] = rand()%2;  // assign missing value randomly
                }
                phasedGenMV[2*i].col = new int[columnnumber];
                phasedGenMV[2*i+1].col = new int[columnnumber];
                for (j = 0; j < columnnumber; j++) {
                    phasedGenMV[2*i].col[j] = distinctHaps[hapIdx].col[j];
                    if (hap2Idx < 0)
                        phasedGenMV[2*i+1].col[j] = secondHap[j];
                    else
                        phasedGenMV[2*i+1].col[j] = distinctHaps[hap2Idx].col[j];
                }
            }
            else {  // genMatMV[i] cannot be phased by any haplotype in _distinctHaps_
			    // phase genMatMV[i] randomly
                phasedGenMV[2*i].col = new int[columnnumber];
                phasedGenMV[2*i+1].col = new int[columnnumber];
                for (j = 0; j < columnnumber; j++) {
                    if ((genMatMV[i].col[j] == 0) || (genMatMV[i].col[j] == 1)) {
                        phasedGenMV[2*i].col[j] = genMatMV[i].col[j];
                        phasedGenMV[2*i+1].col[j] = genMatMV[i].col[j];
                    }
                    else if (genMatMV[i].col[j] == 2) {
                        phasedGenMV[2*i].col[j] = rand()%2;
                        phasedGenMV[2*i+1].col[j] = 1 - phasedGenMV[2*i].col[j];
                    }
                    else {  // phase missing value
                        phasedGenMV[2*i].col[j] = rand()%2;
                        phasedGenMV[2*i+1].col[j] = rand()%2;
                    }
                }
            }
        }

        /// Create the complete solution(_solIdx_) of the input genotype matrix
        /// including phasing for both full genotypes and genotypes with missing values
        completeSln = new CMatrix[2*noGenotype];  // start from index 0
        for (i = 0; i < noGenotype; ++i) {
            completeSln[2*i].col = new int[columnnumber];
            completeSln[2*i+1].col = new int[columnnumber];
        } 
	
        for (i = 0; i < rownumber; ++i) {
            for (j = 0; j < columnnumber; ++j) {
                completeSln[2*fullGenMap[i]].col[j] = orghapmat[2*i+1].col[j+1];
                completeSln[2*fullGenMap[i]+1].col[j] = orghapmat[2*i+2].col[j+1];
            }
        }
	
        for (i = 0; i < noGenotypeMV; ++i) {
            for (j = 0; j < columnnumber; ++j) {
                completeSln[2*mvGenMap[i]].col[j] = phasedGenMV[2*i].col[j];
                completeSln[2*mvGenMap[i]+1].col[j] = phasedGenMV[2*i+1].col[j];
            }
        }
	
        for (i = 1; i <= 2*rownumber; i++)
            delete [] orghapmat[i].col;
        delete [] orghapmat;
        for (i = 0; i < noDistinctHaps; i++)
            delete [] distinctHaps[i].col;
        delete [] distinctHaps;
        for (i = 0; i < 2*noGenotypeMV; i++)
            delete [] phasedGenMV[i].col;
        delete [] phasedGenMV; 

    }
    

    /*
     * ERR 0: release every variable, the input matrix has a PPH solution
     * ERR-1: no need to release memory
     * ERR-2: release inputmat, leafcount, colindex, rightmost1
     * ERR-3: release every variable
     */
    void releasemem (int errorcode)
    {
        if ((errorcode == 0) || (errorcode == -2) || (errorcode == -3)) {
            for (int i = 1; i <= rownumber; i++) 
                delete [] inputmat[i].col;
            delete [] inputmat;
            if (noGenotypeMV > 0) {
                for (int i = 0; i < noGenotypeMV; i++) 
                    delete [] genMatMV[i].col;
                delete [] genMatMV;
            }
            delete [] leafcount;
            delete [] colindex;
            delete [] rightmost1;
        }
        if ((errorcode == 0) || (errorcode == -3)) {
            for (int i = 1; i <= rownumber; i++) 
                delete [] mymat[i].col;
            delete [] mymat;
            delete [] rowindex;
            delete [] TTConn;
            delete [] THConn;
            delete [] STConn;
            delete [] SHConn;
            delete [] boolnewentry;
            delete [] boolInInitialTree;
            delete [] childof;
            delete [] boolNewEntryList;
            delete [] boolOldEntryList;
            delete [] boolCopyofOldEntryList;
            delete [] boolCheckList;
        }
	}

    
    void CConnector::init(int thiscol, int thisedge, int thisconn, int thislinktype, CConnector *thislinkto)
    {
        col = thiscol;
        edgetype = thisedge;
        conntype = thisconn;
        linktype = thislinktype;
        linkto = thislinkto;
    }

    void pph_init (void) 
    {
        TTConn = new CConnector[columnnumber + 1];			// T connectors of tree edges
        THConn = new CConnector[columnnumber + 1];			// H connectors of tree edges
        STConn = new CConnector[columnnumber + 1];			// T connectors of shadow edges
        SHConn = new CConnector[columnnumber + 1];			// H connectors of shadow edges
	
        ConnTreeRoot.init(TREEROOT, TREE, TCONN, FIX, &ConnTreeRoot);	// root of the whole shadow tree
        ConnNoWhere.init (-1, -1, -1, -1, &ConnNoWhere);				// a  connector for program purpose
	
        boolnewentry = new bool[columnnumber + 1];			// whether column i is a new entry
        boolInInitialTree = new bool[columnnumber + 1];		// whether column i is in initial shadow tree
        childof = new int[columnnumber + 1];				// childof[Ci] = Ci.child in the algorithm
	
        boolNewEntryList = new bool[columnnumber + 1];
        boolOldEntryList = new bool[columnnumber + 1];
        boolCopyofOldEntryList = new bool[columnnumber + 1];
        boolCheckList = new bool[columnnumber + 1];
	
        for (int j = 1; j <= columnnumber; j++) {
            THConn[j].init(j, TREE, HCONN, FREE, &ConnNoWhere);
            TTConn[j].init(j, TREE, TCONN, FIX, &THConn[j]);
            SHConn[j].init(j, SHADOW, HCONN, FREE, &ConnNoWhere);
            STConn[j].init(j, SHADOW, TCONN, FIX, &SHConn[j]);
	    
            boolnewentry[j] = true;				// initially every column is a new entry
            boolInInitialTree[j] = false;		// initially no column is in initial shadow tree
        }
    }
    
    
    int FirstPath (int currentrow)
    {
        int Ci = getnextentry (COOEL, columnnumber + 1);	// largest column number in CopyofOldEntryList
        int Cj = getnextentry (COOEL, Ci);	
        /* Cj is the column number left to Ci in COOEL, if Ci is at the head of list, then Cj = rowroot */
	
        CConnector *p1, *root1, *root2, *tmpConn;
	
        if (Ci == rowroot)			// CopyofOldEntryList is empty
            return 0;	
	
        p1 = &THConn[Ci];
	
        while (Ci != rowroot)		// CopyofOldEntryList is not empty
	    {
            while ((p1->linktype == FIX) && (p1->linkto->conntype == HCONN))
		    { p1 = p1->linkto; }
		
            /* if (there is a fixed link form cnt(p1) to a T connector) */
            if (p1->linktype == FIX) {
                if (p1->linkto->edgetype == SHADOW) {
                    p1 = &SHConn[p1->linkto->col];
                }
                else {		// cnt(p1) links to a tree edge
                    /* col(TEp) == p1->linkto->col */
                    if (p1->linkto->col == Cj) {		// the first path extends successfully
                        childof[Cj] = Ci;
                        Ci = Cj;						// remove Ci from CopyofOldEntryList
                        Cj = getnextentry (COOEL, Ci);
                        p1 = &THConn[Ci];				// It's ok if Ci == rowroot
                    }
                    else if (p1->linkto->col < Cj) {	// col(TEp) < Cj
                        if (Cj == rowroot) {			// This is not in the algorithm.
                            return -3;	// ERR-3
                        }
                        else {
                            boolCheckList[Cj] = true;		// put Cj into CheckList
                            Cj = getnextentry (COOEL, Cj);
                        }
                    }
                    else {								// Cj < col(TEp)
                        return -3;	// ERR-3
                    }
                }	// end else (cnt(p1) links to a tree edge)
            }
		
            else {	// there is a free link from cnt(p1)
                root1 = p1;
                root2 = getclassroot (Ci, SHADOW);
                if ((root1->linkto->conntype == HCONN) || ((root1->linkto->conntype == TCONN) && (root1->linkto->edgetype == SHADOW))) {
                    /* flip the class of Ci so that root1 links to the T connector of a tree edge */
                    tmpConn = root1->linkto, root1->linkto = root2->linkto, root2->linkto = tmpConn;
                }
                else if ((root2->linkto->conntype == TCONN) && (root2->linkto->edgetype == TREE) && (root2->linkto->col > root1->linkto->col)) {
                    /* flip the class of Ci */
                    tmpConn = root1->linkto, root1->linkto = root2->linkto, root2->linkto = tmpConn;
                }
		    
                /* col(TEp) == root1->linkto->col */
                if (root1->linkto->col == Cj) {		// the first path extends successfully
                    childof[Cj] = Ci;
                    Ci = Cj;						// remove Ci from CopyofOldEntryList
                    Cj = getnextentry (COOEL, Ci);
                    p1 = &THConn[Ci];				// It's ok if Ci == rowroot.
                }
                else if (root1->linkto->col < Cj) {	// col(TEp) < Cj
                    if (Cj == rowroot) {			// this case is not in the algorithm
                        return -3;  // ERR-3
                    }
                    else {
                        /* sibling case 1 */
                        boolCheckList[Cj] = true;	// put Cj into CheckList
                        Cj = getnextentry (COOEL, Cj);
                    }
                }
                else {								// Cj < col(TEp)
                    /* Flip/Merge Case 1 */
                    tmpConn = root1->linkto, root1->linkto = root2->linkto, root2->linkto = tmpConn;
                    root1->linktype = FIX, root2->linktype = FIX;
                }
            }	// end if there is a fixed link from cnt(p1) to a T connector
	    }	// end while (CopyofOldEntryList is not empty)
	
        return 0;
    }
    

    int SecondPath (int currentrow)
    {
        int Ci = getnextentry (CL, columnnumber + 1);	// largest column number in CheckList
        int Cj = getnextentry (CL, Ci);			
        /* Cj is the column number left to Ci, if Ci is at the head of list, then Cj = rowroot */
	
        CConnector *p1, *root1, *root2, *tmpConn;
        int errorcode;
	
        if (Ci == rowroot)			// CheckList is empty
            return 0;
	
        p1 = &THConn[Ci];
	
        while (Ci != rowroot)		// CheckList is not empty
	    {
            while ((p1->linktype == FIX) && (p1->linkto->conntype == HCONN))
		    { p1 = p1->linkto; }
		
            /* if (there is a fixed link form cnt(p1) to a T connector) */
            if (p1->linktype == FIX) {
                if (p1->linkto->edgetype == SHADOW) {
                    p1 = &SHConn[p1->linkto->col];
                }
                else {		// cnt(p1) links to a tree edge
                    /* col(TEp) == p1->linkto->col */
                    if (p1->linkto->col == Cj) {		// the second path extends successfully
                        Ci = Cj;						// remove Ci from CheckList
                        Cj = getnextentry (CL, Ci);
                        p1 = &THConn[Ci];				// It's ok if Ci == rowroot.
                    }
                    else if ((p1->linkto->col < Cj) || (!boolOldEntryList[p1->linkto->col])) {	
                        /* col(TEp) < Cj, or col(TEp) isn't in OldEntryList */
                        return -3;  // ERR-3
                    }
                    else {								// Cj < col(TEp), and col(TEp) is in OldEntryList
                        /* call Procedure DirectSecondPath */
                        errorcode = DirectSecondPath (p1->linkto->col, Ci);
                        return errorcode;				// Exit Procedure SecondPath
                    }
                }	// end if cnt(p1) links to a shadow edge
            }
		
            else {			// there is a free link from cnt(p1)
                root1 = p1;
                root2 = getclassroot (Ci, SHADOW);
                if ((root1->linkto->conntype == HCONN) || ((root1->linkto->conntype == TCONN) && (root1->linkto->edgetype == SHADOW))) {
                    /* flip the class of Ci so that root1 links to the T connector of a tree edge */
                    tmpConn = root1->linkto, root1->linkto = root2->linkto, root2->linkto = tmpConn;
                }
                else if ((root2->linkto->conntype == TCONN) && (root2->linkto->edgetype == TREE) && (root2->linkto->col > root1->linkto->col)) {
                    /* flip the class of Ci */
                    tmpConn = root1->linkto, root1->linkto = root2->linkto, root2->linkto = tmpConn;
                }
		    
                /* col(TEp) = root1->linkto->col */
                if (root1->linkto->col == Cj) {		// the second path extends successfully
                    Ci = Cj;						// remove Ci from CheckList
                    Cj = getnextentry (CL, Ci);
                    p1 = &THConn[Ci];				// It's ok if Ci == rowroot.
                }
                else if (root1->linkto->col < Cj) {	// col(TEp) < Cj
                    return -3;  // ERR-3
                }
                else if ((Cj < root1->linkto->col) && (!boolOldEntryList[root1->linkto->col])) {
                    /* Cj < col(TEp), and col(TEp) isn't in OldEntryList */
                    /* the same setting as Flip/Merge Case 1 */
                    tmpConn = root1->linkto, root1->linkto = root2->linkto, root2->linkto = tmpConn;
                    root1->linktype = FIX, root2->linktype = FIX;
                }
                else {		// Cj < col(TEp), and col(TEp) is in OldEntryList
                    /* call Procedure DirectSecondPath */
                    errorcode = DirectSecondPath (root1->linkto->col, Ci);
                    return errorcode;				// Exit Procedure SecondPath
                }
            }	// end if there is a fixed link from cnt(p1) to a T connector
	    }	// end while (CheckList is not empty)
	
        return 0;
    }	
    

    int DirectSecondPath (int colTEp,int Ci)
    {
        int Cpc = childof[colTEp];
        CConnector *root1 = getclassroot (Ci, TREE);			// class root of te(Ci)
        CConnector *root2 = getclassroot (Ci, SHADOW);			// class root of se(Ci)
        CConnector *rootpc1 = getclassroot (Cpc, TREE);			// class root of te(Cpc)
        CConnector *rootpc2 = getclassroot (Cpc, SHADOW);		// class root of se(Cpc)
        CConnector *rootTEp = getclassroot (colTEp, TREE);		// class root of TEp

        Ci = getnextentry (CL, Ci);		// remove Ci from CheckList

        /* if te(Ci) and te(Cpc) are in the same class, and have the same class root, then no PPH solution 
           exists */
        if (root1 == rootpc1) 
            return -3;  // ERR-3

        bool boolteCibelowTEp = false; 	// whether TEp must be on the path from te(Ci) to the row root
        bool boolerror = false;
        int CLindex = Ci;				// used to scan all columns in CheckList
        CConnector *p1 = root2;

        if (root1  == rootTEp)		// te(Ci) and TEp are in the same class, and have the same class root
            boolteCibelowTEp = true;	// TEp must be on the path from te(Ci) to the row root
        else {
            /* Compare tree edges on the path from root2 to the row root with tree edges of all columns in 
               CheckList. */
            while (p1->linkto->col != rowroot) {
                if ((p1->linkto->edgetype == TREE) && (p1->linkto->conntype == TCONN)) {
                    if (p1->linkto->col == CLindex) 		// this tree edge matches the one in CheckList
                        CLindex = getnextentry (CL, CLindex);
                    else
                        boolerror = true;		
                }
                p1 = p1->linkto;
            }
            if (CLindex != rowroot)		// there are columns in CheckList not visited on the path from 
                boolerror = true;		// root2 to the row root

            if (boolerror) { 
                /* tree edges on the path from root2 to the row root are NOT exactly tree edges of all 
                   columns in CheckList */
                boolteCibelowTEp = true;	// TEp must be on the path from te(Ci) to the row root
            }
        } 

        bool boolteCpcbelowTEp = false;	// whether TEp must be on the path from te(Cpc) to the row root
        boolerror = false;
        CLindex = Ci;

        if (rootpc1 == rootTEp)		// te(Cpc) and TEp are in the same class, and have the same class root
            boolteCpcbelowTEp = true;		// TEp must be on the path from te(Cpc) to the row root
        else {
            /* rootpc1 should link to TEp, if not, then switch rootpc1 and rootpc2 */
            if ((rootpc1->linkto->col != colTEp) || (rootpc1->linkto->edgetype != TREE)) {
                p1 = rootpc1, rootpc1 = rootpc2, rootpc2 = p1;
            }
            p1 = rootpc2;
            /* Compare tree edges on the path from rootpc2 to the row root with tree edges of all columns in 
               CheckList. */
            while (p1->linkto->col != rowroot) {
                if ((p1->linkto->edgetype == TREE) && (p1->linkto->conntype == TCONN)) {
                    if (p1->linkto->col == CLindex) 		// this tree edge matches the one in CheckList
                        CLindex = getnextentry (CL, CLindex);
                    else
                        boolerror = true;		
                }
                p1 = p1->linkto;
            }
            if (CLindex != rowroot)		// there are columns in CheckList not visited on the path from 
                boolerror = true;		// rootpc2 to row root

            if (boolerror) { 
                /* tree edges on the path from rootpc2 to the row root are NOT exactly tree edges of all 
                   columns in CheckList */
                boolteCpcbelowTEp = true;	// TEp must be on the path from te(Cpc) to the row root
            }
        } 

        /* if flags indicate that TEp must be on the path from both te(Ci) and te(Cpc) to the row root */
        if ((boolteCibelowTEp) && (boolteCpcbelowTEp)) 
            return -3;  // ERR-3
        /* if flags indicate that TEp must be on the path from te(Ci) to the row root */
        else if ((boolteCibelowTEp) && (!boolteCpcbelowTEp)) {
            /* Flip/Merge Case 2 */
            p1 = rootpc1->linkto, rootpc1->linkto = rootpc2->linkto, rootpc2->linkto = p1;
            rootpc1->linktype = FIX, rootpc2->linktype = FIX;
            if (root1 != rootTEp) {	// Ci and TEp are not in the same class
                root1->linktype = FIX, root2->linktype = FIX;
            }
        }
        /* if flags indicate that TEp must be on the path from te(Cpc) to the row root */
        else if (!(boolteCibelowTEp) && (boolteCpcbelowTEp)) {
            /* Flip/Merge Case 2 */
            p1 = root1->linkto, root1->linkto = root2->linkto, root2->linkto = p1;
            root1->linktype = FIX, root2->linktype = FIX;
            if (rootpc1 != rootTEp) {	// Cpc and TEp are not in the same class
                rootpc1->linktype = FIX, rootpc2->linktype = FIX;
            }
        }
        else {		// none of te(Ci) and te(Cpc) must be on the same path as TEp to the row root
            /* Flip/Merge Case 3 */
            p1 = root1->linkto, root1->linkto = root2->linkto, root2->linkto = p1;
            if (root1 != rootpc1) {		// Ci and Cpc are not in the same class
                rootpc1->linkto = root2, rootpc1->linktype = FIX;
                rootpc2->linkto = root1, rootpc2->linktype = FIX;
            }
        }
        return 0;
    }	


    void FixTree (int currentrow)
    {
        /* First we need to find TE1 and E2 which are two ends of the path passing through the row root and
           tree edges corresponding to all columns in OldEntryList. */
        colTE1 = getnextentry (OEL, columnnumber + 1);		// largest column number in OldEntryList
        if (colTE1 == rowroot)		// OldEntryList is empty
            return;

        CConnector *p1 = &THConn[colTE1];		// p1 points to H connector of TE1

        /* TEt is the firt tree edge on the second path to the row root, TEt could be the same as row root */
        colTEt = getnextentry (OEL, colTE1);
        bool boolTEtfound = false;		// whether TEt is found
        while ((!boolTEtfound) && (p1->linkto->col != rowroot)) {
            if ((p1->linkto->edgetype == TREE) && (p1->linkto->conntype == TCONN)) {
                if (p1->linkto->col == colTEt)
                    colTEt = getnextentry (OEL, colTEt);
                else
                    boolTEtfound = true;		
            }
            p1 = p1->linkto;
        }		// colTEt is found

        /* Start a DFS in the subtree below TEt after contracting all links in the shadow tree. Choose a
           shadow edge whose column number is in OldEntryList as the next edge to visit. If no such shadow edge
           exists, then terminates the DFS. E2 is the edge that is visited last in the DFS. */
        if (colTEt == TREEROOT)
            pE2 = &ConnTreeRoot;		// pE2 points to the T connector of E2
        else
            pE2 = &TTConn[colTEt];
        int colcounter;
        for (colcounter = colTEt + 1; colcounter <= columnnumber; colcounter ++) {
            if (boolOldEntryList[colcounter]) {
                if (getParentConn(&SHConn[colcounter]) == pE2)
                    pE2 = &STConn[colcounter];		// DFS visits next shadow edge
            }
        }	// now pE2 points to the T connector of E2
	
        CConnector *crE2;		// class root of E2
        CConnector *crTE2;		// the other class root of E2 (the class root of te(E2))
        if (pE2->col == TREEROOT) {
            crE2 = &ConnTreeRoot;
            crTE2 = &ConnTreeRoot;
        }
        else if (pE2->col == rowroot) {
            crE2 = &TTConn[rowroot];
            crTE2 = &TTConn[rowroot];
        }
        else if (pE2->col == colTEt) {
            crE2 = getclassroot (colTEt, TREE);
            crTE2 = getclassroot (colTEt, SHADOW);
        }
        else {
            crE2 = getclassroot (pE2->col, SHADOW);
            crTE2 = getclassroot (pE2->col, TREE);
        }

        CConnector *crSE1;										// class root of SE1
        CConnector *crTE1 = getclassroot (colTE1, TREE);		// class root of TE1, colTE1 != rowroot
        if (boolInInitialTree[colTE1])
            crSE1 = getclassroot (colTE1, TREE);
        else
            crSE1 = getclassroot (colTE1, SHADOW);
	
        /* repeat until (TE1 and E2 are in the same class, or E2 is the parent of the class root of SE1) */
        while ((crSE1 != crE2) && (pE2 != getParentConn(crSE1))) {
            if (crTE1->col > crE2->col) {
                /* merge the class of TE1 with its attaching class */
                crTE1->linktype = FIX;
                crSE1->linktype = FIX;
                crSE1 = getclassroot (crSE1->col, crSE1->edgetype);
                crTE1 = getclassroot (crTE1->col, crTE1->edgetype);
            }
            else {
                /* merge the class of E2 with its attaching class, in this case E2 is a shadow edge */
                crE2->linktype = FIX;
                crTE2->linktype = FIX;
                crE2 = getclassroot (crE2->col, crE2->edgetype);
                crTE2 = getclassroot (crTE2->col, crTE2->edgetype);
            }
        }
    }	


    int NewEntries (int currentrow)
    {
        int colcounter;
        int Ci = getnextentry (NEL, columnnumber + 1);		// largest column number in NewEntryList
        int Cj = getnextentry (NEL, Ci);					// column number left to Ci in NewEntryList

        if (Ci == rowroot)			// NewEntryList is empty
            return 0;

        while (Cj != rowroot) {		// Ci != Ch
            THConn[Ci].linkto = &TTConn[Cj];		// p(te(Ci)) = te(Cj)
            SHConn[Ci].linkto = &STConn[Cj];		// p(se(Ci)) = se(Cj)
            Ci = Cj;
            Cj = getnextentry (NEL, Ci);
        }
	
        int Ch = Ci;				// the smallest column number in NewEntryList

        /* colTE1, colTEt, and pE2 are global variables, and are obtained in function FixTree */
        CConnector *crSE1;			// class root of SE1
        CConnector *crTE1;
        CConnector *crE2;			// class root of E2

        if (colTE1 < Ch) {
            if (colTE1 == TREEROOT) {		// OldEntryList is empty, and no 1 entry in this row
                THConn[Ch].linkto = &ConnTreeRoot;
                SHConn[Ch].linkto = &ConnTreeRoot;
                THConn[Ch].linktype = FIX;
                SHConn[Ch].linktype = FIX;
            }
            else if (colTE1 == rowroot) {	// OldEntryList is empty
                THConn[Ch].linkto = &TTConn[rowroot];
                SHConn[Ch].linkto = &TTConn[rowroot];
                THConn[Ch].linktype = FIX;
                SHConn[Ch].linktype = FIX;
            }
            else {							// OldEntryList is NOT empty
                /* get crSE1 and crE2 */
                if (boolInInitialTree[colTE1])
                    crSE1 = getclassroot (colTE1, TREE);		// crSE1 is the same as cr
                else
                    crSE1 = getclassroot (colTE1, SHADOW);

                if (pE2->col == TREEROOT)
                    crE2 = &ConnTreeRoot;
                else if (pE2->col == rowroot)
                    crE2 = &TTConn[rowroot];
                else if (pE2->col == colTEt)
                    crE2 = getclassroot (colTEt, TREE);
                else
                    crE2 = getclassroot (pE2->col, SHADOW);

                if (crSE1 == crE2) {						// TE1 and E2 are in the same class
                    /* Add-new-class Case 1 */
                    THConn[Ch].linkto = &TTConn[colTE1];	// p(te(Ch)) = TE1, p(se(Ch)) = E2
                    SHConn[Ch].linkto = pE2;				// pE2 points to the T connector of E2
                }
                else {										// TE1 and E2 are in different classes
                    /* Add-new-class Case 2 */
                    THConn[Ch].linkto = &TTConn[colTE1];	// p(te(Ch)) = TE1
                    SHConn[Ch].linkto = crSE1;				// H connector of se(Ch) links to crSE1, i.e. cr
                }
            }
            /* In this case no class need to be merged. */
        }	// end if (colTE1 < Ch)

        else {		// rowroot < Ch < colTE1
		
            /* Use the same way as to find TE1 and E2 in function FixTree to find two edges TE1' and E2'
               whose column numbers are less than Ch */
		
            /* find TE1' */
            int colTE1pr = colTE1;	// col(TE1')
            while (colTE1pr > Ch)	// the largest col# in OEL that is less than Ch 
                colTE1pr = getnextentry (OEL, colTE1pr);	// colTE1pr could be rowroot

            CConnector *p1;
            CConnector *pE2pr;		// a pointer pointing to T connector of E2'
            CConnector *crSE1pr;	// class root of SE1'
            CConnector *crTE1pr;	// class root of TE1'
            CConnector *crE2pr;		// class root of E2'
            int colTEtpr;			// col(TEt')
		
            /* find E2' */
            if (colTE1pr == TREEROOT)
                pE2pr = &ConnTreeRoot;
            else if (colTE1pr == rowroot)
                pE2pr = &TTConn[rowroot];
            else {								// rowroot < colTE1pr < Ch
			
                p1 = &THConn[colTE1pr];			// p1 points to H connector of TE1'

                /* TEt' is the tree edge of the largest col# in OEL that is less than Ch, whose tree edge is
                   not on the path from TE1' to the row root */
                colTEtpr = getnextentry (OEL, colTE1pr);
                bool boolTEtprfound = false;	// whether TEt' is found
                while ((!boolTEtprfound) && (p1->linkto->col != rowroot)) {
                    if ((p1->linkto->edgetype == TREE) && (p1->linkto->conntype == TCONN)) {
                        if (p1->linkto->col == colTEtpr)
                            colTEtpr = getnextentry (OEL, colTEtpr);
                        else
                            boolTEtprfound = true;		
                    }
                    p1 = p1->linkto;
                }		// colTEtpr is found

                if (colTEtpr == TREEROOT)
                    pE2pr = &ConnTreeRoot;		// pE2pr points to the T connector of E2'
                else
                    pE2pr = &TTConn[colTEtpr];
				
                for (colcounter = colTEtpr + 1; colcounter < Ch; colcounter ++) {
                    if (boolOldEntryList[colcounter]) {
                        if (getParentConn(&SHConn[colcounter]) == pE2pr)
                            pE2pr = &STConn[colcounter];		// DFS visits next shadow edge
                    }
                }	// now pE2pr points to the T connector of E2'
            }	
            /* E2' is found, pE2pr points to the T connector of E2' */

            bool boolTE1belowTE1pr = false;		// whether TE1' is on the path from TE1 to the row root
            if (colTE1pr > rowroot) {
                p1 = &THConn[colTE1];
                while (p1->linkto->col != rowroot) {
                    p1 = p1->linkto;
                    if ((p1->col == colTE1pr) && (p1->conntype == TCONN) && (p1->edgetype == TREE))
                        boolTE1belowTE1pr = true;
                }
            }	// boolTE1belowTE1pr is set

            /* Add-new-edges Cases 3-6 */			// rowroot < Ch < colTE1
            if (colTE1pr == TREEROOT) {
                THConn[Ch].linkto = &ConnTreeRoot;
                SHConn[Ch].linkto = &ConnTreeRoot;
                THConn[Ch].linktype = FIX;
                SHConn[Ch].linktype = FIX;
                crSE1pr = &ConnTreeRoot;
                crTE1pr = &ConnTreeRoot;
            }
            else if (colTE1pr == rowroot) {
                THConn[Ch].linkto = &TTConn[rowroot];
                SHConn[Ch].linkto = &TTConn[rowroot];
                THConn[Ch].linktype = FIX;
                SHConn[Ch].linktype = FIX;
                crSE1pr = &TTConn[rowroot];
                crTE1pr = &TTConn[rowroot];
            }
            else {									// rowroot < colTE1pr < Ch < colTE1
                if (boolInInitialTree[colTE1pr])
                    crSE1pr = getclassroot (colTE1pr, TREE);	// class root of SE1', the same as cr'
                else
                    crSE1pr = getclassroot (colTE1pr, SHADOW);
                crTE1pr = getclassroot (colTE1pr, TREE);

                if (pE2pr->col == TREEROOT)						// pE2pr points to the T connector of E2'
                    crE2pr = &ConnTreeRoot;						// class root of E2'
                else if (pE2pr->col == rowroot)
                    crE2pr = &TTConn[rowroot];
                else if (pE2pr->col == colTEtpr)
                    crE2pr = getclassroot (colTEtpr, TREE);
                else
                    crE2pr = getclassroot (pE2pr->col, SHADOW);

                if (crSE1pr == crE2pr) {		// TE1' and E2' are in the same class
                    if (boolTE1belowTE1pr) {	// TE1' is on the path from TE1 to the row root
                        /* Add-new-edges Case 3 */
                        THConn[Ch].linkto = pE2pr;					// p(te(Ch)) = E2'
                        SHConn[Ch].linkto = &TTConn[colTE1pr];		// p(se(Ch)) = TE1'
                    }
                    else {
                        /* Add-new-edges Case 4 */
                        THConn[Ch].linkto =	&TTConn[colTE1pr];		// p(te(Ch)) = TE1'
                        SHConn[Ch].linkto = pE2pr;					// p(se(Ch)) = E2'
                    }
                }
                else {							// TE1' and E2' are in different classes
                    if (boolTE1belowTE1pr) {	// TE1' is on the path from TE1 to the row root
                        /* Add-new-edges Case 5 */
                        THConn[Ch].linkto = crSE1pr;				// H connector of te(Ch) links to cr'
                        SHConn[Ch].linkto = &TTConn[colTE1pr];		// p(se(Ch)) = TE1'
                    }
                    else {
                        /* Add-new-edge Cases 6 */
                        THConn[Ch].linkto =	&TTConn[colTE1pr];		// p(te(Ch)) = TE1'
                        SHConn[Ch].linkto = crSE1pr;				// H connector of se(Ch) links to cr'
                    }
                }	// end if (TE1' and E2' are in the same class)
            }
            /* Add-new-edges Cases 3-6 are done */					

            /* find Ct */
            int Ct = colTE1;			// Ct is smallest col# in NewEntryList that is larger than col(TE1)
            bool boolCtfound = false;	// if no col# in NEL is larger than col(TE1), then let Ct = col(TE1)
            colcounter = colTE1 + 1;
            while ((!boolCtfound) && (colcounter <= columnnumber)) {
                if (boolNewEntryList[colcounter]) {
                    Ct = colcounter;
                    boolCtfound = true;
                }
                colcounter ++;
            }	// Ct is found (rowroot < Ch < colTE1 <= Ct)

            /* Add-new-edges Case 7 */
            if (Ct > colTE1)
                SHConn[Ct].linkto = &TTConn[colTE1];		// p(se(Ct)) = TE1

            /* merge the class of Ch with classes of col# in NewEntryList that are less than Ct */
            for (colcounter = Ch + 1; colcounter < Ct; colcounter ++) {
                if (boolNewEntryList[colcounter]) {
                    THConn[colcounter].linktype = FIX;
                    SHConn[colcounter].linktype = FIX;
                }
            }

            /* merge the class of TE1 with classes of columns in OldEntryList that are larger than Ch */
            /* tree edges of columns in OEL that are larger than Ch must be on one path to row root */
            colcounter = getnextentry (OEL, colTE1);
            p1 = &THConn[colTE1];
            crTE1 = getclassroot (colTE1, TREE);
            crSE1 = getclassroot (colTE1, SHADOW);
            while (p1->linkto->col > Ch) {
                p1 = p1->linkto;
                if ((p1->conntype == TCONN) && (p1->edgetype == TREE)) {
                    if (p1->col != colcounter) {		// tree edge doesn't match
                        return -3;  // ERR-3
                    }	
                    else {								// tree edge matches
                        colcounter = getnextentry (OEL, colcounter);
                        /* if the tree edge encountered and TE1 aren't in the same class */
                        if (crTE1 != getclassroot(p1->col, TREE)) {
                            crTE1->linktype = FIX;
                            crSE1->linktype = FIX;
                            crTE1 = getclassroot (crTE1->col, crTE1->edgetype);
                            crSE1 = getclassroot (crSE1->col, crSE1->edgetype);
                        }
                    }
                }
            }	// end while
            if (colcounter > Ch) { // one tree edge of column in OEL that is larger that Ch isn't on the path
                return -3;  // ERR-3
            }
		
            /* merge the class of Ch with the class of TE1 */
            if ((getParentConn(&THConn[Ch]) == getParentConn(crSE1)) && (getParentConn(&SHConn[Ch]) == getParentConn(crTE1))) {	
                /* the class of Ch doesn't attaches to the class of TE1, in this case, the classes of Ch and 
                   TE1 both attach to one other class */
                crTE1->linkto = &SHConn[Ch];
                crSE1->linkto = &THConn[Ch];
                crTE1->linktype = FIX;
                crSE1->linktype = FIX;
            }
            else {				// the class of Ch attaches to the class of TE1
                THConn[Ch].linktype = FIX;
                SHConn[Ch].linktype = FIX;
            }
        }	// end else (rowroot < Ch < colTE1 <= Ct)

        return 0;
    }	


    void createentrylist(int currentrow)
    {
        int colcounter;

        /* initialize all lists to be empty */
        for (colcounter = 1; colcounter <= columnnumber; colcounter ++) {
            boolOldEntryList[colcounter] = false;
            boolCopyofOldEntryList[colcounter] = false;
            boolNewEntryList[colcounter] = false;
            boolCheckList[colcounter] = false;
        }

        /* no 2 entry exists left to rowroot */
        for (colcounter = rowroot + 1; colcounter <= columnnumber; colcounter ++) {
            if (mymat[currentrow].col[colcounter] == 2) {
                if (boolnewentry[colcounter]) {			// it's a new entry
                    boolNewEntryList[colcounter] = true; // put it into NewEntryList
                    boolnewentry[colcounter] = false;	// then it becomes an old entry
                }
                else {									// it's an old entry
                    boolOldEntryList[colcounter] = true;
                    boolCopyofOldEntryList[colcounter] = true;
                }
            }
        }
    }	

    int getnextentry (int whichlist, int startcol) 
    {
        /// If no 2 entry is left to startcol in the specified entry list, then nextentry=rowroot.
        int nextentry = rowroot;
        int colcounter;

        for (colcounter = startcol - 1; colcounter > rowroot; colcounter --) {
            if (whichlist == OEL) {				// OldEntryList
                if (boolOldEntryList[colcounter]) return colcounter;
            } else if (whichlist == COOEL) {	// CopyofOldEntryList)
                if (boolCopyofOldEntryList[colcounter]) return colcounter;
            } else if (whichlist == NEL) {		// NewEntryList
                if (boolNewEntryList[colcounter]) return colcounter;
            } else if (whichlist == CL) {		// CheckList
                if (boolCheckList[colcounter]) return colcounter;
            }
        }
        return nextentry;
    }	

    CConnector *getclassroot (int thiscol, int thisedge) 
    {
        CConnector *p1;
	
        if (thiscol == TREEROOT)
            return &ConnTreeRoot;
        else if (thiscol == rowroot)
            return &TTConn[rowroot];
        else {
            if (thisedge == TREE)
                p1 = &THConn[thiscol];
            else
                p1 = &SHConn[thiscol];

            while ((p1->linktype == FIX) && (p1->col != rowroot))
                p1 = p1->linkto;
        }
        return p1;
    }	

    CConnector *getParentConn (CConnector *thisConn)
    {
        CConnector *p1 = thisConn->linkto;
        while ((p1->col > rowroot) && (p1->conntype != TCONN))
            p1 = p1->linkto;

        return p1;
    }	

    int initialtree(void) 
    {
        int rowcounter, colcounter;
        int rightmost1pos;		// the position of rightmost 1 entry of each row in mymat
        CConnector *p1;

        for (rowcounter = 1; rowcounter <= rownumber; rowcounter ++) {
            rightmost1pos = rightmost1[rowindex[rowcounter]];

            if (rightmost1pos >= 1) {		// there is at least one 1 entry in this row
                p1 = &THConn[rightmost1pos];

                for (colcounter = rightmost1pos - 1; colcounter >= 1; colcounter --) {
                    if (mymat[rowcounter].col[colcounter] == 1) {
                        if (boolnewentry[p1->col]) {				// p1 points to a new entry
                            p1->linkto = &TTConn[colcounter];		// build initial shadow tree
                            p1->linktype = FIX;
                            boolnewentry[p1->col] = false;			// now it becomes an old entry
                            boolInInitialTree[p1->col] = true;		// it's in initial tree now
                        }
                        else {					// p1 points to an old entry
                            if (p1->linkto->col != colcounter)
                                return -3;  // ERR-3
                        }
                        p1 = &THConn[colcounter];
                    }
                }

                /* for the first 1 entry in this row */
                if (boolnewentry[p1->col]) {				// p1 points to a new entry
                    p1->linkto = &ConnTreeRoot;				// link to the root of the whole shadow tree
                    p1->linktype = FIX;
                    boolnewentry[p1->col] = false;			// now it becomes an old entry
                    boolInInitialTree[p1->col] = true;		// it's in initial tree now
                }
                else {					// p1 points to an old entry
                    if (p1->linkto->col != TREEROOT)
                        return -3;  // ERR-3
                }
            }	// end if (rightmost1pos >= 1)
        }	// end for

        return 0;
    }

    int rearrangematrix(void) 
    {
        int rowcounter, colcounter;
        int tmp;
        CMatrix *tmpmat;  // tmpmat is got by ordering columns in input matrix by leaf-counts

        /* calculate leaf-count of each column in inputmat */
        leafcount = new int[columnnumber + 1];
        for (colcounter = 1; colcounter <= columnnumber; colcounter ++) {
            leafcount[colcounter] = 0;
            for (rowcounter = 1; rowcounter <= rownumber; rowcounter ++) {
                if (inputmat[rowcounter].col[colcounter] == 1)
                    leafcount[colcounter] += 2;
                else if (inputmat[rowcounter].col[colcounter] == 2)
                    leafcount[colcounter] ++;
            }
        }

        /* calculate colindex, where column i in mymat corresponds to column colindex[i] in inputmat */
        colindex = new int[columnnumber + 1];
        for (colcounter = 1; colcounter <= columnnumber; colcounter ++)
            colindex[colcounter] = colcounter;		// initialize

        bool boolSwap, boolSortDone = false;		// used in bubble sort below
        while (!boolSortDone) {			// bubble sort of leaf-counts
            boolSwap = false;
            for (colcounter = 1; colcounter <= columnnumber - 1; colcounter ++) {
                if (leafcount[colindex[colcounter]] < leafcount[colindex[colcounter + 1]]) {
                    tmp = colindex[colcounter];
                    colindex[colcounter] = colindex[colcounter + 1];
                    colindex[colcounter + 1] = tmp;
                    boolSwap = true;
                }
            }
            if (!boolSwap)
                boolSortDone = true;
        }

        /* re-arrange columns in the input matrix by leaf-counts decreasingly, the result is tmpmat */
        tmpmat = new CMatrix[rownumber + 1];
        for (rowcounter = 1; rowcounter <= rownumber; rowcounter ++) {
            tmpmat[rowcounter].col = new int[columnnumber + 1];
            for (colcounter = 1; colcounter <= columnnumber; colcounter ++) {
                tmpmat[rowcounter].col[colcounter] = inputmat[rowcounter].col[colindex[colcounter]];
            }
        }

        /* calculate the position of rightmost 1 entry of each row in input matrix, and store it in rightmost1[] */
        bool bool2encountered;
        rightmost1 = new int[rownumber + 1];
        for (rowcounter = 1; rowcounter <= rownumber; rowcounter ++) {
            rightmost1[rowcounter] = 0;
            bool2encountered = false;
            for (colcounter = 1; colcounter <= columnnumber; colcounter ++) {
                if (tmpmat[rowcounter].col[colcounter] == 1) {
                    if (!bool2encountered)
                        rightmost1[rowcounter] = colcounter;
                    else {
                        for (int i = 1; i <= rownumber; i++)
                            delete [] tmpmat[i].col;
                        delete [] tmpmat;
                        return -2;	// ERR-2
                    }
                }
                else if (tmpmat[rowcounter].col[colcounter] == 2)
                    bool2encountered = true;
            }
        }

        /* calculate rowindex[], where row i in mymat corresponds to row rowindex[i] in inputmat */
        rowindex = new int[rownumber + 1];
        for (rowcounter = 1; rowcounter <= rownumber; rowcounter ++)
            rowindex[rowcounter] = rowcounter;		// initialize


        pphSortUtil::indirectSort(rowindex + 1, rightmost1, rownumber);


        //boolSortDone = false;			// used in bubble sort below
        //while (!boolSortDone) {			// bubble sort of rightmost1 array
        //    boolSwap = false;
        //    for (rowcounter = 1; rowcounter <= rownumber - 1; rowcounter ++) {
        //        if (rightmost1[rowindex[rowcounter]] < rightmost1[rowindex[rowcounter + 1]]) {
        //            tmp = rowindex[rowcounter];
        //            rowindex[rowcounter] = rowindex[rowcounter + 1];
        //            rowindex[rowcounter + 1] = tmp;
        //            boolSwap = true;
        //        }
        //    }
        //    if (!boolSwap)
        //		boolSortDone = true;
        //}	

        /// re-arrange rows in tmpmat by positions of rightmost 1 entries decreasingly, the result is mymat
        mymat = new CMatrix[rownumber + 1];
        for (rowcounter = 1; rowcounter <= rownumber; rowcounter ++) {
            mymat[rowcounter].col = new int[columnnumber + 1];
            for (colcounter = 1; colcounter <= columnnumber; colcounter ++)
                mymat[rowcounter].col[colcounter] = tmpmat[rowindex[rowcounter]].col[colcounter];
        }

        /// release memory
        for (rowcounter = 1; rowcounter <= rownumber; rowcounter ++)
            delete [] tmpmat[rowcounter].col;
        delete [] tmpmat;

        return 0;
    }	
    
    
    int countClasses(void) 
    {
        int colcounter;
        int numofclass = 1;  // number of classes in the shadow tree
        for (colcounter = 1; colcounter <= columnnumber; colcounter ++)
            if ((THConn[colcounter].linktype == FREE) && (!boolnewentry[colcounter]))
                numofclass ++;

        return numofclass;
    }
    
    
    ///////////////////////////////////////////////////////////////////////////////
    //                    ENT phasing method for Blossoc                         //
    ///////////////////////////////////////////////////////////////////////////////
    void main_ENT(void)
    {	
        bool MODE_greedy = true;  // use batching
        bool free_set = false;  //_ choose nr_free dynamically
        bool locked_set = false;  //_ choose nr_locked dynamically
        srand(int(time(NULL)));

        if (rownumber == 0) {  // every genotype in the input file contains missing data
            specialcase();
            return;
        }
        read_genotypes();  //_ import the genotype matrix from _inputmat
	
        int min_cc;
        double min_cov, cur_cov;

        int sav_locked = nr_locked; int sav_free = nr_free;
        int cc_graph_size = cc_graph.size();

#if EXP
        int total_sol = 0;
        int num_iterations = 0;	
#endif
        int num_recomb = 0;
        int nr_win = 0; int tot_locked = 0; int tot_free = 0;

        CComp ** sorted_config;
        int sorted_size;
        bool recomb_clear = false, dropped = false;
	
        if(MODE_greedy) sorted_config = new CComp*[cc_graph_size];

        // for each window
        for(int i=0 ; i<nr_snp ; i+=nr_free){
            nr_win++;
            if(!free_set) { nr_free = sav_free = compute_free(i); tot_free += nr_free; }
            if(!locked_set) { nr_locked = sav_locked = compute_locked(i); tot_locked += nr_locked; }
			
            // if we used a small nr_free in the last window, return to the original
            if(nr_free != sav_free) nr_free = sav_free;
            if(nr_locked != sav_locked) nr_locked = sav_locked;
		
            if(i-nr_locked < 0) nr_locked = i;
            if(i+nr_free > nr_snp) sav_free = nr_free = nr_snp - i;

            // clear the hap list
            nr_hap = 0;

            // generate random haplotypes and populate hap_list
            for(int ii=0;ii<cc_graph.size();ii++){
                cc_graph[ii]->num_sol = 0;
			
                for(int iii=cc_graph[ii]->size-1;iii>=0;iii--) {
                    cc_graph[ii]->node[iii]->h_count[0] = cc_graph[ii]->node[iii]->h_count[1] = 0;
                }
			
                if(cc_graph[ii]->recomb){
                    // recombination occured - drop its inheritence pattern
                    drop_inheritence(cc_graph[ii]);
                    cc_graph[ii]->recomb = false;
                }
			
                traverse(cc_graph[ii],cc_graph[ii]->size-1,i,0,(double*)NULL,RAND_SOL);
#if EXP
                total_sol += cc_graph[ii]->num_sol;
#endif

                // --- check for recombinations
                recomb_clear = false;
                while(cc_graph[ii]->num_sol == 0) {
                    if(recomb_clear){
                        // we have dropped the inheritence pattern but still no results
                        fprintf(stderr,"no solution - No:%d SNP:%d\n",cc_graph[ii]->node[0]->id,i);
                        if(!MODE_drop) exit(0);
                        fprintf(stderr,"dropping pedigree\n");
                        // -- drop this pedigree
                        nr_gen -= cc_graph[ii]->size;
                        cc_graph.erase(cc_graph.begin() + ii);
                        dropped = true;
                        num_dropped++;
                        nr_free = sav_free;
                        break;
                    }
					
                    cc_graph[ii]->recomb = true; // mark this as having a recombinant event
                    nr_free--;
					
                    if(nr_free == 0){
                        num_recomb++;
                        // drop the inheritence pattern
                        drop_inheritence(cc_graph[ii]);
                        cc_graph[ii]->recomb = false;
                        nr_free = 1;
                        recomb_clear = true;
                    }
                    traverse(cc_graph[ii],cc_graph[ii]->size-1,i,0,(double*)NULL,RAND_SOL);
                }
                if(dropped) {
                    dropped = false;
                    ii--;
                    continue;
                }
                if(cc_graph[ii]->recomb){num_recomb++;}
                // continue the remaining pedigree and ent_phasing with new nr_free value
                cc_save(cc_graph[ii],i,SAVE_NEW);
            }
	
            // compute & save total entropy
            total_cov = min_cov = get_coverage();

            double dif_ent = -1;
            while( dif_ent == -1 || dif_ent > 0 ){
                sorted_size = 0;
#if VERBOSE >= 1
                double total_entropy = (-1.0/(2.0*nr_gen))*(-2.0*nr_gen*log(2.0*nr_gen)/log(2.0) - total_cov);
                //			fprintf(stderr,"min entropy = %f window [%d - %d]\n",total_entropy,i-nr_locked,i+nr_free-1);
#endif
			
                // traverse each connected component
                for(int ii=0;ii<cc_graph_size;ii++){
				
                    if(cc_graph[ii]->num_sol != 1){
                        cc_graph[ii]->num_sol = 0;
                        cur_cov = min_cov;
					
                        for(int iii=0;iii<cc_graph[ii]->size;iii++) 
                            cc_graph[ii]->node[iii]->h_count[0] = cc_graph[ii]->node[iii]->h_count[1] = 0;
					
                        traverse(cc_graph[ii],cc_graph[ii]->size-1,i,total_cov,&cur_cov,MIN_SOL);
					
                        // restore the original haplotypes for this node
                        for(int iii=0;iii<cc_graph[ii]->size;iii++){
                            strncpy(cc_graph[ii]->node[iii]->h[0]+i,cc_graph[ii]->node[iii]->h_init[0],nr_free);
                            strncpy(cc_graph[ii]->node[iii]->h[1]+i,cc_graph[ii]->node[iii]->h_init[1],nr_free);
                        }

#if EXP
                        num_iterations++;
#endif
                        if(MODE_greedy) {
                            cc_graph[ii]->cur_cov = cur_cov;
                            insert_sort_cc(cc_graph[ii],sorted_config,sorted_size++);
                        }
                        if(cur_cov < (min_cov - 1e-9)){
                            min_cov = cur_cov;
                            min_cc = ii;
                        }
                    }
                }
			
                // if we generated a smaller entropy
                if(min_cov < (total_cov - 1e-9)) {
                    // GREEDY: go through all components and save if they're smaller
                    if(MODE_greedy){
                        cc_save(sorted_config[0],i,SAVE_OVER);
                        for(int ii=1;ii<sorted_size;ii++) {
                            if(cc_save_min(sorted_config[ii],0,i,min_cov,&min_cov)){
#if VERBOSE >= 1
                                total_entropy = (-1.0/(2.0*nr_gen))*(-2.0*nr_gen*log(2.0*nr_gen)/log(2.0) - min_cov);
                                //							fprintf(stderr," secondary minimal solution found - %f\n",total_entropy);
#endif
                            }
                        }
                        // NORMAL: save only the minimal component
                    } else { cc_save(cc_graph[min_cc],i,SAVE_OVER); }
                }
                dif_ent = total_cov - min_cov;
                total_cov = min_cov;
            }		
            // for each unsaved component, save it
        }
        outputENTsol();  //_ copy solution to _orghapmat, which contains hapoltypes 
        //  inferred from all full genotypes, index starts from 1 
        return;
    }


    void drop_inheritence(CComp * c){
        for(int i=0;i<c->size;i++) {
            c->node[i]->inherit[0] = c->node[i]->inherit[1] = -1;
            c->node[i]->inherit_set[0] = c->node[i]->inherit_set[1] = false;
        }
    }

    /*
     *	compute the free window size by looking at the number of ambiguous positions in all sequences
     */
    int compute_free(int pos){
        int max = 2 * nr_gen;
        int nr_ambig = 0;
        int i = pos;
	
        while(nr_ambig <= max && i < nr_snp){
            for(int n=0;n<cc_graph.size();n++){
                for(int nn=0;nn<cc_graph[n]->size;nn++){
                    if(cc_graph[n]->node[nn]->seq[i] == '2' || cc_graph[n]->node[nn]->seq[i] == '?'){
                        nr_ambig++;
                    }
                }
            }
            i++;
            if( (i-pos) > 10 ) return 10;
        }
	
        if( (i-pos) <= 1) return 1;
        else return (i-pos) - 1;
    }

    int compute_locked(int pos){
        int max = 2 * nr_gen;
        int nr_ambig = 0;
        int i = pos;
	
        while(nr_ambig <= max && i >= 0){
            for(int n=0;n<cc_graph.size();n++){
                for(int nn=0;nn<cc_graph[n]->size;nn++){
                    if(cc_graph[n]->node[nn]->seq[i] == '2' || cc_graph[n]->node[nn]->seq[i] == '?'){
                        nr_ambig++;
                    }
                }
            }
            i--;
            if( (pos - i) > 10 ) return 10;
        }
	
        if( (pos - i) <= 1) return 1;
        else return (pos - i) - 1;
    }

    /*
     * Copy ENT solution to _orghapmat, which contains hapoltypes inferred from all
     *   full genotypes, index starts from 1.
     */
    void outputENTsol()
    {
        orghapmat = new CMatrix[2*rownumber + 1];
        for (int i = 1; i <= rownumber; i++) {
            orghapmat[2*i - 1].col = new int[columnnumber + 1];
            orghapmat[2*i].col = new int[columnnumber + 1];
        }

        int genIdx = 0;
        char tmpchar;
        for(int i=0;i<cc_graph.size();i++){
            for(int ii=0;ii<cc_graph[i]->size;ii++){
                for(int iii=0;iii<nr_snp;iii++){
                    tmpchar = cc_graph[i]->node[ii]->h[0][iii];
                    orghapmat[2*genIdx+1].col[iii+1] = atoi(&tmpchar);
                }
                for(int iii=0;iii<nr_snp;iii++){
                    tmpchar = cc_graph[i]->node[ii]->h[1][iii];
                    orghapmat[2*genIdx+2].col[iii+1] = atoi(&tmpchar);
                }
                genIdx ++;
            }
        }
        imputeMissingValue();
    }


    double mylog(int cov){
        if(cov <=0 || nr_gen <=0) return 0;
        double p = (double)cov;
        return p*log(p)/log((double)2);
    }

    /* compute overall entropy from hap_list */
    double get_coverage()
    {
        double entropy = 0.0;

        for(int i=0;i<nr_hap;i++){
            double p = (double)hap_table[i].coverage;
            entropy += p*log(p);
        }

        entropy *= (-1);	
        entropy /= log((double)2);
        return entropy;
    }

    void cc_save(CComp * c, int start, int mode){
        // iterate through each node and set its haplotypes (and template)

        for(int i=0;i<c->size;i++){
		
            // remove the old haplotypes
            if( mode == SAVE_OVER && ( MODE_count == 0 || (MODE_count == 1 && c->node[i]->orphan) || (MODE_count == 2 && !c->node[i]->orphan)) ){
                del_hap(hap_string2int(c->node[i]->h[0],c->node[i]->h_init[0],start-nr_locked));
                del_hap(hap_string2int(c->node[i]->h[1],c->node[i]->h_init[1],start-nr_locked));
            }
		
            // copy the min config into the haplotypes
            strncpy(c->node[i]->h[0]+start,c->min_cf1[i],nr_free);
            strncpy(c->node[i]->h[1]+start,c->min_cf2[i],nr_free);
            // copy the min config into the initial position
            strncpy(c->node[i]->h_init[0],c->min_cf1[i],nr_free);
            strncpy(c->node[i]->h_init[1],c->min_cf2[i],nr_free);
		
            // add the new ones in hap_list
            if( MODE_count == 0 || (MODE_count == 1 && c->node[i]->orphan) || (MODE_count == 2 && !c->node[i]->orphan) ){
                add_hap(hap_string2int(c->node[i]->h[0],start-nr_locked));
                add_hap(hap_string2int(c->node[i]->h[1],start-nr_locked));
            }
		
        }
    }

    /*
     *	save the configuration if its entropy is smaller than the given
     */
    bool cc_save_min(CComp * c, int i, int start, double new_cov, double * cur_cov){

        if(i==c->size){
            if(new_cov < (*cur_cov - 1e-9)) {
                *cur_cov = new_cov;
                return true;
            } else return false;
        }

        bool save = true;

        strncpy(c->node[i]->h[0]+start,c->min_cf1[i],nr_free); strncpy(c->node[i]->h[1]+start,c->min_cf2[i],nr_free);

        // --- remove the old haplotypes and add the new ones in hap_list
        int new1_id = hap_string2int(c->node[i]->h[0],start-nr_locked);
        int new2_id = hap_string2int(c->node[i]->h[1],start-nr_locked);
        int old1_id = hap_string2int(c->node[i]->h[0],c->node[i]->h_init[0],start-nr_locked);
        int old2_id = hap_string2int(c->node[i]->h[1],c->node[i]->h_init[1],start-nr_locked);
	
        int old1_cov, old2_cov, new1_cov, new2_cov;

        old1_cov = del_hap(old1_id);
        new_cov = new_cov + mylog(old1_cov+1) - mylog(old1_cov);
        old2_cov = del_hap(old2_id);
        new_cov = new_cov + mylog(old2_cov+1) - mylog(old2_cov);

        new1_cov = add_hap(new1_id);
        new_cov = new_cov + mylog(new1_cov-1) - mylog(new1_cov);
        new2_cov = add_hap(new2_id);
        new_cov = new_cov + mylog(new2_cov-1) - mylog(new2_cov);

        if(!cc_save_min(c,i+1,start,new_cov,cur_cov)){
            // restore the old haps
            strncpy(c->node[i]->h[0]+start,c->node[i]->h_init[0],nr_free);
            strncpy(c->node[i]->h[1]+start,c->node[i]->h_init[1],nr_free);
		
            add_hap(old1_id);
            add_hap(old2_id);
            del_hap(new1_id);
            del_hap(new2_id);
            save = false;
        } else {
            strncpy(c->node[i]->h_init[0],c->min_cf1[i],nr_free); strncpy(c->node[i]->h_init[1],c->min_cf2[i],nr_free);
        }
        return save;
    }

    /* generate int for haplotype seq front start to window size */
    int hap_string2int(char * hap , int start)
    {
        int ret_data = hap[start] - '0';
        for(int i=start+1;i<start+nr_free+nr_locked;i++){
            ret_data <<=1;
            ret_data += hap[i] - '0';
        }

        return ret_data;
    }

    /* generate int for haplotype using last (locked) portion of p and first (free) portion of c */
    int hap_string2int(char * p , char * c , int start)
    {
        int ret_data;

        if(nr_locked > 0) {
            ret_data = p[start] - '0';
            for(int i=start+1;i<start+nr_locked;i++){
                ret_data <<= 1;
                ret_data += p[i] - '0';
            }
            for(int i=0;i<nr_free;i++){
                ret_data <<= 1;
                ret_data += c[i] - '0';
            }
        } else {
            ret_data = c[0] - '0';
            for(int i=1;i<nr_free;i++){
                ret_data <<= 1;
                ret_data += c[i] - '0';
            }
        }

        return ret_data;
    }

    /*
     *	traverse the connected component and find haps that create minimal overall entropy
     */
    int traverse(CComp * c, int i, int seq_start, double cur_cov, double * min_cov, int mode){
	
        // we are at a leaf
        if(i == -1){
            c->num_sol++;
            if(mode == MIN_SOL){
                // if the coverage is smaller (or no solution has been found), return it
                if(cur_cov < (*min_cov - 1e-9)) {
                    *min_cov = cur_cov;
                    return 1;
                } else return 2;

            } else if(mode == RAND_SOL){
                // if this is the first solution - save it
                if(c->num_sol == 1){
                    return 1;
                    // otherwise, randomly save it
                    //_	} else if(unif_rand(c->num_sol) == 0){
                } else if(rand()%2 == 0) {
                    return 1;
                } else {
                    return 2;
                }
            }
            fprintf(stderr,"error: no mode selected for traversal\n");
            exit(0);
        }
	
        int save = 0; int cur_return;
        bool error = false;
        bool tmp_inherit_set[2]; tmp_inherit_set[0] = tmp_inherit_set[1] = false;

        bool h_set[2]; h_set[0] = c->node[i]->h_count[0] > 0; h_set[1] = c->node[i]->h_count[1] > 0;
        bool h_set_tmp[2]; h_set_tmp[0] = h_set_tmp[1] = false;
        bool h_neither = !(h_set[0] || h_set[1]);
	
        int old1_id, old2_id, h1_id, h2_id, h1_cov, h2_cov, old1_cov, old2_cov;
        double sav_cov;
	
        if(mode == MIN_SOL && ( MODE_count == 0 || (MODE_count == 1 && c->node[i]->orphan) || (MODE_count == 2 && !c->node[i]->orphan)) ){
            old1_id = hap_string2int(c->node[i]->h[0],c->node[i]->h_init[0],seq_start-nr_locked);
            old2_id = hap_string2int(c->node[i]->h[1],c->node[i]->h_init[1],seq_start-nr_locked);
        }
	
        // clear out the haplotypes so they can be re-detirmined
        for(int ii=seq_start;ii<seq_start+nr_free;ii++){
            if(!h_set[0]) c->node[i]->h[0][ii] = '\0';
            if(!h_set[1]) c->node[i]->h[1][ii] = '\0';
        }
	
        int consis_0, consis_1;
	
        int k;
	
        // if(h_set[0] && h_set[1]) k = seq_start + nr_free;
        k = seq_start;
	
        char cur_cmp;
	
        while(k>=seq_start){
            // start: has the entire sequence been filled?
            if(k==seq_start + nr_free){		
                consis_0 = consis_1 = -1;
			
                // set the transmitted haplotypes
                if(mode == MIN_SOL && ( MODE_count == 0 || (MODE_count == 1 && c->node[i]->orphan) || (MODE_count == 2 && !c->node[i]->orphan)) ){
                    // -- increment these haps into the hap_table
                    // save all hap ids

                    sav_cov = cur_cov;

                    h1_id = hap_string2int(c->node[i]->h[0],seq_start-nr_locked);
                    h2_id = hap_string2int(c->node[i]->h[1],seq_start-nr_locked);
                    // remove the original haps

                    old1_cov = del_hap(old1_id);
                    cur_cov = cur_cov + mylog(old1_cov+1) - mylog(old1_cov);
                    old2_cov = del_hap(old2_id);
                    cur_cov = cur_cov + mylog(old2_cov+1) - mylog(old2_cov);

                    // add the new haps
                    h1_cov = add_hap(h1_id);
                    cur_cov = cur_cov + mylog(h1_cov-1) - mylog(h1_cov);
                    h2_cov = add_hap(h2_id);
                    cur_cov = cur_cov + mylog(h2_cov-1) - mylog(h2_cov);
                }

                bool parent_done = false;
                int trav_ctr = 0;
			
                while(!parent_done){
                    if(consis_0 == 2 && consis_1 != 1){ // try both positions in both parents
                        switch (trav_ctr) {
                            case 0:
                                c->node[i]->inherit[0] = 0;
                                c->node[i]->inherit[1] = 0;
                                break;
                            case 1:
                                c->node[i]->inherit[0] = 0;
                                c->node[i]->inherit[1] = 1;
                                break;	
                            case 2:
                                c->node[i]->inherit[0] = 1;
                                c->node[i]->inherit[1] = 0;
                                break;
                            case 3:
                                c->node[i]->inherit[0] = 1;
                                c->node[i]->inherit[1] = 1;
                                parent_done = true;
                                break;
                        }
                    } else if(consis_0 == 2){ // try both positions in first parent
                        switch (trav_ctr) {
                            case 0:
                                c->node[i]->inherit[0] = 0;
                                break;
                            case 1:
                                c->node[i]->inherit[0] = 1;
                                parent_done = true;
                                break;
                        }
                    } else if(consis_1 == 2){ // try both positions in second parent
                        switch (trav_ctr) {
                            case 0:
                                c->node[i]->inherit[1] = 0;
                                break;
                            case 1:
                                c->node[i]->inherit[1] = 1;
                                parent_done = true;
                                break;
                        }
                    } else { // only try the preset positions
                        parent_done = true;
                    }
				
                    // copy the haplotypes into the parent
                    if(consis_0 != -1) {
                        if(++c->node[i]->pp[0]->h_count[c->node[i]->inherit[0]] == 1) {
                            strncpy(c->node[i]->pp[0]->h[c->node[i]->inherit[0]]+seq_start,c->node[i]->h[0]+seq_start,nr_free);
                        }
                    }
                    if(consis_1 != -1) {
                        if(++c->node[i]->pp[1]->h_count[c->node[i]->inherit[1]] == 1) {
                            strncpy(c->node[i]->pp[1]->h[c->node[i]->inherit[1]]+seq_start,c->node[i]->h[1]+seq_start,nr_free);
                        }
                    }
			
                    // go to the next node
                    cur_return = traverse(c,i-1,seq_start,cur_cov,min_cov,mode);
                    if( cur_return != 0 ){
                        parent_done = true; // we have found a solution, stop swapping
                        if(cur_return == 1) {
                            // an entropy is returned - save current configuration
                            strncpy(c->min_cf1[i],c->node[i]->h[0]+seq_start,nr_free);
                            strncpy(c->min_cf2[i],c->node[i]->h[1]+seq_start,nr_free);

                            c->node[i]->sav_inherit[0] = -1;
                            c->node[i]->sav_inherit[1] = -1;
                            save = 1;
                        } else if(save == 0) { save = 2; }
                    }
				
                    // lower the counts of the parent
                    if(consis_0 != -1) c->node[i]->pp[0]->h_count[c->node[i]->inherit[0]]--;
                    if(consis_1 != -1) c->node[i]->pp[1]->h_count[c->node[i]->inherit[1]]--;
				
                    trav_ctr++;
                }

                if(mode == MIN_SOL && ( MODE_count == 0 || (MODE_count == 1 && c->node[i]->orphan) || (MODE_count == 2 && !c->node[i]->orphan)) ){
                    // --- restore the old haps into the hap table
                    // add the old haps
                    add_hap(old1_id);
                    add_hap(old2_id);
                    // remove the new haps
                    del_hap(h1_id);
                    del_hap(h2_id);
                    // restore cur_cov
                    cur_cov = sav_cov;
                }
			
                if(h_set[0] && h_set[1]) return save; // both haplotypes are pre-set, we are done here
                else k--; // otherwise, continue to the next hap variant
            } else {
                // start: found an ambiguity - fill it
			
                // if both haplotypes are set, make sure they are consistent with the sequence
                if(h_set[0] && h_set[1]) {
                    if(c->node[i]->seq[k] == '2'){
                        if(c->node[i]->h[0][k] != compliment(c->node[i]->h[1][k])) return false;
                    } 
                    k++; continue;
                }
			
                if(c->node[i]->t[0][k] == '?' || c->node[i]->t[1][k] == '?'){
                    if(c->node[i]->t[0][k] != '?' && !h_set[0]){
                        if(h_set[1]){
                            if(c->node[i]->h[0][k] == c->node[i]->t[0][k]) {
                                c->node[i]->h[0][k] = '\0';
                                k--; continue;
                            } else {
                                c->node[i]->h[0][k] = c->node[i]->t[0][k];
                                k++; continue;
                            }
                        } else { c->node[i]->h[0][k] = c->node[i]->t[0][k]; }
					
                        // otherwise temporarily set this one
                        h_set[0] = true; h_neither = false;
                        h_set_tmp[0] = true; 
                    }
                    if(c->node[i]->t[1][k] != '?' && !h_set[1]){
                        if(h_set[0]){
                            if(c->node[i]->h[1][k] == c->node[i]->t[1][k]) {
                                c->node[i]->h[1][k] = '\0';
                                k--; continue;
                            } else {
                                c->node[i]->h[1][k] = c->node[i]->t[1][k];
                                k++; continue;
                            }
                        } else { c->node[i]->h[1][k] = c->node[i]->t[1][k]; }
					
                        // otherwise temporarily set this one
                        h_set[1] = true; h_neither = false;
                        h_set_tmp[1] = true;
                    }
					
                    // start: iterate through each char combination
                    if(c->node[i]->seq[k] == '2'){
                        if(h_set[0]){ // one hap has been set, generate the compliment for it
                            cur_cmp = compliment(c->node[i]->h[0][k]);
                            if(c->node[i]->h[1][k] == cur_cmp){
                                c->node[i]->h[1][k] = '\0';
                                k--; continue;
                            } else { c->node[i]->h[1][k] = cur_cmp; k++; }					
                        } else 
                            if(h_set[1]){
                                cur_cmp = compliment(c->node[i]->h[1][k]);
                                if(c->node[i]->h[0][k] == cur_cmp){
                                    c->node[i]->h[0][k] = '\0';
                                    k--; continue;
                                } else { c->node[i]->h[0][k] = cur_cmp; k++; }
                            } else { // neither hap has been set , generate both haplotypes		
                                if(c->node[i]->h[0][k] == '0' && c->node[i]->h[1][k] == '1'){
                                    c->node[i]->h[0][k] = '1';
                                    c->node[i]->h[1][k] = '0';
                                    k++;
                                }else
                                    if(c->node[i]->h[0][k] == '1' && c->node[i]->h[1][k] == '0'){
                                        c->node[i]->h[0][k] = '\0';
                                        c->node[i]->h[1][k] = '\0';
                                        k--; continue;
                                    }else
                                    {
                                        c->node[i]->h[0][k] = '0';
                                        c->node[i]->h[1][k] = '1';
                                        k++;
                                    }
                            }
                    } else
                        if(c->node[i]->seq[k] == '?'){		
                            if( c->node[i]->h[0][k] == '0' && c->node[i]->h[1][k] == '0' ){
                                if(h_set[0] || h_neither){
                                    c->node[i]->h[0][k] = '0';
                                    c->node[i]->h[1][k] = '1';
                                } else { // h1 is set
                                    c->node[i]->h[0][k] = '1';
                                }
                                k++;
                            }else
                                if( c->node[i]->h[0][k] == '0' && c->node[i]->h[1][k] == '1'){
                                    if(h_neither){
                                        c->node[i]->h[0][k] = '1';
                                        c->node[i]->h[1][k] = '0';
                                        k++;
                                    } else if(h_set[1]) {
                                        c->node[i]->h[0][k] = '1';
                                        k++;
                                    } else { // h0 is set
                                        c->node[i]->h[1][k] = '\0';
                                        k--; continue;
                                    }
                                }else
                                    if(c->node[i]->h[0][k] == '1' && c->node[i]->h[1][k] == '0'){
                                        if(h_set[0] || h_neither){							
                                            c->node[i]->h[0][k] = '1';
                                            c->node[i]->h[1][k] = '1';
                                            k++;
                                        } else { // h1 is set
                                            c->node[i]->h[0][k] = '\0';
                                            k--; continue;
                                        }
                                    }else
                                        if(c->node[i]->h[0][k] == '1' && c->node[i]->h[1][k] == '1'){
                                            if(!h_set[0]) c->node[i]->h[0][k] = '\0';
                                            if(!h_set[1]) c->node[i]->h[1][k] = '\0';
                                            k--; continue;
                                        }else
                                        {
                                            if(!h_set[0]) c->node[i]->h[0][k] = '0';
                                            if(!h_set[1]) c->node[i]->h[1][k] = '0';
                                            k++;
                                        }
                        } // end: iterate through each char combination
				
                    if(h_set_tmp[0]){ h_set[0] = false; h_set_tmp[0] = false; h_neither = true;}
                    if(h_set_tmp[1]){ h_set[1] = false; h_set_tmp[1] = false; h_neither = true;}
				
                } else if(c->node[i]->t[0][k] == c->node[i]->h[0][k] && c->node[i]->t[1][k] == c->node[i]->h[1][k]){
                    if(!h_set[0]) c->node[i]->h[0][k] = '\0';
                    if(!h_set[1]) c->node[i]->h[1][k] = '\0';
                    k--; continue;
                } else {
                    if(!h_set[0]) { c->node[i]->h[0][k] = c->node[i]->t[0][k];}
                    if(!h_set[1]) { c->node[i]->h[1][k] = c->node[i]->t[1][k];}
                    k++;
                } // end: found an ambiguity - fill it

                // does this match the template of parent's haplotype it comes from?
                if(c->node[i]->inherit_set[0]){
                    if( c->node[i]->pp[0]->t[c->node[i]->inherit[0]][k-1] != '?' && 
                        c->node[i]->pp[0]->t[c->node[i]->inherit[0]][k-1] != c->node[i]->h[0][k-1] ) { k--; continue; }
                }
                if(c->node[i]->inherit_set[1]){
                    if( c->node[i]->pp[1]->t[c->node[i]->inherit[1]][k-1] != '?' && 
                        c->node[i]->pp[1]->t[c->node[i]->inherit[1]][k-1] != c->node[i]->h[1][k-1] ) { k--; continue; }
                }
				
            } // end: has the entire sequence been filled?
        } // end: while

        return save;
    }

    char compliment(char c){
        if(c == '0') return '1';
        if(c == '1') return '0';
        return '?';
    }

    /*
     *  use info from parents to replace as many '?' as possible
     */
    bool fill_template(Gen * g){
        /* assume t[0] comes from p[0] and t[1] comes from p[1] always */
        bool compSet;
        bool error;

        for(int i=0;i<nr_snp;i++){
            error = false;
            if(g->seq[i] == '2' || g->seq[i] == '?'){		
                compSet = false;
                g->t[0][i] = g->t[1][i] = '?';
            } else if(g->seq[i] == '0' || g->seq[i] == '1'){ 
                g->t[0][i] = g->t[1][i] = g->seq[i];
            } else {
                fprintf(stderr,"sequence error: must be of 0/1/2/? format\n");
                exit(0);
            }
        }
        return true;
    }

    /* 
     *	combine two connected components into one
     */
    void merge_cc(CComp * c1, CComp * c2){
        for(int i=0;i<c1->member.size();i++){
            c1->member[i]->cc = c2;
            c2->member.push_back(c1->member[i]);
        }
        // remove the 1st CC and renumber the rest
        cc_graph.erase(cc_graph.begin()+c1->id);
        for(int i=0;i<cc_graph.size();i++){
            cc_graph[i]->id = i;
        }
        delete c1;
    }

    /*
     *	start with a node and build connected component for that node
     */
    CComp* get_cc(Gen * g, CComp * c){

        // the current node is not in a component - add it to c
        if(g->cc == 0){
            g->cc = c;
            c->member.push_back(g);
            // traverse its children
            for(int i=0;i < g->child.size();i++){
                c = get_cc(g->child[i],c);
            }
            // this node is in a different component  - merge the two components
        } else if(g->cc->id != c->id) {
            merge_cc(c,g->cc);
            return g->cc;
        }
        // otherwise, the node is already in this component - do nothing
        return c;
    }

    /*
     *	swap two elements within a connected component
     */
    void cc_swap(int cc, int a, int b){
        if(a==b) return;
        Gen * x = cc_graph[cc]->member[a];
        cc_graph[cc]->member[a] = cc_graph[cc]->member[b];
        cc_graph[cc]->member[b] = x;
    }

    /*
     *	use id to add to the coverage of hap in table
     *	or create the new hap
     *	return NEW coverage
     */
    int add_hap(int id){

        for(int i=0;i<nr_hap;i++){
            if(hap_table[i].id == id){
                hap_table[i].coverage++;
                return hap_table[i].coverage;
            }
        }

        hap_table[nr_hap].id = id;
        hap_table[nr_hap].coverage = 1;

        nr_hap++;
        return 1;
    }

    /*
     *	remove a haplotype based on its ID
     *	returns the NEW coverage
     */
    int del_hap(int id){

        // see if hap is in the table
        for(int i=0;i<nr_hap;i++){
            if(hap_table[i].id == id){
                if(hap_table[i].coverage == 1){
                    remove_hap(i);
                } else {
                    hap_table[i].coverage--;
                    return hap_table[i].coverage;
                }
            }
        }
        return 0;
    }

    /* resize the array when deleting a haplotype */
    void remove_hap(int i){
        for(int ctr=i;ctr<nr_hap;ctr++){
            hap_table[ctr] = hap_table[ctr+1];
        }
        nr_hap--;
    }

    /*
     * read genotypes from _inputmat and create topological connected components for all genotypes.
     * _inputmat contains full genotypes from input data, index starts from 1.
     */
    void read_genotypes(void)
    {
        int order_ctr = 0;

        nr_gen = rownumber;
        nr_snp = columnnumber;
        //	assert(nr_snp <= MAXNOSITES);
        //	assert(nr_gen <= MAXNOGENOTYPES);

        if (noRuns == 0)  // temporary fix the memory leak
            gen_array = new Gen[noGenotype+10000];
        else
            init_ENT();  // initilize global variables
        noRuns ++;

        hap_table = new Hap[2*nr_gen];
        gen_array[0].id = 0;

        // --- REGULAR INPUT ---
        // read each genotype from _inputmat
        int tmpvalue;
        for(int i = 1; i <= nr_gen; i++){
            gen_array[i].id = i;
            for (int j = 0; j < nr_snp; j++) {
                tmpvalue = inputmat[i].col[j+1];
                if (tmpvalue < 0)  // missing value
                    gen_array[i].seq[j] = '?';
                else if (tmpvalue == 0)
                    gen_array[i].seq[j] = '0';
                else if (tmpvalue == 1)
                    gen_array[i].seq[j] = '1';
                else
                    gen_array[i].seq[j] = '2';
            }
            gen_array[i].seq[nr_snp] = '\0';  // genotype sequence
        }		

        // set all parent/child pointers
        int cur_nr_gen = nr_gen;
        for(int i=1;i<=cur_nr_gen;i++){
            gen_array[i].pp[0] = 0;
            gen_array[i].pp[1]= 0;
            gen_array[i].orphan = true;
            gen_array[0].child.push_back(&gen_array[i]);
        }

        // --- isolate connected components
        // traverse all nodes with no parents and build their connected components
        for(int i=0;i<gen_array[0].child.size();i++){
            CComp * new_cc = new CComp(cc_graph.size());  // new_cc.id = cc_graph.size()
            cc_graph.push_back(new_cc);
            get_cc(gen_array[0].child[i],new_cc);
        }

        // order each connected component
        for(int i=0;i<cc_graph.size();i++){
            // create the min configuration array to qual the number of members in the component
            cc_graph[i]->min_cf1 = new char*[cc_graph[i]->member.size()];
            cc_graph[i]->min_cf2 = new char*[cc_graph[i]->member.size()];
            // initialize empty string for each member
            for(int ii=0;ii<cc_graph[i]->member.size();ii++){
                cc_graph[i]->min_cf1[ii] = new char[nr_free];
                cc_graph[i]->min_cf2[ii] = new char[nr_free];
            }
            order_ctr = 0;
            // loop until every element in the CC has been ordered
            while(order_ctr < cc_graph[i]->member.size()){
                // find nodes which have no parents and mark them
                for(int ii=0;ii<cc_graph[i]->member.size();ii++){

                    if(cc_graph[i]->member[ii]->flag == F_REMOVED) continue;

                    // if it has passed all checks, flag it for removal
                    cc_graph[i]->member[ii]->flag = F_MARKED;
                }
                // remove all flagged nodes
                for(int ii=0;ii<cc_graph[i]->member.size();ii++){
                    if(cc_graph[i]->member[ii]->flag == F_MARKED){
                        cc_graph[i]->member[ii]->flag = F_REMOVED;
                        cc_swap(i,order_ctr,ii);
                        order_ctr++;
                    }
                }
            }
        }

        // convert each connected component to an array (the array will be used from now on)
        bool removed = false;
	
        for(int i=0;i<cc_graph.size();i++){
            cc_graph[i]->size = cc_graph[i]->member.size();
            cc_graph[i]->node = new Gen*[cc_graph[i]->size];
            for(int ii=0;ii<cc_graph[i]->size;ii++){
                cc_graph[i]->node[ii] = cc_graph[i]->member[ii];
                if(!fill_template(cc_graph[i]->node[ii])){
                    fprintf(stderr,"mendelian inconsistency at No.%d\n",cc_graph[i]->node[ii]->id);
                    if(!MODE_drop) exit(0);
                    fprintf(stderr,"dropping pedigree\n");
                    // remove this pedigree
                    nr_gen -= cc_graph[i]->size;
                    cc_graph.erase(cc_graph.begin() + i);
                    i--;
                    removed = true;
                    num_dropped++;
                    break;
                }
            }
            if(removed) removed = false;
            else cc_graph[i]->member.clear();
        }
    }

    /* initilize global variables for ENT, if it is NOT run for the first time */
    void init_ENT(void)
    {
        nr_cc = 0, nr_hap = 0, nr_free = 0, nr_locked = 0; 
        MODE_drop = false, MODE_count = 1;  
        num_dropped = 0, num_merrors = 0;

        for (int i = 0; i <= nr_gen; i++) {
            gen_array[i].seq[nr_snp] = '\0';
            gen_array[i].t[0][0] = gen_array[i].t[1][0] = '\0';
            gen_array[i].flag = false;
            gen_array[i].cc = 0; 
            gen_array[i].id = 0;
            gen_array[i].h[0][0] = gen_array[i].h[1][0] = '\0';
            gen_array[i].h_init[0][0] = gen_array[i].h_init[1][0] = '\0';
            gen_array[i].old_saved = false;
            gen_array[i].inherit[0] = gen_array[i].inherit[1] = -1;
            gen_array[i].sav_inherit[0] = gen_array[i].sav_inherit[1] = -1;
            gen_array[i].inherit_set[0] = gen_array[i].inherit_set[1] = false;
            gen_array[i].orphan = false;
            gen_array[i].child.clear();
        }
        cc_graph.clear();
    }

    void insert_sort_cc(CComp * c, CComp ** arr, int size){
        for(int i=0;i<size;i++){
            if(c->cur_cov < arr[i]->cur_cov){
                for(int ii=size;ii>i;ii--){
                    arr[ii] = arr[ii-1];
                }
                arr[i] = c;
                return;
            }
        }
        arr[size] = c;
    }

    Gen::Gen() {
        seq = new char[nr_snp+500];
        t[0] = new char[nr_snp+500];
        t[1] = new char[nr_snp+500];
        h[0] = new char[nr_snp+500];
        h[1] = new char[nr_snp+500];
        h_init[0] = new char[nr_free];
        h_init[1] = new char[nr_free];		
        flag = false;
        cc = 0; 
        id = 0;
        h[0][0] = h[1][0] = h_init[0][0] = h_init[1][0] = '\0';
        old_saved = false;
        inherit[0] = inherit[1] = sav_inherit[0] = sav_inherit[1] = -1;
        inherit_set[0] = inherit_set[1] = false;
        orphan = false;
    }

    void specialcase(void)
    {
        // Create the complete solution of the input genotype matrix.
        // In this special case every input genotype contains missing data.
        CMatrix *completeSln;
        completeSln = new CMatrix[2*noGenotype];  // start from index 0
        for (int i = 0; i < noGenotype; ++i) {
            completeSln[2*i].col = new int[columnnumber];
            completeSln[2*i+1].col = new int[columnnumber];
            for (int j = 0; j < columnnumber; ++j) {
                if (genMatMV[i].col[j] == 0) {
                    completeSln[2*i].col[j] = 0;
                    completeSln[2*i+1].col[j] = 0;
                }
                else if (genMatMV[i].col[j] == 1) {
                    completeSln[2*i].col[j] = 1;
                    completeSln[2*i+1].col[j] = 1;
                }
                else if (genMatMV[i].col[j] == 2) {
                    completeSln[2*i].col[j] = rand()%2;
                    completeSln[2*i+1].col[j] = 1 - completeSln[2*i].col[j];
                }
                else {  // missing value
                    completeSln[2*i].col[j] = rand()%2;
                    completeSln[2*i+1].col[j] = rand()%2;
                }
            }
        }
    }


} // namespace {}

namespace pph {

    CompatibleCache::CompatibleCache(const ImmutableMatrix &haplotypes)
        : ::CompatibleCache(haplotypes)
    {
    }


    void
    CompatibleCache::calc_right(int i) const
    {
        int j;
        for (j = i+1;	
             j < i_haplotypes.noCols()
                 && hasPPHSolution(i_haplotypes, i, j);
             ++j)
            ;
	
        i_right[i] = j-1;	// j is now the first column _not_
        // compatible with i.
    }


    int getNumOfClasses(const MatrixView &haplotypes, int left, int right)
    {
		int errorCode = pph_main(haplotypes, left, right);
		if (errorCode != 0)    // no PPH solution
	    	return errorCode;  // release memory in _getCCSolution
		else
	    	return countClasses();  // release memory in _getPPHSolution
    }


    void getPPHSolution(int solIdx, int noPPHSolutions, Matrix &matrix)
    {
		getOnePPHSln(solIdx);
		imputeMissingValue();

		matrix.resize(2*noGenotype, columnnumber);
		for (int i = 0; i < 2*noGenotype; ++i)
	    	for (int j = 0; j < columnnumber; ++j)
				matrix(i, j) = completeSln[i].col[j];

		for (int i = 0; i < 2*noGenotype; ++i)
			delete [] completeSln[i].col;
		delete [] completeSln;
	
		if (solIdx == noPPHSolutions)  releasemem(0);
    }
    
    
    void getCCSolution(int errorCode, Matrix &matrix)
    {
    	main_ENT();
    	delete [] hap_table;

    	matrix.resize(2*noGenotype, columnnumber);
		for (int i = 0; i < 2*noGenotype; ++i)
	    	for (int j = 0; j < columnnumber; ++j)
				matrix(i, j) = completeSln[i].col[j];

		for (int i = 0; i < 2*noGenotype; ++i)
			delete [] completeSln[i].col;
		delete [] completeSln;
		releasemem(errorCode);
    }

} // namespace pph
