/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "config.hh"
#include "qt_score.hh"
#include <snpfile/matrix.hh>
using namespace BiRC::SNPFile;
#include "tree.hh"

#include <cmath>
#include <limits>
#include <gsl/gsl_cdf.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_nan.h>

#ifndef VECTOR_INCLUDED
# include <vector>
# define VECTOR_INCLUDED
#endif


  // --- Local, generally used, functions --------------------------------
//static
long double PleafQT(Tree &node, long double sigma){
    long double avg = 0.0;
    size_t n = node.getQtValues().size();
    assert(n > 0);

    for(size_t i = 0; i < n; ++i)
	avg += node.getQtValues()[i];
    avg /= n;

    long double logL = 0.0;
    for(size_t i = 0; i < n; ++i)
	logL += logl(gsl_ran_gaussian_pdf(avg - node.getQtValues()[i], sigma));

    return -2.0*logL;
}


long double Qleaf(Tree &node, int nIndividuals, const QTValue &qt_value){
    std::vector<bool> cluster(nIndividuals, false);
    int m = node.getQtValues().size();
    for(int i=0; i<m; i++){
	assert(cluster[(int)node.getQtValues()[i]]==false);
	cluster[(int)node.getQtValues()[i]]=true;
    }
    
    int n = nIndividuals - m;

    if(m==0 or n==0){
	return 0.0L;
    }

    long double avgC = 0.0;
    long double avgR = 0.0;

    for(int i = 0; i < nIndividuals; i++){
	if(cluster[i]){
	    avgC += qt_value.value(i);
	}
	else{
	    avgR += qt_value.value(i);
	}
    }
    avgC /= m;
    avgR /= n;

    long double varC = 0.0;
    long double varR = 0.0;
    
    for(int i = 0; i < nIndividuals; i++){
	if(cluster[i]){
	    varC += (avgC - qt_value.value(i))*(avgC - qt_value.value(i));
	}
	else{
	    varR += (avgR - qt_value.value(i))*(avgR - qt_value.value(i));
	}
    }
    return (((avgC - avgR)*sqrtl(m+n-2))/
	    sqrtl((varC + varR)*((1/(long double)m)+(1/(long double)n))));
}


// --- QTValue ---------------------------------------------------------
void 
QTValue::permuteValues()
{
    std::random_shuffle(i_values.begin(), i_values.end());

}


// --- QTScore ---------------------------------------------------------
void
QTScore::permute()
{
    if (!i_backup_qtvalues.get())
	i_backup_qtvalues.reset(new QTValue(i_qtvalues));
    i_qtvalues.permuteValues();
}

void
QTScore::resetPermutation()
{ 
    if (i_backup_qtvalues.get())
	i_qtvalues = *i_backup_qtvalues;
}





// --- Q Score --------------------------------------------------------
namespace {
    struct QScoreVisitor : TreeVisitor {
	const QTValue &qt_value;
	const int nIndividuals;
	const double m;
	QScoreVisitor(const QTValue &qt_value,
		 const int nIndividuals)
	    : qt_value(qt_value), nIndividuals(nIndividuals), m(1.0) 
	{}
	virtual void visitInnerNode(InnerNode &node);
	virtual void visitLeafNode(LeafNode &node);
    };


    void QScoreVisitor::visitInnerNode(InnerNode &node)
    {
	std::vector<double>& LeftData = node.getLeftChild()->getQtValues();
	std::vector<double>& RightData = node.getRightChild()->getQtValues();

	node.getQtValues().clear();
	for (unsigned int i = 0; i< LeftData.size(); i++){
	    node.getQtValues().push_back(LeftData[i]);
	}
	for (unsigned int i = 0; i< RightData.size(); i++){
	    node.getQtValues().push_back(RightData[i]);
	}
	
	long double leaf = Qleaf(node, nIndividuals, qt_value);
	long double left =  node.getLeftChild()->getScore();
	long double right = node.getRightChild()->getScore();

	if(left>right)
	    {
		if(left>leaf){
		    node.setScore(left);
		}
		else{
		    node.setScore(leaf);
		}
	    }
	else
	    {
		if(right>leaf){
		    node.setScore(right);
		}
		else{
		    node.setScore(leaf);
		}
	    }
    }
    

    void QScoreVisitor::visitLeafNode(LeafNode &node)
    {
	LeafNode::IDsIterator i;
	
	node.getQtValues().clear();
	for (i = node.leafIDsBegin(); i != node.leafIDsEnd(); ++i){
	    node.getQtValues().push_back(*i);
	}
	node.setScore(Qleaf(node, nIndividuals, qt_value));
    }
}

long double
QTQScore::scoreTree(Tree *t) const
{
    int nIndividuals = i_qtvalues.size();
    QScoreVisitor sv(i_qtvalues, nIndividuals);
    t->acceptVisitor(sv);
    return t->getScore() ;
}



namespace {
    struct CountVisitor : TreeVisitor {
	const QTValue &qt_value;
	uint dfs;
	CountVisitor(const QTValue &qt_value) : 
	    qt_value(qt_value), dfs(0) {}
	virtual void visitInnerNode(InnerNode &node);
	virtual void visitLeafNode(LeafNode &node);
    };
    
    
    void CountVisitor::visitInnerNode(InnerNode &node)
    {
	node.setAffected(dfs);
	dfs++;
    }
    

    void CountVisitor::visitLeafNode(LeafNode &node)
    {
	node.setAffected(dfs);
	dfs++;
	LeafNode::IDsIterator i;
	
	node.getQtValues().clear();
	for (i = node.leafIDsBegin(); i != node.leafIDsEnd(); ++i){
	    node.getQtValues().push_back(qt_value.value(*i));
	}
    }
}

void get_clusters(Tree *node, std::vector<std::vector<double> > &clusters, 
		  std::vector<bool> &is_mutation, uint cur_cluster){

    InnerNode* IN;
    LeafNode* LN;
    
    if(is_mutation[node->getAffected()]){
	std::vector<double> newC;
	clusters.push_back(newC);
	cur_cluster = clusters.size()-1;
    }
    IN = dynamic_cast<InnerNode*>(node);
    if(IN!=0){
	get_clusters(IN->getLeftChild(), clusters, is_mutation, cur_cluster);
	get_clusters(IN->getRightChild(), clusters, is_mutation, cur_cluster);
    }
    else{
	LN = dynamic_cast<LeafNode*>(node);
	std::vector<double>& data = node->getQtValues();
	for(uint i=0; i<data.size(); i++){
	    clusters[cur_cluster].push_back(data[i]);
	}
    }
}    

long double get_sigma(std::vector<std::vector<double> > &clusters){
    double sigma = 0.0; 
    int n =0;
    
    double avg =0.0;
    for(size_t i = 0; i < clusters.size(); i++){
	for(size_t j = 0; j < clusters[i].size(); j++){
	    avg += clusters[i][j];
	}
	n += clusters[i].size();
    }
    avg /= n;

    for(size_t i = 0; i < clusters.size(); i++){
	for(size_t j = 0; j < clusters[i].size(); j++)
	    sigma += (avg - clusters[i][j]) * (avg - clusters[i][j]);
    }
    
    return sqrt(sigma/(long double)(n-1));
    
}

long double get_sigma2(const QTValue &qt_value){
    double sigma = 0.0; 
    double avg =0.0;
    for(size_t i = 0; i < qt_value.size(); i+=2){
	avg += qt_value.value(i)/2;
    }
    avg /= qt_value.size();
    for(size_t i = 0; i < qt_value.size(); i+=2){
	sigma += (avg - qt_value.value(i)) * (avg - qt_value.value(i));
    }
    return sqrt(sigma/(long double)((qt_value.size()/2)-1));
}


long double calculate_LogL(std::vector<std::vector<double> > &clusters, 
			   long double sigma){
   
    // Sigma could be 0 if all values in all the subtree are the same 
    // in this case the log-likelihood is also 0.
    /*
    if (sigma == 0) {
        return  0.0L; 
    }
    */
    long double logL = 0.0;
    for(size_t i = 0; i < clusters.size(); i++){
	long double avg =0.0;
	for(size_t j = 0; j < clusters[i].size(); j++)
	    avg += clusters[i][j];
	avg /= clusters[i].size();
	for(size_t j = 0; j < clusters[i].size(); j++)
	    logL += logl(gsl_ran_gaussian_pdf(avg - clusters[i][j], sigma));
    }
    return -2.0*logL;
}


// --- BIC Score -------------------------------------------------------

long double
QTnewBICScore::scoreTree(Tree *t) const
{
    int nIndividuals = i_qtvalues.size();
     CountVisitor cv(i_qtvalues);
    t->acceptVisitor(cv);
    double maxscore = - 100000.0;
    double score = 0.0;
    int tsize = t->getAffected();
    std::vector<bool> is_mutation(tsize+1, false);
    std::vector<std::vector<double> > clusters(1);
    get_clusters(t, clusters, is_mutation, 0); 
    int n_mut =0;
    long double sigma = get_sigma(clusters);
    long double leaf = calculate_LogL(clusters, sigma);
    for(int i = 1; i<tsize; i++){ /// Start == 0 ???
	is_mutation[i]=true;
	std::vector<std::vector<double> > clusters(1);
	get_clusters(t, clusters, is_mutation, 0); 

	score = leaf - calculate_LogL(clusters, sigma) - 
	    (clusters.size()-1) * logl((long double)nIndividuals);
	    //2 * (clusters.size()-1) * logl((long double)nIndividuals);
	    //4 * (clusters.size()-1) * logl(logl((long double)nIndividuals));

	if(score>maxscore){
	    maxscore = score;
	    n_mut = 1;
	}
	
	if(muts_to_check>1){
	    for(int j = i+1; j<tsize; j++){
		is_mutation[j]=true;
		std::vector<std::vector<double> > clusters(1);
		get_clusters(t, clusters, is_mutation, 0); 
		score = leaf - calculate_LogL(clusters, sigma) - 
		    (clusters.size()-1) * logl((long double)nIndividuals);
		//4 * (clusters.size()-1) * logl(logl((long double)nIndividuals));
		if(score>maxscore){
		    maxscore = score;
		    n_mut = 2;
		}
	    
	    	if(muts_to_check>2){
		    for(int k = j+1; k<tsize; k++){
			is_mutation[k]=true;
			std::vector<std::vector<double> > clusters(1);
			get_clusters(t, clusters, is_mutation, 0); 
			score = leaf - calculate_LogL(clusters, sigma) - 
			    (clusters.size()-1) * logl((long double)nIndividuals);
			if(score>maxscore){
			    maxscore = score;
			    n_mut = 3;
			}
			is_mutation[k]=false;
		    }
		}    
		is_mutation[j]=false;  
	    }
	}	
	is_mutation[i]=false;
    }
    return maxscore;
}





/* =================
 *  GENOTYPE SCORE:
 * ================= */

namespace {
    struct DfsVisitor : TreeVisitor {
	const QTValue &qt_value;
	std::vector<int>& hap_to_dfs;
	uint dfs;
	DfsVisitor(const QTValue &qt_value, std::vector<int>& hap_to_dfs) : 
	    qt_value(qt_value), hap_to_dfs(hap_to_dfs), dfs(0) {}
	virtual void visitInnerNode(InnerNode &node);
	virtual void visitLeafNode(LeafNode &node);
    };
    
    
    void DfsVisitor::visitInnerNode(InnerNode &node)
    {
	node.setAffected(dfs);
	dfs++;
    }
    

    void DfsVisitor::visitLeafNode(LeafNode &node)
    {
	node.setAffected(dfs);
	dfs++;
	LeafNode::IDsIterator i;
	for (i = node.leafIDsBegin(); i != node.leafIDsEnd(); ++i){
	    hap_to_dfs[*i] = dfs;
	}
    }
}


void get_dfs_clust(Tree *node, std::vector<bool> &is_mutation, 
		   uint &cur_cluster, std::vector<int> &dfs_to_cluster){
    InnerNode* IN;
    
    //std::cout << "Get_Dfs_Clust" << std::endl;
    //LeafNode* LN;  
    if(is_mutation[node->getAffected()]){
	cur_cluster++;
    }
    IN = dynamic_cast<InnerNode*>(node);
    if(IN!=0){
	get_dfs_clust(IN->getLeftChild(), is_mutation, 
		      cur_cluster, dfs_to_cluster);
	get_dfs_clust(IN->getRightChild(), is_mutation, 
		      cur_cluster, dfs_to_cluster);
    }
    dfs_to_cluster[node->getAffected()]=cur_cluster;
}


void get_diploid_clusters(std::vector<std::vector<double> > &diclusters,
			  uint n_individuals,
			  const QTValue &qt_value,
			  std::vector<int> &hap_to_dfs,
			  std::vector<int> &dfs_to_cluster){
    for(uint i=0; i<n_individuals/2; i++){
	int x = dfs_to_cluster[hap_to_dfs[i*2]];
	int y = dfs_to_cluster[hap_to_dfs[(i*2)+1]];
	assert(qt_value.value(i*2) == qt_value.value((i*2)+1));
	if(x<y){
	    diclusters[(x*(x+1)/2)+y].push_back(qt_value.value(i*2));
	}
	else{
	    diclusters[(y*(y+1)/2)+x].push_back(qt_value.value(i*2));
	}
    }
}


long double
QTgBICScore::scoreTree(Tree *t) const
{
    int nIndividuals = i_qtvalues.size();
    std::vector<int> hap_to_dfs(i_qtvalues.size());
    DfsVisitor dv(i_qtvalues, hap_to_dfs);
    t->acceptVisitor(dv);
    double maxscore = - 100000.0;
    double score = 0.0;
    int tsize = t->getAffected();
    std::vector<bool> is_mutation(tsize+1, false);
    std::vector<std::vector<double> > clusters(1);
    std::vector<int> dfs_to_cluster(tsize);
    uint tmp = 0;
    get_dfs_clust(t, is_mutation, tmp, dfs_to_cluster); 
    int n_mut =0;
    long double sigma = get_sigma2(i_qtvalues);
    get_diploid_clusters(clusters, nIndividuals, i_qtvalues, 
			 hap_to_dfs, dfs_to_cluster);
    long double leaf = calculate_LogL(clusters, sigma);
    for(int i = 1; i<tsize; i++){
	is_mutation[i]=true;
	std::vector<std::vector<double> > clusters(3);
	uint tmp = 0;
	get_dfs_clust(t, is_mutation, tmp, dfs_to_cluster); 
	get_diploid_clusters(clusters, nIndividuals, i_qtvalues, 
			    hap_to_dfs, dfs_to_cluster);
	int m = 0;
	for(size_t x = 0; x< clusters.size(); x++){
	    if(clusters[x].size()>0){
		m++;
	    }
	}
	score = leaf - calculate_LogL(clusters, sigma) - 
	    (m-1) * logl((long double)nIndividuals/2);
	if(score>maxscore){
	    maxscore = score;
	    n_mut = 1;
	}
	if(muts_to_check>1){
	    
	    for(int j = i+1; j<tsize; j++){
		is_mutation[j]=true;
		std::vector<std::vector<double> > clusters(6);
		uint tmp = 0;
		get_dfs_clust(t, is_mutation, tmp, dfs_to_cluster); 
		get_diploid_clusters(clusters, nIndividuals, i_qtvalues, 
				     hap_to_dfs, dfs_to_cluster);
		int m = 0;
		for(size_t x = 0; x< clusters.size(); x++){
		    if(clusters[x].size()>0){
			m++;
		    }
		}
		score = leaf - calculate_LogL(clusters, sigma) - 
		    (m-1) * logl((long double)nIndividuals/2);
		if(score>maxscore){
		    maxscore = score;
		    n_mut = 2;
		}
		if(muts_to_check>2){
		    
		    for(int k = j+1; k<tsize; k++){
			is_mutation[k]=true;
			std::vector<std::vector<double> > clusters(10);
			uint tmp = 0;
			get_dfs_clust(t, is_mutation, tmp, dfs_to_cluster); 
			get_diploid_clusters(clusters, nIndividuals, 
					     i_qtvalues, hap_to_dfs, 
					     dfs_to_cluster);
			int m = 0;
			for(size_t x = 0; x< clusters.size(); x++){
			    if(clusters[x].size()>0){
				m++;
			    }
			}
			score = leaf - calculate_LogL(clusters, sigma) - 
			    (m-1) * logl((long double)nIndividuals/2);
			if(score>maxscore){
			    maxscore = score;
			    n_mut = 3;
			}	
			is_mutation[k]=false;
		    }
		}
		is_mutation[j]=false;
	    }
	}
	is_mutation[i]=false;
    }
    return maxscore;
}



