/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef QT_SCORE_HH_INCLUDED
#define QT_SCORE_HH_INCLUDED

/** \file
 * \brief Functions for scoring quantitative trait datasets.
 */

#include "score.hh"

#ifndef CMATH_INCLUDED
# include <cmath>
# define CMATH_INCLUDED
#endif
#include<iostream>

long double PleafQT(Tree &node, long double sigma);

/// Representation of quantitative traits for a set of individuals.
class QTValue {
    /// Internal representation of the traits.
    std::vector<double> i_values;

public:
    /**
     * \brief Construct new trait value object based on a sequence of
     * trait values.
     *
     * The sequence of traits is given through a pair of iterators.
     * The values in the sequence are copied into the trait value
     * object, so changes to the sequence later on will not affect the
     * trait value object.
     *
     * \param valuesBegin       Iterator to the first value.
     * \param valuesEnd         Iterator one past the last value.
     */
    template <typename Itr>
    QTValue(Itr valuesBegin, Itr valuesEnd)
        : i_values(valuesBegin, valuesEnd)
    {
	
	long double avg = 0.0;
 	for(uint i = 0; i < i_values.size() ; i++){
	    avg += i_values[i];
	}
	avg /= i_values.size();
	sigma = 0.0; 
 	for(uint i=0; i<i_values.size(); i++){
	    sigma += (avg - i_values[i]) * (avg - i_values[i]);		
	}
	sigma = sqrtl(sigma/(long double)i_values.size());
	
    }

    /// Get total number of individuals in the trait values.
    size_t size() const { return 2 * i_values.size(); }
    
    // minimal difference between two different qt_values.
    long double sigma;
    
    /**
     * \brief Get the value for a given individual.
     * \param i         Individual to access.
     * \returns         The trait value for individual \a i.
     */
    double value(size_t i)   const { return i_values.at(i/2); }

    /**
     * \brief Permute values for a permutation test.
     *
     * The permutation destroys the original values.
     */
    void permuteValues();
};


/// Abstract class for scoring quantitative trait datasets
class QTScore : public BasicScore {
protected:
    /// Classification of dataset.
    QTValue &i_qtvalues;
    std::auto_ptr<QTValue> i_backup_qtvalues;

    QTScore(BiRC::SNPFile::ImmutableMatrix &haplotypes,
	    std::vector<int> &positions,
            int minMarkers, double maxlen, 
            QTValue &values)
        : BasicScore(haplotypes, positions, minMarkers, maxlen), 
          i_qtvalues(values), i_backup_qtvalues(0)
    {
    }

    virtual void permute();
    virtual void resetPermutation();
};


/// Q score
class QTQScore : public QTScore {
public:
    QTQScore(BiRC::SNPFile::ImmutableMatrix &haplotypes,
	     std::vector<int> &positions,
	     int minMarkers, double maxlen, QTValue &qt_values)
	: QTScore(haplotypes,positions,minMarkers,maxlen,qt_values)
    {
    }

protected:
    /// Calculate the probability score for \a tree.
    virtual long double scoreTree(Tree *tree) const;
};


/// new BIC score
class QTnewBICScore : public QTScore {
    int muts_to_check;
public:
    QTnewBICScore(BiRC::SNPFile::ImmutableMatrix &haplotypes, 
		  std::vector<int> &positions,
		  int minMarkers, double maxlen, int nmut, QTValue &qt_values)
	: QTScore(haplotypes,positions,minMarkers,maxlen,qt_values),  
	  muts_to_check(nmut)
    {
    }

protected:
    /// Calculate the probability score for \a tree.
    virtual long double scoreTree(Tree *tree) const;
};




/// genotype BIC score
class QTgBICScore : public QTScore {
    int muts_to_check;
public:
    QTgBICScore(BiRC::SNPFile::ImmutableMatrix &haplotypes, 
		std::vector<int> &positions,
		int minMarkers, double maxlen, int nmut, QTValue &qt_values)
	: QTScore(haplotypes,positions,minMarkers,maxlen,qt_values),
	  muts_to_check(nmut)
    {
    }

protected:
    /// Calculate the probability score for \a tree.
    virtual long double scoreTree(Tree *tree) const;
};



#endif
