/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "config.hh"
#include "score.hh"

#include <snpfile/array_matrix.hh>
using namespace BiRC::SNPFile;

#include "matrix_utils.hh"
#include "incompatibility.hh"
#include "tree.hh"
#include "pph.hh"
#include "tree_builder.hh"

#include <vector>
#include <iomanip>
#include <memory>

Score::~Score()
{
}

BasicScore::BasicScore(ImmutableMatrix &haplotypes, 
		       std::vector<int> &positions,
                       int minMarkers, double maxlen)
  : Score(haplotypes.isPhased()), // if phased, we score through
                                  // trees, otherwise not
    i_haplotypes(haplotypes), i_positions(positions), 
    i_minMarkers(minMarkers), i_maxlen(maxlen),
    i_cc(haplotypes.isPhased() 
         ? new CompatibleCache(haplotypes)
         : new pph::CompatibleCache(haplotypes))
{
    if (maxlen <= 0) maxlen = positions[positions.size()-1] - positions[0];
}



long double
BasicScore::scoreLocus(int locus) const
{
    if (i_haplotypes.isPhased())
	{
	    std::auto_ptr<Tree> t(getTree(locus));
	    //std::cout << *t << std::endl;
	    return scoreTree(t.get());
	}
    else
        {
	    int leftBreak, rightBreak;
	    getRegion(locus, leftBreak, rightBreak);
	    
	    /// Get the number of classes in the shadow tree for the
	    /// given region
	    int noSTClasses = pph::getNumOfClasses(i_haplotypes.view(), 
						   leftBreak, rightBreak-1);
	    
	    if(noSTClasses > 0)  // it is a compatible region
		{
		    if (noSTClasses > 10) noSTClasses = 10;
		    
		    int noPPHSolutions = 1 << (noSTClasses - 1); // 2**(noSTClasses-1)
		    Matrix pphSolution(newArrayMatrix(0,0,true));
		    long double averScore = 0;
		    long double totalScore = 0;
		
		    for (int solIdx = 0; solIdx < noPPHSolutions; ++solIdx) 
			{
			    pph::getPPHSolution(solIdx, 
						noPPHSolutions, 
						pphSolution);
			    MatrixView view = pphSolution.view();
			    assert(view.noRows() == 2*i_haplotypes.noRows());
			    std::auto_ptr<Tree> t(getTree(locus, view));
			    totalScore += scoreTree(t.get());
			}
		    averScore = totalScore/noPPHSolutions;
		    return averScore;
		}
	    else 
		{  // incompatible region
		    Matrix ccSolution(newArrayMatrix(0,0,true));
		    // _noSTClasses is used as _errorCode here
		    pph::getCCSolution(noSTClasses, ccSolution);
		    MatrixView view = ccSolution.view();
		    std::auto_ptr<Tree> t(getTree(locus, view));
		    return scoreTree(t.get());
		}
	}
}

void
BasicScore::getRegion(int locus, int &leftBreak, int &rightBreak) const
{
    ::calculateIncompatibilities(i_haplotypes, i_positions, *i_cc,
				 locus, leftBreak, rightBreak,
				 i_minMarkers, i_maxlen);
}

Tree *
BasicScore::getTree(int locus) const
{
    int leftBreak, rightBreak;
    getRegion(locus, leftBreak, rightBreak);
    MatrixView view = i_haplotypes.subMatrixColumns(leftBreak, rightBreak);
    return getTree(locus, view);
}


Tree *
BasicScore::getTree(int locus, MatrixView &view) const
{
    sortColumns(view, i_positions, locus);
    return ::buildTree(view, 0, locus, i_positions);
}


PermutationTestScore::PermutationTestScore(BasicScore &orig, 
					   unsigned long noPermTests) 
	: Score(orig.scoresThroughTree()),
	  i_original(orig), i_noPermTests(noPermTests)
{
}

long double
PermutationTestScore::scoreLocus(int locus) const
{
    if (i_original.scoresThroughTree())
	{
	    std::auto_ptr<Tree> t(i_original.getTree(locus));
	    return scoreTree(t.get());
	}

    else
	{
	    i_original.resetPermutation();
	    long double realScore = i_original.scoreLocus(locus);
	    unsigned long cnt = 0;
	    for (unsigned int i = 0; i < i_noPermTests; ++i)
		{
		    i_original.permute();
		    if (realScore <= i_original.scoreLocus(locus)) ++cnt;
		}
	    i_original.resetPermutation();
	    return static_cast<long double>(cnt) / i_noPermTests;
	}

    assert(!"We shouldn't reach this point!");
    return -1.0;
}

long double
PermutationTestScore::scoreTree(Tree *tree) const
{
    i_original.resetPermutation();
    long double realScore = i_original.scoreTree(tree);
    unsigned long cnt = 0;
    for (unsigned int i = 0; i < i_noPermTests; ++i)
	{
	    i_original.permute();
	    if (realScore <= i_original.scoreTree(tree)) ++cnt;
	}
    i_original.resetPermutation();
    return static_cast<long double>(cnt) / i_noPermTests;
}

// fixme: is this the best impl?
void
PermutationTestScore::permute() 
{}

void
PermutationTestScore::resetPermutation() 
{}

