/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef SCORE_HH_INCLUDED
#define SCORE_HH_INCLUDED

/** \file
 * \brief Functions for scoring association for local trees. 
 */

#include "compatible_cache.hh"

#include <vector>
#include <algorithm>
#include <memory>

#include "config.hh"
namespace BiRC { 
    namespace SNPFile { 
	class ImmutableMatrix;
	class Matrix;
	class MatrixView;
    }
}
class Tree;

/// Abstract interface for scoring loci
class Score {
protected:
    bool i_scoresThroughTree;
    Score(bool scoresThroughTree) : i_scoresThroughTree(scoresThroughTree)
    {}

public:
    virtual ~Score();

    /**
     * \brief Score a locus.
     *
     * Calculate the score (as specified in concrete sub-classes) for
     * a given locus.
     *
     * \param locus		Index of the locus to score.
     * \returns			Score for \a locus.
     */
    virtual long double scoreLocus(int locus) const = 0;

    /**
     * \brief Score a tree.
     *
     * Calculate the score (as specified in concrete sub-classes) for
     * a given tree.
     *
     * \param tree		A tree to score.
     * \returns			Score for \a locus.
     */
    virtual long double scoreTree(Tree *tree) const = 0;

    /// predicate specifying if this score should be used to score
    /// trees (SMA tests, for example, do not support this and should
    /// not be scored with a tree).
    bool scoresThroughTree() const { return i_scoresThroughTree; }

    /**
     * \brief Permute trait data for a permutation test.
     *
     * The permutation destroys the existing trait/haplotype
     * association.  It can be restored using resetPermutation().
     */
    virtual void permute() = 0;

    /**
     * \brief Reset trait data for a permutation test.
     */
    virtual void resetPermutation() = 0;

};

/// Concreate class for scoring (i.e. one that simply calculates a
/// score based on the underlying information criteria).
class BasicScore : public Score {
protected:
    /// Reference to underlying haplotypes.
    BiRC::SNPFile::ImmutableMatrix &i_haplotypes;
    /// Reference to haplotypes positions.
    std::vector<int> &i_positions;
    /// Minimum markers to include when building regions
    int i_minMarkers;
    /// Maximal length of regions
    double i_maxlen;
    /// Cache for speeding up region calculations
    std::auto_ptr<CompatibleCache> i_cc;

    BasicScore(BiRC::SNPFile::ImmutableMatrix &haplotypes,
	       std::vector<int> &positions,
	       int minMarkers, double maxlen);

public:

    /**
     * \brief Score a locus.
     *
     * Calculate the score (as specified in concrete sub-classes) for
     * a given locus.
     *
     * \param locus		Index of the locus to score.
     * \returns			Score for \a locus.
     */
    virtual long double scoreLocus(int locus) const;

    /**
     * \brief Calculate the region, around a locus, to build trees
     * for.
     *
     * \param locus             Index of the marker this region is for.
     * \param leftBreak         Reference that will be assigned the index of
     *                          the leftmost marker in the region.
     * \param rightBreak        Reference that will be assigned the index
     *                          one past the rightmost marker in the region.
     */
    virtual void getRegion(int locus, int &leftBreak, int &rightBreak) const;

    /**
     * \brief Calculate a tree for a locus.
     *
     * \param locus		Index of the locus.
     * \returns			A local tree for this locus.
     */
    virtual Tree *getTree(int locus) const;

    /**
     * \brief Calculate a tree for a locus (used in the PPH version)
     *
     * \param locus		Index of the locus.
     * \param view		A matrix encoding a current PPH solution
     * \returns			A local tree for this locus.
     */
    virtual Tree *getTree(int locus, BiRC::SNPFile::MatrixView &view) const;

    /**
     * \brief Score a tree.
     *
     * Calculate the score (as specified in concrete sub-classes) for
     * a given tree.
     *
     * \param tree		A tree to score.
     * \returns			Score for \a locus.
     */
    virtual long double scoreTree(Tree *tree) const = 0;

};


/**
 * \brief Permutation test scores.
 *
 * Class wrapping a score class used to obtain empirical p-values
 * through permutation tests.
 *
 * \see Score
 */
class PermutationTestScore : public Score {
    BasicScore &i_original; // used for getting the non-permuted score
    unsigned long i_noPermTests;

public:
    PermutationTestScore(BasicScore &orig, unsigned long noPermTests);

    virtual long double scoreLocus(int locus) const;
    virtual long double scoreTree(Tree *tree) const;

    virtual void permute();
    virtual void resetPermutation();
};




#endif
