/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include "config.hh"
#include "io.hh"
#include "tree.hh"
#include "tree_builder.hh"
#include "incompatibility.hh"
#include "score.hh"
#include "case_control_score.hh"
#include "genotype_score.hh"
#include "qt_score.hh"
#include "config.hh"

#include <snpfile/snpfile.hh>
#include <snpfile/matrix.hh>
#include <snpfile/file_matrix.hh>
#include <snpfile/metadata_access.hh>
using namespace BiRC::SNPFile;

#include <boost/program_options.hpp>
#include <boost/serialization/vector.hpp>
#include <boost/lexical_cast.hpp>

#include <numeric>
#include <string>
#include <cmath>
#include <fstream>
#include <cstdio>
#include <sys/time.h>
#include <ext/algorithm>
#include <signal.h>



namespace options {
    bool verbose = false;
    std::string inputFileName;
    std::string outputFileName;

    int noPerms = 0;
    int minMarkers = 10;
    double maxlen = -1.0;

    std::string scoreFunction;
    std::string interval;

    int maxMutations = 1;

    std::string phenotypeMetadataKey;
}

static int
getInterval(int &from, int &to, int noMarkers)
{
    if (options::interval == "")
    {
        from = 0;
        to = noMarkers;
        return 0;
    }

    std::string &s = options::interval;
    std::string::size_type pos = s.find(":",0);
    if (pos == std::string::npos)
    {
        from = to = atol(s.c_str());
    }
    else
    {
        from = atol(s.substr(0,pos).c_str());
        to = atol(s.substr(pos+1,s.size()).c_str());
    }

    if (from < 0) return 1;
    if (to > noMarkers) return 1;
    return 0;
}


// signal handling to shut down properly if the user asks for termination
namespace {
    volatile bool termination_signalled = false;
}
static void 
signal_handler(int signum)
{
    termination_signalled = true;
}

static BasicScore *
selectScoringFunction(ImmutableMatrix &haplotypes,
                      std::vector<int> &positions,
                      Classification &classification)
{
    switch (options::scoreFunction[0])
	{
	case 'S': return new SMAScore(haplotypes, positions,
                                      options::minMarkers, options::maxlen, 
                                      classification);
        case 'B': return new BICScore(haplotypes, positions,
                                      options::minMarkers, options::maxlen, 
                                      classification);
        case 'P': return new ProbScore(haplotypes, positions,
                                       options::minMarkers, options::maxlen,
                                       classification);
        case 'H': return new HQScore(haplotypes, positions,
                                     options::minMarkers, options::maxlen, 
                                     classification);
	case 'A': return new AICScore(haplotypes, positions,
                                      options::minMarkers, options::maxlen,
                                      classification);
	case 'N': return new newBICScore(haplotypes, positions,
					options::minMarkers, 
					 options::maxlen, 
					 options::maxMutations, 
					 classification);
	case 'W': return new newHQScore(haplotypes, positions,
					 options::minMarkers, 
					 options::maxlen, 
					 options::maxMutations, 
					 classification);
        case 'G': return new GiniScore(haplotypes, positions,
                                       options::minMarkers, options::maxlen,
                                       classification);
        case 'X': return new GenotypeScore(haplotypes, positions,
                                           options::minMarkers, 
                                           options::maxlen, 
                                           classification,
                                           options::maxMutations);
    }
    assert(!"shouldn't reach this point!");
    return 0;
}

static BasicScore *
selectScoringFunction(ImmutableMatrix &haplotypes, 
		      std::vector<int> &positions,
		      QTValue &qt_values)
{

    switch (options::scoreFunction[0])
	{
	case 'Q': return new QTQScore(haplotypes, positions,
				      options::minMarkers, options::maxlen,
				      qt_values);
	case 'B': return new QTnewBICScore(haplotypes, positions,
					   options::minMarkers, 
					   options::maxlen, 
					   options::maxMutations, qt_values);
 	case 'X': return new QTgBICScore(haplotypes, positions,
					 options::minMarkers, 
					 options::maxlen, 
					 options::maxMutations, qt_values);
	}
    assert(!"Shouldn't reach this point!");
    return 0;
}


int
main(int argc, char *argv[])
{
    using namespace TextManip;
    namespace po = boost::program_options;

    po::options_description visible("Allowed options");
    po::options_description hidden("Allowed options");
    po::options_description all("Allowed options");
    po::variables_map vm;
    visible.add_options()
        ("help,h", "Produce help message.")
        ("verbose,v","Turns on verbose output.")
        
        ("scoreFunction,f",
         po::value<std::string>(&options::scoreFunction),
	 "Score funcition (H = HQC, B = BIC, A = AIC, G = Gini, S = SMA, "
	 "P = Prob, or X = eXperimental genotype likelihood score) "
	 "(default: H if #haplotypes > 400, otherwise P). For Quantitative "
	 "data: (B = BIC independent chromosomes, X = BIC chromosomes in pairs"
	 ", Q)")
        ("interval,i",
         po::value<std::string>(&options::interval),
         "Specify sub-interval to score (only markers in the interval will be "
         "scored/used for permutation tests).  The interval is given in the "
         "form <from>:<to>.")    

        ("minmarkers,m",
         po::value<int>(&options::minMarkers)->default_value(10),
         "Minimum number of markers used (default: 10).")
        ("deslen,x",
         po::value<double>(&options::maxlen)->default_value(-1),
         "Maximum length of interval (default: whole range).")

        ("max-mutations,z",
         po::value<int>(&options::maxMutations)->default_value(1),
         "Maximal number of mutations to consider when using the -fX score.")

        ("phenotype-metadata-key,k",
         po::value<std::string>(&options::phenotypeMetadataKey)->default_value("phenotype"),
         "Key under which phenotype data is stored")
	
        ("permutations,p",
         po::value<int>(&options::noPerms)->default_value(0),
         "Perform this number of permutations in a permutation test.")
	;
    hidden.add_options()
        ("input",
         po::value<std::string>(&options::inputFileName),
         "File where data is read from.")
        ("output",
         po::value<std::string>(&options::outputFileName),
         "File where scores will be written to.")
        ;
    all.add(visible).add(hidden);
    po::positional_options_description p;
    p.add("input", 1);
    p.add("output", 1);
    
    try {
        po::store(po::command_line_parser(argc, argv)
                  .options(all).positional(p).run(), 
                  vm, false);
        po::notify(vm);
    } catch (std::exception &e) {
        std::cerr << "Error parsing options:" << e.what()
                  << std::endl
                  << "Usage: " << argv[0] << " [options] input output"
                  << std::endl
                  << visible << std::endl;
        return 2;
    }

    if (vm.count("help")) {
        std::cout << "Usage: " << argv[0] << " [options] input output"
                  << std::endl
                  << visible << std::endl;
        return 0;
    }


    if (options::inputFileName == "")
    {
        std::cerr << "No input file specified."
                  << std::endl
                  << "Usage: " << argv[0] << " [options] input output"
                  << std::endl
                  << visible << std::endl;
        return 2;
    }
    if (options::outputFileName == "")
    {
        std::cerr << "No output file specified."
                  << std::endl
                  << "Usage: " << argv[0] << " [options] input output"
                  << std::endl
                  << visible << std::endl;
        return 2;
    }

    std::ofstream output(options::outputFileName.c_str());

    options::verbose = vm.count("verbose") > 0;

    SNPFile input(options::inputFileName);



    if (options::verbose)
        std::cout << bold
                  << "Running " << PACKAGE_NAME << " " << PACKAGE_VERSION
                  << reset
                  << std::endl << std::endl;

    struct timeval tv; struct timezone tz;
    gettimeofday(&tv,&tz);
    unsigned int random_seed = tv.tv_usec;
    std::srand(random_seed);  

    std::vector<int>  positions;
    std::vector<std::string>  markerNames;

    ImmutableMatrix genotypes(newReadOnlyFileMatrix(input));

    getMetadata(input, "positions", positions);



    if(input.containsMetadata("markerNames"))
    {
        getMetadata(input, "markerNames", markerNames);
    }
    else
    {
	std::cout << "no marker names" << std::endl;
        using boost::lexical_cast;
        for (size_t i = 0; i < positions.size(); ++i)
            markerNames.push_back("marker"+lexical_cast<std::string>(i));

    }

    std::auto_ptr<BasicScore> scoreFun;
    std::auto_ptr<QTValue> qt_value;
    std::auto_ptr<Classification> classification;


    if(!input.containsMetadata(options::phenotypeMetadataKey)){
	std::cerr << "Unknown phenotypeMetadataKey: "
		  << options::phenotypeMetadataKey
		  << std::endl;
	return 2;
    }
    
    std::string type = const_cast<const SNPFile &>(input).getMetadataBuffer(options::phenotypeMetadataKey).getType();
    
    if (type == "std::vector< bool >")
	{
	    if(!vm.count("scoreFunction"))
		{
		    if(input.noRows() > 400)
			options::scoreFunction = "H";
		    else
			options::scoreFunction = "P";
		}
	    else {
		
		if(options::scoreFunction[0]!='B' &&
		   options::scoreFunction[0]!='P' &&
		   options::scoreFunction[0]!='N' &&
		   options::scoreFunction[0]!='W' &&
		   options::scoreFunction[0]!='H' &&
		   options::scoreFunction[0]!='A' &&
		   options::scoreFunction[0]!='S' &&
		   options::scoreFunction[0]!='G' &&
		   options::scoreFunction[0]!='X')
		    {
			std::cerr << "Not a valid name for a scoreFunction, "
				  << "use B, P, H, N, G, S, A, or X." 
				  << std::endl;
			return 2;
		    }
	    }

	    
	    std::vector<bool> diseaseStatus;

	    getMetadata(input, options::phenotypeMetadataKey, diseaseStatus);
	        
	    classification.reset(new Classification(diseaseStatus.begin(),
						    diseaseStatus.end()));
   
	    scoreFun.reset(selectScoringFunction(genotypes,
						 positions, 
						 *classification));
	   
	}
    else if (type == "std::vector< double >"){
	
	if(!vm.count("scoreFunction"))
	    {
		options::scoreFunction = "B";
	    }
	else{
	    if(options::scoreFunction[0]!='B' &&
	       options::scoreFunction[0]!='X' &&
	       options::scoreFunction[0]!='Q')
		{
		    std::cerr << "Not a valid name for a scoreFunction "
			      << "on quantitative data use B, X or Q." 
			      << std::endl;
		    return 2;
		}
	}
	
	if((options::scoreFunction[0] == 'B' ||
	    options::scoreFunction[0] == 'X') &&
	   (options::maxMutations<1 || options::maxMutations>3)){
	    std::cerr << "max mutations , "
		      << "should be between 1 and 3" << std::endl;
	    return 2;
	}
	
	std::vector<double> phenotypes;
	
	getMetadata(input, options::phenotypeMetadataKey, phenotypes);
	
	
	qt_value.reset(new QTValue(phenotypes.begin(), phenotypes.end()));

	scoreFun.reset(selectScoringFunction(genotypes,
					     positions, 
					     *qt_value));
    }
    else{

	std::cerr << "The phenotypes should be of type bool or double" 
		  << std::endl;
	return 2;
    }
    
    
    std::auto_ptr<Score> ptest(new PermutationTestScore(*scoreFun, options::noPerms));

    int from, to;
    if (0 != getInterval(from, to, genotypes.noCols()))
    {
        std::cerr << "Specified interval outside the range of the markers."
                  << std::endl;
        return 2;
    }


    termination_signalled = false;
    sighandler_t prev_handler = signal(SIGINT, signal_handler);
    signal(SIGKILL, signal_handler);

    using namespace TextManip;
    const int verbose_step_size = (genotypes.noCols() < 100) ?
        1 : genotypes.noCols() / 100;
    TextProgressBar progressBar;

    if (options::verbose) progressBar.start();
    output << "index position name score";
    if (options::noPerms > 0) output << " pvalue";
    output << std::endl;

    for (int i = from; i < to && !termination_signalled; ++i)
    {
        // FIXME: write more info to output file
        output << i << ' ' 
               << positions[i] << ' '
               << markerNames[i] << ' '
               << scoreFun->scoreLocus(i);
	if (options::noPerms > 0)
	    output << ' ' << ptest->scoreLocus(i);
	output << std::endl << std::flush;
        if (options::verbose and (i+1) % verbose_step_size == 0)
            progressBar.showProgress(i, genotypes.noCols());
    }
    if (options::verbose) progressBar.end();

    if (options::verbose && termination_signalled)
        std::cout << std::endl
                  << "Termination signal recieved; writing results so far and shutting down." << std::endl
                  << "Warning: the results are not reliable since the process was not allowed to" << std::endl
                  << "complete, so analyse them with care!" << std::endl;
    
    signal(SIGINT,  prev_handler);
    signal(SIGKILL, prev_handler);

    return 0;
}
