/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Sren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef TREE_HH_INCLUDED
#define TREE_HH_INCLUDED

/** \file
 * \brief Representation of local phylogenies.
 */


#include <iostream>
#include <vector>

struct TreeVisitor;

/**
 * \brief Abstract tree node.
 *
 * This class represents a sub-tree that can be either an inner node
 * or a leaf node.  As such, it is also implementing the Composite
 * Design Pattern.
 */
class Tree {
protected:
    long double score;
    size_t affected;
    size_t unaffected;
    int clusters;
    int level;
    int noNodes;
    int noInnerNodes;
    std::vector<double> QTData;

public:
    /** Construct a tree at a given level (tree depth).
     * \param level	Tree depth for this (sub-)tree.
     */
    Tree(int level, int noNodes, int noInnerNodes)
	: score(0), 
	  affected(0), unaffected(0), 
	  clusters(0), level(level), 
	  noNodes(noNodes), noInnerNodes(noInnerNodes),
	  QTData() 
    {
    }
    virtual ~Tree();

    /// Get the number of affected individuals in this (sub-)tree.
    size_t               getAffected()     const { return affected; }
    /// Get the number of unaffected individuals in this (sub-)tree.
    size_t               getUnaffected()   const { return unaffected; }
    /// Get the (sub-)score calculated from this (sub-)tree.
    long double          getScore()        const { return score; }
    /// Get the number of clusters in this (sub-)tree.
    int                  getClusters()     const { return clusters; }
    /// Get the tree depth of this sub-tree.
    int                  getLevel()        const { return level; }
    /// Get the number of nodes in this sub-tree.
    int                  getNoNodes()      const { return noNodes; }
    /// Get the number of inner nodes in this sub-tree.
    int                  getNoInnerNodes() const { return noInnerNodes; }


    /// Get the number of individuals in this (sub-)tree.
    size_t               getIndividuals() const { return affected+unaffected; }

    /// Get the list of qt-values in this (sub-) tree. 
    std::vector<double>& getQtValues()          { return QTData; }

    /** Set the number of affected individuals in this (sub-)tree.
     * There is no consistency check to ensure that this matches the sum
     * of affected in sub-trees under this (sub-)tree.
     *
     * \param newAffected	New affected count for this (sub-)tree.     
     */
    void setAffected(size_t newAffected)         { affected = newAffected; }
    /** Set the number of unaffected individuals in this (sub-)tree.
     * There is no consistency check to ensure that this matches the sum
     * of unaffected in sub-trees under this (sub-)tree.
     *
     * \param newUnaffected	New unaffected count for this (sub-)tree.     
     */
    void setUnaffected(size_t newUnaffected)     { unaffected = newUnaffected; }

    /** Set the score for this (sub-)tree.
     * \param newScore	New score for this (sub-)tree.
     */
    void setScore(long double newScore)       { score = newScore; }

    /** Set the number of clusters in this (sub-)tree.
     * There is no consistency check to ensure that this matches the sum
     * of clusters in sub-trees under this (sub-)tree.
     *
     * \param newClusters	New cluster count for this (sub-)tree.     
     */
    void setClusters(int newClusters)         { clusters = newClusters; }

    /**
     * \brief Traverse (sub-)tree.
     *
     * Use the Visitor Design Pattern to traverse this (sub-)tree.
     *
     * \param v		Visitor for the traversal.
     */
    virtual void acceptVisitor(TreeVisitor &v) = 0;

    /**
     * \brief Print this (sub-)tree to a stream.
     * \param os	Output stream to print the tree to.
     */
    virtual void print(std::ostream &os) const = 0;
};

inline std::ostream &
operator<<(std::ostream &os, const Tree &t)
{
    t.print(os); os << ';';
    return os;
}


/// An inner node sub-tree.
class InnerNode : public Tree {
    Tree *leftChild, *rightChild;
    double pos;
public:
    /**
     * Construct an inner node from its sub-trees.  The left/right
     * ordering of sub-trees is arbitrary.
     *
     * \param level		Tree depth of this node.
     * \param leftChild		Left sub-tree.
     * \param rightChild	Right sub-tree.
     * \param pos		Genomic position this tree is local for.
     */
    InnerNode(int level, Tree *leftChild, Tree *rightChild, double pos)
	: Tree(level,
	       leftChild->getNoNodes()+rightChild->getNoNodes()+1,
	       leftChild->getNoInnerNodes()+rightChild->getNoInnerNodes()+1),
	  leftChild(leftChild), rightChild(rightChild), pos(pos)
    {
    }
    virtual ~InnerNode();

    /// Get the left subtree.  The left/right ordering is arbitrary.
    Tree *getLeftChild()  const { return leftChild; }
    /// Get the right subtree.  The left/right ordering is arbitrary.
    Tree *getRightChild() const { return rightChild; }
    /// get the genomic position for this tree.
    double getPos()       const { return pos; }
    
    virtual void acceptVisitor(TreeVisitor &v);
    virtual void print(std::ostream &os) const;
};

/// An leaf tree node.
class LeafNode : public Tree {
    std::vector<size_t> i_leafIDs;

public:

    /**
     * Construct a leaf containing a set of individuals.
     *
     * \param level		Tree depth of this leaf.
     * \param leafIDsBegin	Iterator pointing to the first individual.
     * \param leafIDsEnd	Iterator pointing one past the last individual.
     */
    template <typename Itr>
    LeafNode(int level, Itr leafIDsBegin, Itr leafIDsEnd) 
	: Tree(level,1,0), i_leafIDs(leafIDsBegin, leafIDsEnd)
    {
    }
    /**
     * Construct a leaf containing a single individual.
     *
     * \param level	Tree depth of this leaf.
     * \param leafID	The single individual.
     */
    LeafNode(int level, size_t leafID) : Tree(level,1,0)
    {
	i_leafIDs.push_back(leafID);
    }

    /// Iterator type for the set of individuals contained in this leaf.
    typedef std::vector<size_t>::const_iterator IDsIterator;
    /// First individual in this leaf.
    IDsIterator leafIDsBegin() const { return i_leafIDs.begin(); }
    /// One past the last individual in this leaf.
    IDsIterator leafIDsEnd()   const { return i_leafIDs.end(); }

    virtual ~LeafNode();
    virtual void acceptVisitor(TreeVisitor &v);
    virtual void print(std::ostream &os) const;
};

/**
 * Abstract visitor for traversing trees.  Part of the Visitor Design
 * Pattern.
 */
struct TreeVisitor {
    virtual ~TreeVisitor();

    /** Callback for inner nodes.
     * \param n	The node being visited. */
    virtual void visitInnerNode(InnerNode &n) = 0;
    /** Callback for leaf nodes.
     * \param n	The node being visited. */
    virtual void visitLeafNode(LeafNode   &n) = 0;
};


#endif
