/*
 * Blossoc -- Association Mapping by Incompatibilities
 *
 * Copyright (C) 2006 by Thomas Mailund <mailund@mailund.dk>
 *                    and Søren Besenbacher <besen@birc.au.dk>
 *
 * Copyright (C) 2007 by Bioinformatics Research Center,
 *                       University of Aarhus,
 *                       Contact: Thomas Mailund <mailund@birc.au.dk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#ifndef TREE_BUILDER_HH_INCLUDED
#define TREE_BUILDER_HH_INCLUDED

/** \file
 * \brief Interface for building trees from haplotype matrix data.
 */

#include "config.hh"
#include <snpfile/matrix.hh>
#include <stdexcept>
#include <vector>


class Tree;

/**
 * \brief Exception throw if a matrix view is empty.
 *
 * It is an error to try to build a (sub-)tree for an empty matrix; if
 * it is attempted, this exception will be thrown.
 *
 * \see MatrixView
 * \see Tree
 */
struct EmptyMatrixView : std::logic_error {
    EmptyMatrixView() : std::logic_error("noRows() must be greater than zero.")
    {}
};

/**
 * \brief Construct a (sub-)tree representing a (sub-)matrix view.
 *
 * Construct a tree from a matrix view -- conceptually a perfect
 * phylogeny for a compatible matrix, but in cases where there are
 * incompatibilities (accepted to achieve a minimal matrix size) the
 * tree will have the same split represented as multiple mutations.
 *
 * \param view		A matrix view to build the tree over.
 * \param level		Depth of the (sub-)tree in the local phylogeny.
 * \param pos		Index of the locus this tree is local for.
 * \param positions	Vector of marker positions.
 * \param minLeafSize	The minimal size accepted for a leaf (to prevent too
 *                      small leaves when using counts to score the tree).
 *
 * \returns A new tree representing the matrix view.
 * \throws  EmptyMatrixView
 */
Tree *buildTree(BiRC::SNPFile::MatrixView &view, 
		int level, int pos, 
		const std::vector<int>& positions,
		int minLeafSize = 1);


#endif
